# Pose2Lux Pro

from __future__ import generators
import poser, os, inspect, time, operator, string, sys, re, warnings
import Tkinter as tk
import tkFileDialog, tkMessageBox, tkColorChooser, tkSimpleDialog, tkFont
from Tkinter import *
from math import *
try:
	warnings.filterwarnings("ignore")
	import xmllib
except: pass

# Use Numeric if available, otherwise use NumPy
try:
	import Numeric as num
except:
	import numpy.oldnumeric as num

# Compatibility with other versions of Python
try:
    from math import pi, radians
except:
    pi=3.1415926535897931
    def radians(angle):
        return ((float(angle) / 180.0) * pi)	
if not hasattr(__builtins__, 'True'):
    True = 1
    False = not True
if not hasattr(__builtins__, 'enumerate'):
    def enumerate(seq):
        i = 0
        for x in seq:
            yield i, x
            i = i + 1
if not hasattr(__builtins__, 'sum'):
    def sum(seq):
        s = 0
        for x in seq:
            s = s + x
        return s

# Global constants
NAME = "Pose2Lux pro"
DEBUG = False
TEST = False
P2L_VERSION = "0.8.9"
POSER_VERSION = int(float(poser.Version()))
HOME_FOLDER = "L:/Poser development/Plop"		# For Poser 6 users only
FORCE_HOME_FOLDER = False

# XML stuff
__all__ = [
    # public symbols
    "Comment",
    "dump",
    "Element", "ElementTree",
    "fromstring",
    "iselement", "iterparse",
    "parse",
    "PI", "ProcessingInstruction",
    "QName",
    "SubElement",
    "tostring",
    "TreeBuilder",
    "VERSION", "XML",
    "XMLParser", "XMLTreeBuilder",
    ]
VERSION = "1.2.6"
if sys.version[:3] == "1.5":
    _escape = re.compile(r"[&<>\"\x80-\xff]+") # 1.5.2
else:
    _escape = re.compile(eval(r'u"[&<>\"\u0080-\uffff]+"'))
_escape_map = {
    "&": "&amp;",
    "<": "&lt;",
    ">": "&gt;",
    '"': "&quot;",}
_namespace_map = {
    # "well-known" namespace prefixes
    "http://www.w3.org/XML/1998/namespace": "xml",
    "http://www.w3.org/1999/xhtml": "html",
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
    "http://schemas.xmlsoap.org/wsdl/": "wsdl",}
class _SimpleElementPath:
    # emulate pre-1.2 find/findtext/findall behaviour
    def find(self, element, tag):
        for elem in element:
            if elem.tag == tag:
                return elem
        return None
    def findtext(self, element, tag, default=None):
        for elem in element:
            if elem.tag == tag:
                return elem.text or ""
        return default
    def findall(self, element, tag):
        if tag[:3] == ".//":
            return element.getiterator(tag[3:])
        result = []
        for elem in element:
            if elem.tag == tag:
                result.append(elem)
        return result
ElementPath = _SimpleElementPath()
class _ElementInterface:
    tag = None
    attrib = None
    text = None
    tail = None

    def __init__(self, tag, attrib):
        self.tag = tag
        self.attrib = attrib
        self._children = []

    def __repr__(self):
        return "<Element %s at %x>" % (self.tag, id(self))

    def makeelement(self, tag, attrib):
        return Element(tag, attrib)

    def __len__(self):
        return len(self._children)

    def __getitem__(self, index):
        return self._children[index]

    def __setitem__(self, index, element):
        assert iselement(element)
        self._children[index] = element

    def __delitem__(self, index):
        del self._children[index]

    def __getslice__(self, start, stop):
        return self._children[start:stop]

    def __setslice__(self, start, stop, elements):
        for element in elements:
            assert iselement(element)
        self._children[start:stop] = list(elements)

    def __delslice__(self, start, stop):
        del self._children[start:stop]

    def append(self, element):
        assert iselement(element)
        self._children.append(element)

    def insert(self, index, element):
        assert iselement(element)
        self._children.insert(index, element)

    def remove(self, element):
        assert iselement(element)
        self._children.remove(element)

    def getchildren(self):
        return self._children

    def find(self, path):
        return ElementPath.find(self, path)

    def findtext(self, path, default=None):
        return ElementPath.findtext(self, path, default)

    def findall(self, path):
        return ElementPath.findall(self, path)
		
    def clear(self):
        self.attrib.clear()
        self._children = []
        self.text = self.tail = None

    def get(self, key, default=None):
        return self.attrib.get(key, default)

    def set(self, key, value):
        self.attrib[key] = value

    def keys(self):
        return self.attrib.keys()

    def items(self):
        return self.attrib.items()

    def getiterator(self, tag=None):
        nodes = []
        if tag == "*":
            tag = None
        if tag is None or self.tag == tag:
            nodes.append(self)
        for node in self._children:
            nodes.extend(node.getiterator(tag))
        return nodes
_Element = _ElementInterface
def Element(tag, attrib={}, **extra):
    attrib = attrib.copy()
    attrib.update(extra)
    return _ElementInterface(tag, attrib)
def SubElement(parent, tag, attrib={}, **extra):
    attrib = attrib.copy()
    attrib.update(extra)
    element = parent.makeelement(tag, attrib)
    parent.append(element)
    return element
def Comment(text=None):
    element = Element(Comment)
    element.text = text
    return element
def ProcessingInstruction(target, text=None):
    element = Element(ProcessingInstruction)
    element.text = target
    if text:
        element.text = element.text + " " + text
    return element
PI = ProcessingInstruction
class QName:
    def __init__(self, text_or_uri, tag=None):
        if tag:
            text_or_uri = "{%s}%s" % (text_or_uri, tag)
        self.text = text_or_uri
    def __str__(self):
        return self.text
    def __hash__(self):
        return hash(self.text)
    def __cmp__(self, other):
        if isinstance(other, QName):
            return cmp(self.text, other.text)
        return cmp(self.text, other)
class ElementTree:

	def __init__(self, element=None, file=None):
		assert element is None or iselement(element)
		self._root = element # first node
		if file:
			self.parse(file)

	def getroot(self):
		return self._root

	def _setroot(self, element):
		assert iselement(element)
		self._root = element

	def parse(self, source, parser=None):
		global POSER_VERSION
		if not hasattr(source, "read"):
			source = open(source, "rb")
		if not parser:
			if POSER_VERSION == 6:
				parser = SimpleTreeBuilder()
			else:
				parser = XMLTreeBuilder()
		while 1:
			data = source.read(32768)
			if not data:
				break
			parser.feed(data)
		self._root = parser.close()
		return self._root

	def getiterator(self, tag=None):
		assert self._root is not None
		return self._root.getiterator(tag)

	def find(self, path):
		assert self._root is not None
		if path[:1] == "/":
			path = "." + path
		return self._root.find(path)

	def findtext(self, path, default=None):
		assert self._root is not None
		if path[:1] == "/":
			path = "." + path
		return self._root.findtext(path, default)

	def findall(self, path):
		assert self._root is not None
		if path[:1] == "/":
			path = "." + path
		return self._root.findall(path)

	def write(self, file, encoding="us-ascii"):
		assert self._root is not None
		if not hasattr(file, "write"):
			file = open(file, "wb")
		if not encoding:
			encoding = "us-ascii"
		elif encoding != "utf-8" and encoding != "us-ascii":
			file.write("<?xml version='1.0' encoding='%s'?>\n" % encoding)
		self._write(file, self._root, encoding, {})

	def _write(self, file, node, encoding, namespaces):
		# write XML to file
		tag = node.tag
		if tag is Comment:
			file.write("<!-- %s -->" % _escape_cdata(node.text, encoding))
		elif tag is ProcessingInstruction:
			file.write("<?%s?>" % _escape_cdata(node.text, encoding))
		else:
			items = node.items()
			xmlns_items = [] # new namespaces in this scope
			try:
				if isinstance(tag, QName) or tag[:1] == "{":
					tag, xmlns = fixtag(tag, namespaces)
					if xmlns: xmlns_items.append(xmlns)
			except TypeError:
				_raise_serialization_error(tag)
			file.write("<" + _encode(tag, encoding))
			if items or xmlns_items:
				items.sort() # lexical order
				for k, v in items:
					try:
						if isinstance(k, QName) or k[:1] == "{":
							k, xmlns = fixtag(k, namespaces)
							if xmlns: xmlns_items.append(xmlns)
					except TypeError:
						_raise_serialization_error(k)
					try:
						if isinstance(v, QName):
							v, xmlns = fixtag(v, namespaces)
							if xmlns: xmlns_items.append(xmlns)
					except TypeError:
						_raise_serialization_error(v)
					file.write(" %s=\"%s\"" % (_encode(k, encoding),
											   _escape_attrib(v, encoding)))
				for k, v in xmlns_items:
					file.write(" %s=\"%s\"" % (_encode(k, encoding),
											   _escape_attrib(v, encoding)))
			if node.text or len(node):
				file.write(">")
				if node.text:
					file.write(_escape_cdata(node.text, encoding))
				for n in node:
					self._write(file, n, encoding, namespaces)
				file.write("</" + _encode(tag, encoding) + ">")
			else:
				file.write(" />")
			for k, v in xmlns_items:
				del namespaces[v]
		if node.tail:
			file.write(_escape_cdata(node.tail, encoding))
def iselement(element):
	return isinstance(element, _ElementInterface) or hasattr(element, "tag")
def dump(elem):
	if not isinstance(elem, ElementTree):
		elem = ElementTree(elem)
	elem.write(sys.stdout)
	tail = elem.getroot().tail
	if not tail or tail[-1] != "\n":
		sys.stdout.write("\n")
def _encode(s, encoding):
	try:
		return s.encode(encoding)
	except AttributeError:
		return s
def _raise_serialization_error(text):
	raise TypeError(
		"cannot serialize %r (type %s)" % (text, type(text).__name__)
		)
def _encode_entity(text, pattern=_escape):
	def escape_entities(m, map=_escape_map):
		out = []
		append = out.append
		for char in m.group():
			text = map.get(char)
			if text is None:
				text = "&#%d;" % ord(char)
			append(text)
		return string.join(out, "")
	try:
		return _encode(pattern.sub(escape_entities, text), "ascii")
	except TypeError:
		_raise_serialization_error(text)
def _escape_cdata(text, encoding=None, replace=string.replace):
	# escape character data
	try:
		if encoding:
			try:
				text = _encode(text, encoding)
			except UnicodeError:
				return _encode_entity(text)
		text = replace(text, "&", "&amp;")
		text = replace(text, "<", "&lt;")
		text = replace(text, ">", "&gt;")
		return text
	except (TypeError, AttributeError):
		_raise_serialization_error(text)
def _escape_attrib(text, encoding=None, replace=string.replace):
	# escape attribute value
	try:
		if encoding:
			try:
				text = _encode(text, encoding)
			except UnicodeError:
				return _encode_entity(text)
		text = replace(text, "&", "&amp;")
		text = replace(text, "'", "&apos;")
		text = replace(text, "\"", "&quot;")
		text = replace(text, "<", "&lt;")
		text = replace(text, ">", "&gt;")
		return text
	except (TypeError, AttributeError):
		_raise_serialization_error(text)
def fixname(name, split=string.split):
    # xmllib in 2.0 and later provides limited (and slightly broken)
    # support for XML namespaces.
    if " " not in name:
        return name
    return "{%s}%s" % tuple(split(name, " ", 1))
def fixtag(tag, namespaces):
	if isinstance(tag, QName):
		tag = tag.text
	namespace_uri, tag = string.split(tag[1:], "}", 1)
	prefix = namespaces.get(namespace_uri)
	if prefix is None:
		prefix = _namespace_map.get(namespace_uri)
		if prefix is None:
			prefix = "ns%d" % len(namespaces)
		namespaces[namespace_uri] = prefix
		if prefix == "xml":
			xmlns = None
		else:
			xmlns = ("xmlns:%s" % prefix, namespace_uri)
	else:
		xmlns = None
	return "%s:%s" % (prefix, tag), xmlns
def parse(source, parser=None):
	tree = ElementTree()
	tree.parse(source, parser)
	return tree
class iterparse:

	def __init__(self, source, events=None):
		global POSER_VERSION
		if not hasattr(source, "read"):
			source = open(source, "rb")
		self._file = source
		self._events = []
		self._index = 0
		self.root = self._root = None
		if POSER_VERSION == 6:
			self._parser = SimpleTreeBuilder()
		else:
			self._parser = XMLTreeBuilder()

		parser = self._parser._parser
		append = self._events.append
		if events is None:
			events = ["end"]
		for event in events:
			if event == "start":
				try:
					parser.ordered_attributes = 1
					parser.specified_attributes = 1
					def handler(tag, attrib_in, event=event, append=append,
								start=self._parser._start_list):
						append((event, start(tag, attrib_in)))
					parser.StartElementHandler = handler
				except AttributeError:
					def handler(tag, attrib_in, event=event, append=append,
								start=self._parser._start):
						append((event, start(tag, attrib_in)))
					parser.StartElementHandler = handler
			elif event == "end":
				def handler(tag, event=event, append=append,
							end=self._parser._end):
					append((event, end(tag)))
				parser.EndElementHandler = handler
			elif event == "start-ns":
				def handler(prefix, uri, event=event, append=append):
					try:
						uri = _encode(uri, "ascii")
					except UnicodeError:
						pass
					append((event, (prefix or "", uri)))
				parser.StartNamespaceDeclHandler = handler
			elif event == "end-ns":
				def handler(prefix, event=event, append=append):
					append((event, None))
				parser.EndNamespaceDeclHandler = handler

	def next(self):
		while 1:
			try:
				item = self._events[self._index]
			except IndexError:
				if self._parser is None:
					self.root = self._root
					try:
						raise StopIteration
					except NameError:
						raise IndexError
				# load event buffer
				del self._events[:]
				self._index = 0
				data = self._file.read(16384)
				if data:
					self._parser.feed(data)
				else:
					self._root = self._parser.close()
					self._parser = None
			else:
				self._index = self._index + 1
				return item

	try:
		iter
		def __iter__(self):
			return self
	except NameError:
		def __getitem__(self, index):
			return self.next()
def XML(text):
	global POSER_VERSION
	if POSER_VERSION == 6:
		parser = SimpleTreeBuilder()
	else:
		parser = XMLTreeBuilder()
	parser.feed(text)
	return parser.close()
fromstring = XML
def XMLID(text):
	global POSER_VERSION
	if POSER_VERSION == 6:
		parser = SimpleTreeBuilder()
	else:
		parser = XMLTreeBuilder()
	parser.feed(text)
	tree = parser.close()
	ids = {}
	for elem in tree.getiterator():
		id = elem.get("id")
		if id:
			ids[id] = elem
	return tree, ids
def tostring(element, encoding=None):
    class dummy:
        pass
    data = []
    file = dummy()
    file.write = data.append
    ElementTree(element).write(file, encoding)
    return string.join(data, "")
class TreeBuilder:

	def __init__(self, element_factory=None):
		self._data = [] # data collector
		self._elem = [] # element stack
		self._last = None # last element
		self._tail = None # true if we're after an end tag
		if element_factory is None:
			element_factory = _ElementInterface
		self._factory = element_factory

	def close(self):
		assert len(self._elem) == 0, "missing end tags"
		assert self._last != None, "missing toplevel element"
		return self._last

	def _flush(self):
		if self._data:
			if self._last is not None:
				text = string.join(self._data, "")
				if self._tail:
					assert self._last.tail is None, "internal error (tail)"
					self._last.tail = text
				else:
					assert self._last.text is None, "internal error (text)"
					self._last.text = text
			self._data = []

	def data(self, data):
		self._data.append(data)

	def start(self, tag, attrs):
		self._flush()
		self._last = elem = self._factory(tag, attrs)
		if self._elem:
			self._elem[-1].append(elem)
		self._elem.append(elem)
		self._tail = 0
		return elem

	def end(self, tag):
		self._flush()
		self._last = self._elem.pop()
		assert self._last.tag == tag,\
			   "end tag mismatch (expected %s, got %s)" % (
				   self._last.tag, tag)
		self._tail = 1
		return self._last
class SimpleTreeBuilder(xmllib.XMLParser):

	def __init__(self, html=0):
		self.__builder = TreeBuilder()
		if html:
			import htmlentitydefs
			self.entitydefs.update(htmlentitydefs.entitydefs)
		xmllib.XMLParser.__init__(self)

	##
	# Feeds data to the parser.
	#
	# @param data Encoded data.

	def feed(self, data):
		xmllib.XMLParser.feed(self, data)

	##
	# Finishes feeding data to the parser.
	#
	# @return An element structure.
	# @defreturn Element

	def close(self):
		xmllib.XMLParser.close(self)
		return self.__builder.close()

	def handle_data(self, data):
		self.__builder.data(data)

	handle_cdata = handle_data

	def unknown_starttag(self, tag, attrs):
		attrib = {}
		for key, value in attrs.items():
			attrib[fixname(key)] = value
		self.__builder.start(fixname(tag), attrib)

	def unknown_endtag(self, tag):
		self.__builder.end(fixname(tag))
class XMLTreeBuilder:

    def __init__(self, html=0, target=None):
        try:
            from xml.parsers import expat
        except ImportError:
            raise ImportError(
                "No module named expat; use SimpleXMLTreeBuilder instead"
                )
        self._parser = parser = expat.ParserCreate(None, "}")
        if target is None:
            target = TreeBuilder()
        self._target = target
        self._names = {}
        
        parser.DefaultHandlerExpand = self._default
        parser.StartElementHandler = self._start
        parser.EndElementHandler = self._end
        parser.CharacterDataHandler = self._data

        try:
            self._parser.buffer_text = 1
        except AttributeError:
            pass

        try:
            self._parser.ordered_attributes = 1
            self._parser.specified_attributes = 1
            parser.StartElementHandler = self._start_list
        except AttributeError:
            pass
        encoding = None
        if not parser.returns_unicode:
            encoding = "utf-8"

        self._doctype = None
        self.entity = {}

    def _fixtext(self, text):

        try:
            return _encode(text, "ascii")
        except UnicodeError:
            return text

    def _fixname(self, key):
 
        try:
            name = self._names[key]
        except KeyError:
            name = key
            if "}" in name:
                name = "{" + name
            self._names[key] = name = self._fixtext(name)
        return name

    def _start(self, tag, attrib_in):
        fixname = self._fixname
        tag = fixname(tag)
        attrib = {}
        for key, value in attrib_in.items():
            attrib[fixname(key)] = self._fixtext(value)
        return self._target.start(tag, attrib)

    def _start_list(self, tag, attrib_in):
        fixname = self._fixname
        tag = fixname(tag)
        attrib = {}
        if attrib_in:
            for i in range(0, len(attrib_in), 2):
                attrib[fixname(attrib_in[i])] = self._fixtext(attrib_in[i+1])
        return self._target.start(tag, attrib)

    def _data(self, text):
        return self._target.data(self._fixtext(text))

    def _end(self, tag):
        return self._target.end(self._fixname(tag))

    def _default(self, text):
        prefix = text[:1]
        if prefix == "&":

            try:
                self._target.data(self.entity[text[1:-1]])
            except KeyError:
                from xml.parsers import expat
                raise expat.error(
                    "undefined entity %s: line %d, column %d" %
                    (text, self._parser.ErrorLineNumber,
                    self._parser.ErrorColumnNumber)
                    )
        elif prefix == "<" and text[:9] == "<!DOCTYPE":
            self._doctype = []
        elif self._doctype is not None:

            if prefix == ">":
                self._doctype = None
                return
            text = string.strip(text)
            if not text:
                return
            self._doctype.append(text)
            n = len(self._doctype)
            if n > 2:
                type = self._doctype[1]
                if type == "PUBLIC" and n == 4:
                    name, type, pubid, system = self._doctype
                elif type == "SYSTEM" and n == 3:
                    name, type, system = self._doctype
                    pubid = None
                else:
                    return
                if pubid:
                    pubid = pubid[1:-1]
                self.doctype(name, pubid, system[1:-1])
                self._doctype = None

    def doctype(self, name, pubid, system):
        pass

    def feed(self, data):
        self._parser.Parse(data, 0)


    def close(self):
        self._parser.Parse("", 1)
        tree = self._target.close()
        del self._target, self._parser
        return tree
if POSER_VERSION == 6:
	XMLParser = SimpleTreeBuilder
else:
	XMLParser = XMLTreeBuilder	
	
# Geometry classes and functions (based on ODF's PyDough code)
def preprocess(geometry, exporter, options={}):
	if options.get('compute_normals', True) in [True, 1, '1', 'true']:
		exporter.UpdateProgressBar("Computing normals %s" % geometry.material_key)
		geometry.compute_normals()

	for i in xrange(int(options.get('subdivisionlevel', 0))):
		exporter.UpdateProgressBar("Subdividing %s" % geometry.material_key)
		geometry.subdivide()
def normalize(rows):
	norms = num.sqrt(num.maximum(num.sum(rows * rows, 1), 1e-16))
	return rows / norms[:, num.NewAxis]
def topological_order(children):
	seen = {}
	result = []

	def visit(v):
		if not seen.has_key(v):
			seen[v] = True
			for w in children[v]: visit(w)
			result.insert(0, v)

	is_child = {}
	for v in children.keys():
		for w in children[v]:
			is_child[w] = True
	
	for v in children.keys():
		if not is_child.has_key(v):
			visit(v)

	return result
def good_actor(actor):
	geom = actor.Geometry()
	return (actor.Visible() and not actor.IsProp() and geom
			and geom.NumVertices() > 0 and geom.NumPolygons() > 0)
def get_hair_group_for(actor):
	name = actor.InternalName()
	parent = actor.Parent()
	for i in xrange(parent.NumHairGroups()):
		g = parent.HairGroup(i)
		if name.startswith(g.Name()):
			return g
	return None
def get(subject):
	if isinstance(subject, poser.ActorType):
		figure = subject.ItsFigure()
		mesh = SimpleMesh(subject)
	elif isinstance(subject, poser.FigureType):
		figure = subject
		mesh = WeldedFigureMesh(subject)
	else:
		raise TypeError("Argument must be an actor or figure.")

	if mesh.polys:
		if hasattr(subject, 'IsHairProp') and subject.IsHairProp():
			group = get_hair_group_for(subject)
			if group == None:
				options = {}
			else:
				options = {
					'root_radius': group.RootWidth() * 1e-4,
					'tip_radius' : group.TipWidth()  * 1e-4
					}
			geom = HairGeometry(mesh.verts, mesh.polys, mesh.poly_mats,
							None, mesh.tverts, mesh.tpolys, options)
		else:
			geom = Geometry(mesh.verts, mesh.polys, mesh.poly_mats,
							None, mesh.tverts, mesh.tpolys)
		geom.materials = mesh.materials
		geom.material_key = subject.Name()
	else:
		geom = None
		
	return geom	
def write(file, geometry, keys, exporter):

	if geometry.is_empty:
		exporter.UpdateProgressBar("Ignoring %s" % geometry.material_key, increase=1)
		return
	
	for i, key in enumerate(keys):
	
		isPortal, processing_portals = exporter.CheckForPortal(key)
		
		if (isPortal and not processing_portals) or (processing_portals and not isPortal):
			exporter.UpdateProgressBar("Ignoring %s" % key, increase=1)
			continue
			
		if exporter.IsSkydome(key):
			exporter.UpdateProgressBar("Ignoring %s" % key, increase=1)
			continue
	
		exporter.UpdateProgressBar("Processing %s" % key)
		
		sub = geometry.extract_by_material(i)
		
		if not sub.is_empty:
		
			if not isinstance(geometry, HairGeometry):
				sub.convert_to_per_vertex_uvs()
				
				thickness = exporter.MakeBackside(key)
				if thickness: sub.add_backside(thickness)
			
			if not isPortal: 
				
				print >>file, '\nAttributeBegin'
			
				# Write out volume data
				print >>file, '\n', exporter.InteriorVolume(key)
				print >>file, exporter.ExteriorVolume(key), '\n'
				
				# Write out NamedMaterial reference
				print >>file, exporter.GetNamedMaterial(key)
			
				# Write Emitter parameters, if any
				if exporter.IsEmitter(key):
					texAttr, emitAttr, group, matfile = exporter.GetEmitter(key)
					print >>file, 'LightGroup "%s"' % group
					if texAttr: texAttr.Write(matfile)
					emitAttr.Write(file)
			
			# Check for subdivision
			attr = exporter.GetSubDText(key)
			
			# Check for displacement
			exporter.DoDisplacement(key, attr)
		
			print >>file, '\n%s "mesh"' % ("Shape", "PortalShape")[isPortal]
			attr.Write(file, nodesonly=True)

			print >>file, '"integer triindices" ['
			for u, v, w in sub.triangles: print >>file, u, v, w
			print >>file, ']\n'

			print >>file, ' "point P" ['
			for x, y, z in sub.points: print >>file, x, -z, y
			print >>file, ']\n'

			if sub.number_of_normals:
				print >>file, ' "normal N" ['
				for x, y, z in sub.normals: print >>file, x, -z, y
				print >>file, ']\n'

			if sub.number_of_texture_points:
				print >>file, ' "float uv" ['
				for u, v in sub.texture_points: print >>file, u, v
				#if addDepth:
				#	for u, v in sub.texture_points: print >>file, u, v
				print >>file, ']\n'
			
			if not isPortal: print >>file, 'AttributeEnd'
		
		exporter.UpdateProgressBar("Processing %s complete" % key, increase=1)
def cross(u, v):
	return num.array([u[1]*v[2] - u[2]*v[1],
					  u[2]*v[0] - u[0]*v[2],
					  u[0]*v[1] - u[1]*v[0]])
def make_fiber(verts_in, r_start = 0.01, r_end = 0.01, gon = 5):
	n = len(verts_in)
	directions = normalize(num.subtract(verts_in[1:], verts_in[:-1]))
	bisectors = normalize(num.add(directions[1:], directions[:-1]))

	d = directions[0]
	if abs(num.dot(d, [1,0,0])) > 0.9:
		u = cross(d, [0,1,0])
	else:
		u = cross(d, [1,0,0])
	v = cross(d, u)
	u, v = normalize(num.array([u, v]))
	a = 2 * pi / gon

	section = num.array([u * cos(a * i) + v * sin(a * i)
						 for i in xrange(gon)])

	verts_out = num.zeros([n * gon, 3], "double")
	verts_out[:gon] = section * r_start + verts_in[0]

	for i in xrange(n - 2):
		d = bisectors[i]
		u = directions[i]
		f = num.sum(section * d, 1)[:, num.NewAxis]
		section = section - f / num.dot(d, u) * u
		r = ((i + 1) * r_end + (n - i - 1) * r_start) / n
		verts_out[(i + 1)*gon : (i + 2)*gon] = section * r + verts_in[i + 1]

	d = directions[-1]
	section = section - num.sum(section * d, 1)[:, num.NewAxis] * d
	verts_out[-gon:] = section * r_end + verts_in[-1]

	return verts_out, fiber_polygons(n - 1, gon)
def fiber_polygons(n, m):
	polys = [None] * (2 + n * m)
	ring = num.zeros([m, 4], "int")
	for i in xrange(m):
		i1 = (i + 1) % m
		ring[i] = [i, i1, i1 + m, i + m]
	for k in xrange(0, n * m, m):
		polys[k : k + m] = (ring + k).tolist()
	polys[-2] = range(m-1, -1, -1)
	polys[-1] = range(n * m, (n + 1) * m)

	return polys
class WeldingInfo(object):
	def __init__(self, figure):
		self.figure = figure
		self.collect_info()

	def welding_order(self):
		after = {}
		for actor in self.figure.Actors():
			after.setdefault(actor.Name(), [])
			for goal in actor.WeldGoalActors():
				after.setdefault(goal.Name(), []).append(actor.Name())
		return topological_order(after)

	def collect_info(self):
		figure = self.figure

		self.map_vertex  = map_vertex = {}
		self.new_vertex  = new_vertex = {}
		self.map_tvertex = map_tvertex = {}
		self.new_tvertex = new_tvertex = {}

		count = 0
		tcount = 0

		for name in self.welding_order():
			actor = figure.Actor(name)
			if not good_actor(actor): continue

			nv = actor.Geometry().NumVertices()
			a_map = map_vertex[name] = [None] * nv
			a_new = new_vertex[name] = [False] * nv
			
			for goal, new_map in zip(actor.WeldGoalActors(), actor.WeldGoals()):
				if not good_actor(goal): continue
				g_map = map_vertex[goal.Name()]
				for v, w in enumerate(new_map):
					if w >= 0: a_map[v] = g_map[w]

			for v in xrange(nv):
				if a_map[v] is None:
					a_map[v] = count
					a_new[v] = True
					count += 1

			nt = actor.Geometry().NumTexVertices()
			map_tvertex[name] = [i + tcount for i in xrange(nt)]
			new_tvertex[name] = [True] * nt
			tcount += nt

		self.used_actors = [a for a in figure.Actors() if good_actor(a)]
		self.vertex_count = count
		self.tvertex_count = tcount
class WeldedFigureMesh(object):
	def __init__(self, figure):
		self.figure = figure
		self.collect_data()
		
	def convert_materials_for_figure(self, actors):
		material_key = self.figure.Name()
		material_index = {}
		map_material = {}
		materials = []
		for actor in actors:
			a_name = actor.Name()
			for i, mat in enumerate(actor.Geometry().Materials() or []):
				m_name = mat.Name()
				# -- materials from geometry are incorrect (hat tip Dizzi)
				mat = actor.Material(m_name)
				if not material_index.has_key(m_name):
					material_index[m_name] = len(materials)
					materials.append(mat)
				map_material[(a_name, i)] = material_index[m_name]

		return materials, map_material

	def collect_data(self):
		welding_info = WeldingInfo(self.figure)
		actors = welding_info.used_actors
		nr_verts = welding_info.vertex_count
		nr_tverts = welding_info.tvertex_count
		nr_polys = sum([actor.Geometry().NumPolygons() for actor in actors])
		nr_tpolys = sum([actor.Geometry().NumTexPolygons() for actor in actors])

		materials, map_material = self.convert_materials_for_figure(actors)
		self.materials = materials

		self.verts     = verts     = num.zeros([nr_verts, 3], "double")
		self.tverts    = tverts    = num.zeros([nr_tverts, 2], "double")
		self.polys     = polys     = [None] * nr_polys
		self.poly_mats = poly_mats = [None] * nr_polys
		self.tpolys    = tpolys    = [None] * nr_polys

		pcount  = 0
		tpcount = 0 

		for actor in actors:
			name = actor.Name()
			geom = actor.Geometry()
			sets = geom.Sets()
			tsets = geom.TexSets()

			new_vert = welding_info.new_vertex[name]
			map_vert = welding_info.map_vertex[name]
			for i, v in enumerate(geom.WorldVertices()):
				if new_vert[i]:
					verts[map_vert[i]] = [v.X(), v.Y(), v.Z()]

			for p in (geom.Polygons() or []):
				start = p.Start()
				indices = sets[start : start + p.NumVertices()]
				polys[pcount] = [map_vert[v] for v in indices]
				poly_mats[pcount] = map_material[(name, p.MaterialIndex())]
				pcount += 1

			if nr_tpolys:
				new_tvert = welding_info.new_tvertex[name]
				map_tvert = welding_info.map_tvertex[name]
				for i, v in enumerate(geom.TexVertices() or []):
					if new_tvert[i]:
						tverts[map_tvert[i]] = [v.U(), v.V()]

				for p in (geom.TexPolygons() or []):
					start = p.Start()
					indices = tsets[start : start + p.NumTexVertices()]
					tpolys[tpcount] = [map_tvert[v] for v in indices]
					tpcount += 1

				for i in xrange(geom.NumPolygons() - geom.NumTexPolygons()):
					tpolys[tpcount] = []
					tpcount += 1
class SimpleMesh(object):
	def __init__(self, actor):
		grab  = lambda p, sets: sets[p.Start() : p.Start() + p.NumVertices()]
		tgrab = lambda p, sets: sets[p.Start() : p.Start() + p.NumTexVertices()]
		
		geom = actor.Geometry()
		if geom and actor.Visible():
			verts  = geom.WorldVertices() or []
			tverts = geom.TexVertices() or []
			polys  = geom.Polygons() or []
			tpolys = geom.TexPolygons() or []
			sets   = geom.Sets()
			tsets  = geom.TexSets()

			self.verts  = num.array([[v.X(), v.Y(), v.Z()] for v in verts], 'd')
			self.tverts = num.array([[v.U(), v.V()] for v in tverts], 'd')
			self.polys  = [grab(p, sets) for p in polys]
			if tsets:
				self.tpolys = [tgrab(p, tsets) for p in tpolys]
			else:
				self.tpolys = []
			self.poly_mats = [p.MaterialIndex() for p in polys]
			self.materials = geom.Materials()
		else:
			self.polys = []
class SubmeshData(object):
	def __init__(self, verts, polys):
		self.used  = list(dict([(v, True) for p in polys for v in p]))
		vmap = dict(zip(self.used, xrange(len(self.used))))
		self.polys = [[vmap[v] for v in p] for p in polys]
		self.verts = num.take(verts, self.used)
class Geometry(object):
	def __init__(self, verts, polys, poly_mats = None, normals = None, tverts = None, tpolys = None, options = {}):
		self.verts     = verts
		self.polys     = polys
		self.poly_mats = poly_mats
		self._normals  = normals
		self.tverts    = tverts
		self.tpolys    = tpolys
		self.options   = options

		self.log = options.get('logger', self.default_logger)
		if not options.get('skip_check', False):
			self.check_tpolys()

	def check_tpolys(self):
		if (self.tpolys is not None) and (self.tverts is not None):
			if len(self.polys) != len(self.tpolys):
				self.tverts = self.tpolys = None
				self.log("Corrupted UVs removed.")
			else:
				count = 0
				for poly, tpoly in zip(self.polys, self.tpolys):
					if len(tpoly) != len(poly):
						count += 1
						if len(tpoly) == 0:
							tpoly.append(0)
						while len(tpoly) < len(poly):
							tpoly.append(tpoly[-1])

				if count: self.log(count, "bad UV polygons")

	def default_logger(self, *args):
		# print " ".join(map(str, args))
		pass

	def extract_by_material(self, material_index):
		return self.extract([i for i, n in enumerate(self.poly_mats or [])
							 if n == material_index])

	def extract(self, poly_indices):
		geomesh = SubmeshData(self.verts, [self.polys[i] for i in poly_indices])
		poly_mats = [self.poly_mats[i] for i in poly_indices]
		if self._normals is not None:
			normals = num.take(self._normals, geomesh.used)
		else:
			normals = None

		if (self.tpolys is not None) and (self.tverts is not None):
			texmesh = SubmeshData(self.tverts,
								  [self.tpolys[i] for i in poly_indices])
			tverts = texmesh.verts
			tpolys = texmesh.polys
		else:
			tverts = tpolys = None
			
		return self.__class__(geomesh.verts, geomesh.polys, poly_mats,
							  normals, tverts, tpolys, self.options)

	def compute_normals(self):
		rotate = lambda v: num.take(v, [1,2,0], 1)
		
		p = num.take(self.verts, num.concatenate(self.polys))
		q = num.take(self.verts, num.concatenate([poly[1:] + poly[:1]
												  for poly in self.polys]))
		cross = rotate(p * rotate(q) - q * rotate(p))

		ends    = num.add.accumulate([len(p) for p in self.polys])
		starts  = num.concatenate([[0], ends])
		normals = num.array([num.sum(num.take(cross, range(a,o)))
							 for a, o in zip(starts, ends)])
		self.face_normals = fnormals = normalize(normals)

		normals = num.zeros([len(self.verts), 3], "double")
		for i, poly in enumerate(self.polys):
			n = fnormals[i]
			for v in poly: normals[v] += n
		self._normals = normalize(normals)

	def interpolate(self, center, adj):
		self.verts[center] = num.sum(num.take(self.verts, adj)) / len(adj)
		if self._normals is not None:
			n = num.sum(num.take(self._normals, adj))
			self._normals[center] = n / num.sqrt(num.maximum(num.dot(n, n),
															 1e-16))

	def subdivide(self):
		# -- grab some instance data
		verts   = self.verts
		normals = self._normals
		
		original_polys        = self.polys
		original_poly_mats    = self.poly_mats
		original_vertex_count = len(verts)
		
		# -- holds next available vertex index
		next_index = len(verts)

		# -- count edges and assign vertex indices for edge centers
		edge_centers  = {}
		edge_center_to_poly_centers = {}
		vertex_to_edge_centers = [[] for v in xrange(original_vertex_count)]

		for poly in original_polys:
			for u, v in zip(poly, poly[1:] + poly[:1]):
				if edge_centers.has_key((v, u)):
					ec = edge_centers[(u, v)] = edge_centers[(v, u)]
				else:
					ec = edge_centers[(u, v)] = next_index
					next_index += 1
					edge_center_to_poly_centers[ec] = []
					vertex_to_edge_centers[u].append(ec)
					vertex_to_edge_centers[v].append(ec)

		# -- make room for edge and polygon centers in vertex arrays
		new_size = next_index + len(original_polys)
		self.verts = verts = num.resize(verts, (new_size, 3))
		if self._normals is not None:
			self._normals = normals = num.resize(normals, (new_size, 3))

		# -- create the polygon centers and new polygons
		vertex_to_poly_centers = [[] for v in xrange(original_vertex_count)]

		self.polys     = polys     = []
		self.poly_mats = poly_mats = []

		for i, poly in enumerate(original_polys):
			if len(poly) < 1: continue

			# -- claim next free vertex index
			center = next_index
			next_index += 1

			# -- interpolate 3d and texture coordinates and normals
			self.interpolate(center, poly)

			# -- get the edge centers on the polygon's boundary
			ecs = [edge_centers[(u, v)]
				   for u, v in zip(poly, poly[1:] + poly[:1])]

			# -- subdivide the polygon
			mat = original_poly_mats[i]
			
			for u, v, w in zip(ecs[-1:] + ecs[:-1], poly, ecs):
				polys.append([u, v, w, center])
				poly_mats.append(mat)
				vertex_to_poly_centers[v].append(center)

			# -- link edge centers to new polygon center
			for ec in ecs:
				edge_center_to_poly_centers[ec].append(center)

		# -- subdivide the texture polygons
		if self.tverts is not None:
			tverts  = self.tverts
			n = len(tverts)
			tverts_needed = n + sum([1 + len(tpoly) for tpoly in self.tpolys])
			self.tverts = tverts = num.resize(tverts, (tverts_needed, 2))
		
			original_tpolys = self.tpolys
			self.tpolys = tpolys = []

			for i, poly in enumerate(original_tpolys):
				nv = len(poly)
				if nv < 1: continue
				ecs = range(n, n+nv)
				for u, v in zip(poly, poly[1:] + poly[:1]):
					tverts[n] = (tverts[u] + tverts[v]) / 2
					n += 1
				tverts[n] = num.sum(num.take(tverts, ecs)) / nv
				for u, v, w in zip(ecs[-1:] + ecs[:-1], poly, ecs):
					tpolys.append([u, v, w, n])
				n += 1

		# -- flag border vertices
		on_border = [False] * len(verts)
		for (u,v), ec in edge_centers.items():
			if len(edge_center_to_poly_centers[ec]) == 1:
				on_border[u] = on_border[v] = on_border[ec] = True

		# -- adjust edge center positions
		for (u, v), ec in edge_centers.items():
			if u > v and edge_centers.has_key((v, u)): continue
			if on_border[ec]:
				self.interpolate(ec, [u, v])
			else:
				self.interpolate(ec, [u, v] + edge_center_to_poly_centers[ec])

		# -- adjust positions of original vertices
		for v in xrange(original_vertex_count):
			ecs = vertex_to_edge_centers[v]
			pcs = vertex_to_poly_centers[v]
			k = len(ecs)

			if k < 2: continue
			
			if on_border[v]:
				self.interpolate(v, [u for u in ecs if on_border[u]])
			else:
				a = num.sum(num.take(verts, ecs)) * 4 / k
				b = num.sum(num.take(verts, pcs)) / k
				verts[v] = (verts[v] * (k - 3) + a - b) / k

	def add_backside(self, distance):
		if self._normals is not None:
			backverts = self.verts - self._normals * distance
			self._normals = num.concatenate([self._normals, -self._normals])
		else:
			backverts = self.verts
		n = len(self.verts)
		self.verts = num.concatenate([self.verts, backverts])

		self.polys += [[i + n for i in a[::-1]] for a in self.polys]
		
		self.poly_mats += self.poly_mats

		if self.tverts is not None:
			self.tverts = num.concatenate([self.tverts, self.tverts])
			if self.tpolys is not None:
				self.tpolys += [[i + n for i in a[::-1]] for a in self.tpolys]

	def convert_to_per_vertex_uvs(self):
		if (self.tpolys is not None) and (self.tverts is not None):
			self.fix_texture_seams()
			self.reorder_tex_verts()

	def corners_by_vertex(self):
		result = [[] for v in self.verts]
		for i, poly in enumerate(self.polys):
			for j, v in enumerate(poly):
				result[v].append((i, j))
		return result

	def fix_texture_seams(self):
		polygons  = self.polys
		tpolygons = self.tpolys
		corners   = self.corners_by_vertex()

		indices = range(len(self.verts))
		for v, corners_for_v in enumerate(corners):
			by_tvert = dict([(tpolygons[i][j], True) for i, j in corners_for_v])
			if len(by_tvert) < 2:
				continue

			by_texture_position = {}
			for tv in by_tvert:
				key = tuple([int(x * 5000) for x in self.tverts[tv]])
				by_texture_position.setdefault(key, []).append(tv)

			remap = {}
			for colliding in by_texture_position.values():
				for tv in colliding:
					remap[tv] = colliding[0]

			for i, j in corners_for_v: tpolygons[i][j] = remap[tpolygons[i][j]]

			by_tvert = {}
			for i, j in corners_for_v:
				by_tvert.setdefault(tpolygons[i][j], []).append((i, j))

			for tv, corners_for_tv in by_tvert.items()[1:]:
				for i, j in corners_for_tv:
					polygons[i][j] = len(indices)
				indices.append(v)

		self.verts = num.take(self.verts, indices)
		if self._normals is not None: self._normals = num.take(self._normals, indices)

	def reorder_tex_verts(self):
		tpolys = self.tpolys
		corner_to_tex = [None] * len(self.verts)
		corner_from_tex = {}
		
		for i, p in enumerate(self.polys):
			for j, v in enumerate(p):
				if corner_to_tex[v] is None:
					tv = tpolys[i][j]
					corner_to_tex[v] = tv
					corner_from_tex[tv] = v

		self.tverts = num.take(self.tverts, corner_to_tex)
		self.tpolys = [[corner_from_tex[v] for v in p] for p in tpolys]

	def is_empty(self):
		return (self.number_of_points == 0) or (self.number_of_polygons == 0)
	is_empty = property(is_empty)

	def number_of_normals(self):
		if self._normals is not None:
			return len(self._normals)
		else:
			return 0
	number_of_normals = property(number_of_normals)

	def number_of_points(self):
		if self.verts is not None:
			return len(self.verts)
		else:
			return 0
	number_of_points = property(number_of_points)

	def number_of_polygons(self):
		if self.polys is not None:
			return len(self.polys)
		else:
			return 0
	number_of_polygons = property(number_of_polygons)

	def number_of_texture_points(self):
		if self.tverts is not None:
			return len(self.tverts)
		else:
			return 0
	number_of_texture_points = property(number_of_texture_points)

	def triangles(self):
		for poly in (self.polys or []):
			for v in xrange(1, len(poly) - 1):
				yield poly[0], poly[v], poly[v + 1]
	triangles = property(triangles)

	def points(self):
		if self.verts is not None:
			arr = self.verts
		else:
			arr = []
		for v in (arr):
			yield tuple(v)
	points = property(points)

	def normals(self):
		if self._normals is not None:
			arr = self._normals
		else:
			arr = []
		for n in (arr):
			yield tuple(n)
	normals = property(normals)

	def texture_points(self):
		if self.tverts is not None:
			arr = self.tverts
		else:
			arr = []
		for v in (arr):
			yield tuple(v)
	texture_points = property(texture_points)
class HairGeometry(Geometry):
	def __init__(self, verts, polys, poly_mats = None, normals = None,
				 tverts = None, tpolys = None, options = {}):
		r_root = options.get('root_radius', 0.0001)
		r_tip  = options.get('tip_radius', 0.00004)
		hair_verts = num.zeros([0,3], "double")
		hair_polys = []
		hair_poly_mats = []
		for p, m in map(None, polys, poly_mats):
			new_verts, new_polys = make_fiber(num.take(verts, p), r_root, r_tip)
			n = len(hair_verts)
			hair_polys.extend([[v + n for v in p] for p in new_polys])
			hair_poly_mats.extend([m] * len(new_polys))
			hair_verts = num.concatenate([hair_verts, new_verts])

		Geometry.__init__(self, hair_verts, hair_polys, hair_poly_mats)

# The exporter
class Exporter:
	def __init__(self, gui):
		self.count = 0
		self.exporting_lights = False
		self.gui = gui
		self.data = gui.GetData()
		self.done = 0
		self.library = self.data.LibraryManager()
		self.materials = self.data.Materials()
		self.matlist = {}					# Dictionary of Owners / exported materials
		self.portals = self.data.Portals()	# Dictionary of all Portals.
		self.skydome = None
		self.textures = {}					# All textures that have been output
		self.texture_maps = [{}, {}]		# All texture maps that have been output. [0]=colur, [1]=data
		self.verbosity = 0					# Higher values generate more verbose reports. Range 0-2
		self.volumes = {}					# All volumes that have been output
		
		# Output files
		self.LXSfile = None
		self.LXOfile = None
		self.LXMfile = None
		self.LXLfile = None
	
	def AddTextureMap(self, file, map, data):
		attr = LuxTexture()
		name = self.UniqueRef()
		attr.SetType(name)
		attr.SetTextureType('"%s" "imagemap"' % ("color", "float")[data])
		attr.AddNode("string", "filename", map[0])
		attr.AddNode("float", "gamma", "%s" % (2.2, 1.0)[data])
		attr.AddNode("float", "uscale", map[1])
		attr.AddNode("float", "vscale", map[2])
		attr.Write(file)
		self.texture_maps[data][map[0]] = name
		
		return name
	
	def CheckForPortal(self, key):
		m = self.materials[key]
		return m.IsPortal(), self.exporting_lights
		
	def DoDisplacement(self, key, attr=None):
	
		if not attr: attr = LuxAttribute("Mesh")

		m = self.materials[key]					# Material information as held by the Data object
		lm = self.matlist.get(key, None)		# Information from library material
		
		if lm:
		
			disp = False
			
			# Prefer user Material settings, if any
			if m.displacement and m.HasDisplacementMap():
				
				disp = True
				map = self.WriteTextureMap(key, m.simplemat.DisplacementMap, True, datatype="displacement")
				dmscale = 1.0 / (m.simplemat.DisplacementValue * 2.54)
				offset = m.disp_base
			
			# Otherwise use library material settings
			elif lm.displacement is not None:
				
				disp = True
				map = lm.displacement[0]
				dmscale = lm.displacement[1]
				offset = 0
				
			if disp == True:	
				#Add displacement parameters to geometry output
				attr.AddNode("string", "displacementmap", map)
				attr.AddNode("float", "dmscale", dmscale)
				attr.AddNode("float", "dmoffset", offset)
				
				# Set subdivision if none has been set yet
				if attr.nodes.get("subdivscheme", None) == None:
					self.GetSubDText(key, attr, override=True)
					
				if m.use_micro:
					attr.AddNode("string", "subdivscheme", "microdisplacement")
					attr.AddNode("integer", "nsubdivlevels", m.disp_levels * 5)
				else:
					attr.AddNode("integer", "nsubdivlevels", m.disp_levels)
				
		
		return attr
	
	def Export(self):
		
		global P2L_VERSION
		
		# Ensure there is at least one output file
		if self.LXSfile == None:
			tkMessageBox.showwarning("Oops!", "Unable to open output file")
			return
		
		self.done = 0
		target = len(self.data.ListOfMaterials()) + len(self.data.ListOfActors())
		self.gui.Progress(self.done, max=target, text="Beginning export...", colour="red")
		
		# Write Lux file(s)		
		print >> self.LXSfile, "# Exported using Pose2Lux Pro v%s\n" % P2L_VERSION
		
		# Write GLOBAL settings
		print >> self.LXSfile, "# GLOBAL INFORMATION\n"
		
		# Check for consistency
		r = self.gui.GetRenderer()
		s = self.gui.GetSurfaceIntegrator()
		if (r == 'sppm' or s == 'sppm') and not r == s :
			tkMessageBox.showerror("Oops!", "Renderer incompatible with Surface Integrator\nAborting export")
			self.gui.Progress(self.done, max=target, text="Export aborted", colour="red")
			return
			
		self.WriteCamera()
		self.WriteRenderer()
		self.WriteFilm()
		self.WritePixelFilter()
		self.WriteSampler()
		self.WriteAccelerator()
		self.WriteSurfaceIntegrator()
		self.WriteVolumeIntegrator()
		
		# Write SCENE SPECIFIC settings	
		print >> self.LXSfile, "\n# SCENE SPECIFIC INFORMATION\n"
		print >> self.LXSfile, "WorldBegin"
		
		self.WriteMaterials()
		self.WriteGeometry()
		self.WriteLights()

		print >> self.LXSfile, '\nExterior "WorldVolume"'
		print >> self.LXSfile, '\nWorldEnd'
		
		self.gui.Progress(self.done, text="Export completed successfully", colour="green")
	
	def ExteriorVolume(self, key):

		material = self.matlist.get(key, None)
		if material == None: return 'Exterior "WorldVolume"'
		
		volume = material.ExteriorVolume()
		
		if volume:
			v = volume.replace("USE_KEY",  key)
			text = 'Exterior "%s"' % v
		else:
			text = 'Exterior "WorldVolume"'
		
		return text
	
	def GetDataPath(self):
		return self.gui.DefaultFolders("tabbeddata")
	
	def GetEmitter(self, key):
		m = self.materials[key]
		
		# Colour
		ref = "%s_emitter" % key
		texAttr = LuxTexture()
		texAttr.SetType(ref)
		
		# Emitter uses texture map
		if m.use_texture and m.HasDiffuseMap():	
			if m.IsTransparent() and m.HasTransparencyMap():
				texAttr.SetTextureType('"color" "scale"')
				texAttr.AddNode("texture", "tex1", self.WriteTextureMap(key, m.simplemat.TransparencyMap, False))
				texAttr.AddNode("texture", "tex2", self.WriteTextureMap(key, m.simplemat.DiffuseMap, False))
			else:
				ref = self.WriteTextureMap(key, m.simplemat.DiffuseMap, False)
				texAttr = None	
					
		# Simple colour only
		else:	
			if m.IsTransparent() and m.HasTransparencyMap():
				texAttr.SetTextureType('"color" "scale"')
				texAttr.AddNode("texture", "tex1", self.WriteTextureMap(key, m.simplemat.TransparencyMap, False))
				texAttr.AddNode("color", "tex2", str3(m.colour))
			else:
				texAttr.SetTextureType('"color" "constant"')
				texAttr.AddNode("color", "value", str3(m.colour))
		
		# Light source
		emitAttr = LuxAttribute("AreaLightSource")
		emitAttr.SetType("area")
		emitAttr.AddNode("float", "power", m.wattage)
		emitAttr.AddNode("float", "gain", m.gain)
		emitAttr.AddNode("float", "efficacy", m.efficacy)
		emitAttr.AddNode("texture", "L", ref)
		
		# Check for IES file
		if YN(m.ies) == "Yes":
			attr.AddNode("string", "iesname", m.ies)
			attr.AddNode("bool", "flipz", "true")
		
		return texAttr, emitAttr, m.group, self.GetOutputFile("materials")
		
	def GetNamedMaterial(self, key):
		material = self.matlist.get(key, None)
		if material == None:
			text = '# No Material definition for %s' % key
		else:
			text = 'NamedMaterial "%s"' % material.Type()
			
		return text
	
	def GetOutputFile(self, type):
		if type.lower() == "lights":
			if self.LXLfile:
				return self.LXLfile
			else:
				return self.LXSfile
		elif type.lower() == "geometry":
			if self.LXOfile:
				return self.LXOfile
			else:
				return self.LXSfile
		elif type.lower() == "materials":
			if self.LXMfile:
				return self.LXMfile
			else:
				return self.LXSfile
		else:
			return self.LXSfile
	
	def GetSubDText(self, key, attr=None, override=False):
	
		m = self.materials[key]		# Material information as held by the Data object
		lm = self.matlist[key]		# Information from library material

		if not attr: attr = LuxAttribute("Mesh")
		
		if m.subdivide or (override == True):
			attr.AddNode("integer", "nsubdivlevels", m.lux_levels)
			attr.AddNode("bool", "dmnormalsmooth", ("false", "true")[m.smooth_normals])
			attr.AddNode("bool", "dmsharpboundary", ("false", "true")[m.sharp_boundary])
			attr.AddNode("bool", "dmnormalsplit", ("false", "true")[m.create_edges])
			attr.AddNode("string", "subdivscheme", "loop")
			attr.AddNode("string", "tritype", "auto")
			
		return attr
	
	def GetTexturePath(self):
		return self.gui.DefaultFolders("textures")
	
	def InteriorVolume(self, key):
	
		material = self.matlist.get(key, None)
		
		if material:
			volume = material.InteriorVolume()
			
			if volume:
				v = volume.replace("USE_KEY",  key)
				text = 'Interior "%s"' % v
			else:
				text = '# No interior volume'
		else:
			text = ""
		
		return text
	
	def IsEmitter(self, key):
		m = self.materials[key]
		return m.IsEmitter() and not m.Group() == "ignore"
	
	def IsSkydome(self, key):
		m = self.materials[key]
		if m.Luxname() == "Special/Skydome":
			self.skydome = key
			return True
		else:
			return False
	
	def MakeBackside(self, key):
		m = self.materials[key]
		return (None, m.distance)[m.double_sided]
			
	def SetOutputFiles(self, files):
		self.LXSfile, self.LXOfile, self.LXMfile, self.LXLfile = files
	
	def SetVerbosity(self, level):
		try:
			self.verbosity = min(int(level), 2)
		except:
			self.verbosity = 0
	
	def TextureNotUsedYet(self, key):
		return key not in self.textures
	
	def UniqueRef(self):
		ref = '%04d' % self.count
		self.count += 1
		return ref
	
	def UpdateProgressBar(self, text, increase=0):
		self.done += increase
		self.gui.Progress(self.done, text=text)
	
	def VolumeNotUsedYet(self, key):
		return key not in self.volumes
		
	def WriteAccelerator(self):
		attr = LuxAttribute("Accelerator")
		self.gui.GetParmsAccelerator(attr)
		attr.Write(self.LXSfile)
	
	def WriteCamera(self):
		camera = self.data.camera
		print >> self.LXSfile, camera.LookAtString()
		if camera.IsValid():
			attr = camera.MakeLuxAttribute()
			self.gui.GetParmsCamera(attr, camera)
			attr.Write(self.LXSfile)
			
	def WriteFilm(self):
		attr = LuxAttribute("Film")
		self.gui.GetParmsFilm(attr)
		attr.Write(self.LXSfile)
	
	def WriteGeometry(self):	
		file = self.GetOutputFile("geometry")
		
		# Default options
		options = {}
		options['subdivisionlevel'] = 0
		options['compute_normals'] = self.gui.ComputeNormals()
		
		print >>file, "#### GEOMETRY ####\n"
	
		for actor in self.data.ListOfActors():
		
			self.UpdateProgressBar("Exporting actor %s" % actor.Name())
			
			# Get the actor's geometry using ODF's code
			geometry = get(actor.actor)
			
			# If there is some geometry ...
			if not (geometry is None or geometry.is_empty):
				
				ignored_materials = len(actor.materials) - len(geometry.materials)
				self.UpdateProgressBar("Analysis completed for %s" % actor.Name(), increase=1 + ignored_materials)
				
				# Provide a list of material keys
				materials = [MakeKey(actor.Name(), mat.ExtName()) for mat in geometry.materials]
				
				# Update the default options with actor-specific ones
				parms = options.copy()
				parms.update(actor.Options())
				# Compute normals and apply Catmull-Clark Subdivision if appropriate
				preprocess(geometry, self, parms)
				
				# Export the geometry
				write(file, geometry, materials, self)
				
			else:
				ignored_materials = len(actor.materials)
				self.UpdateProgressBar("Analysis completed for %s" % actor.Name(), increase=1 + ignored_materials)				
		
		if self.LXOfile: print >> self.LXSfile, 'Include "%s"' % (self.gui.GetOutputBasename() + '.lxo')
	
	def WriteLibraryTexture(self, texKey, p2lmat, file):
		
		tAttr = LuxTexture()
		tAttr.SetType(texKey, p2lmat.Key())
		if self.TextureNotUsedYet(tAttr.Type()):
			self.library.MakeTexture(texKey, tAttr)
			maxD, maxS = self.gui.GetMaxColourValues()
			tAttr.ProcessInputs(p2lmat, maxD, maxS, self)
			tAttr.Write(file)
			self.textures[tAttr.Type()] = None

	def WriteLibraryVolume(self, volKey, p2lmat, file, maxD, maxS):
		vAttr = LuxVolume()
		vAttr.SetType(volKey, p2lmat.Key())
		if self.VolumeNotUsedYet(vAttr.Type()):
			usedTextures = self.library.MakeVolume(volKey, vAttr)
			for t in usedTextures:
				self.WriteLibraryTexture(t, p2lmat, file)
			vAttr.ProcessInputs(p2lmat, maxD, maxS, self)
			vAttr.Write(file)
			self.volumes[vAttr.Type()] = None
		
	def WriteLights(self):
	
		self.exporting_lights = True
		
		file = self.GetOutputFile("lights")
		
		print >>file, "#### LIGHTS ####\n"
		
		# Sun
		sun, attr = self.gui.GetSun()
		if attr:
			attr.AddNode("integer", "nsamples", self.gui.GetSamples())
			print >>file, '\nAttributeBegin'
			attr2 =  LuxAttribute("LightGroup")
			attr2.SetType("Sun")
			attr2.Write(file)
			attr.Write(file)
			self.WritePortals(file, "sun")
			print >>file, '\nExterior "WorldVolume"'
			print >>file, '\nAttributeEnd'
			
		# Sky
		attr = self.gui.GetSky()
		if attr:
			attr.AddNode("integer", "nsamples", self.gui.GetSamples())
			print >>file, '\nAttributeBegin'
			attr2 =  LuxAttribute("LightGroup")
			attr2.SetType("Sky")
			attr2.Write(file)
			attr.Write(file)
			self.WritePortals(file, "sky")
			print >>file, '\nExterior "WorldVolume"'
			print >>file, '\nAttributeEnd'
		
		# Other lights
		for l in self.data.ListOfLights():
		
			# TO DO - check if light is Sun
			
			if not l.Group() == "ignore" and l.Name() != self.gui.GetSun(justname=True):
				
				list = l.MakeLuxAttribute(self.gui.GCforIBL(), antigc=self.gui.AntiGCLights())
				attr2 = LuxAttribute("LightGroup")
				attr2.SetType(l.Group())
				
				print >> file, '\n# Light name is %s\nAttributeBegin' % l.name
				attr2.Write(file)
				
				# Bug in Luxrender requires HDRI maps to to flipped horizontally
				if l.GetTypeByName() == "IBL":
					if self.gui.FlipHDRI() and YN(l.image) == "Yes": print >> file, 'Scale -1 1 1'
				
				# Write out transforms for IBL and Point light types
				if l.GetTypeByName() == "Point":
					print >>file, 'TransformBegin'
					print >>file, 'Translate %s %s %s' % l.PointFromTo()[0]
				elif l.GetTypeByName() == "IBL":
					print >>file, 'TransformBegin'
					print >>file, 'Rotate 270 0 0 1'
					print >>file, 'Rotate %s 0 0 1' % l.yrot
					
				for attr in list: attr.Write(file)
				
				if l.GetTypeByName() in ("IBL", "Point"): print >>file, 'TransformEnd'
					
				if l.GetTypeByName() == "IBL": self.WritePortals(file, "infinite")	
				
				print >>file, '\nExterior "WorldVolume"'
				print >>file, '\nAttributeEnd'
				
		# Check for Skydome
		if self.skydome:
		
			mat = self.materials[self.skydome]
			act = mat.Owner().Actor()
			
			if not mat.HasDiffuseMap(): return
			
			print >> file, 'AttributeBegin'
			
			# Bug in Luxrender requires HDRI maps to to flipped horizontally
			if self.gui.FlipHDRI(): print >> file, 'Scale -1 1 1'
		
			# Transformation
			print >>file, 'TransformBegin'
			print >> file, 'Rotate 270 0 0 1'
			try:
				yrot = act.ParameterByCode(poser.kParmCodeYROT).Value()
				if self.gui.FlipHDRI(): yrot *= -1	
				yrot += 180
			except:
				yrot = 0
				tkMessageBox.showwarning("Pose2Lux", "Skydome does not have a Y rotation parameter, 0 degree rotation used")			
			print >> file, 'Rotate %s 0 0 1' % yrot
			
			# Lightsource
			attr = LuxAttribute("LightSource")
			attr.SetType("infinite")
			attr.AddNode("float", "gain", 25000)
			attr.AddNode("integer", "nsamples", self.gui.GetSamples())
			attr.AddNode("float", "gamma", self.gui.GCforIBL())
			attr.AddNode("string", "mapname", mat.simplemat.DiffuseMap[0])
			attr.Write(file)

			print >>file, '\nExterior "WorldVolume"'
			print >> file, 'AttributeEnd\n'			
			
			
		self.exporting_lights = False
					
		if self.LXLfile: print >> self.LXSfile, 'Include "%s"' % (self.gui.GetOutputBasename() + '.lxl')

	def WriteMaterial(self, p2lmat, file, generateName=False, mat=None, parent=None):
	
		maxD, maxS = self.gui.GetMaxColourValues()

		mAttr = LuxMaterial()
		if not parent: parent = mAttr
		
		# Get name for MakeNamedMaterial
		name = p2lmat.Key()
		if generateName == True: name += self.UniqueRef()
		mAttr.SetType(name)
		
		# Process library parameters & update LuxMaterial
		if self.gui.ClayMode():
			key = "p2l/Clay"
		else:
			key = mat or p2lmat.Luxname()
			if not key: key = "p2l/Clay"
			
		libdata = self.library.MakeMaterial(mAttr, p2lmat, key)
		mAttr.SetAdditionalData(libdata)

		# Stop processing if there has been an error
		if libdata == None: return None
		
		# Cache certain information in root node
		if parent:
			parent.SetDisplacement(libdata.Displacement())
		else:
			mAttr.SetDisplacement(libdata.Displacement())
		
		# Write out dependencies ...
		
		# Displacement maps
		if p2lmat.HasDisplacementMap():
			self.WriteTextureMap(p2lmat.Key(), p2lmat.simplemat.DisplacementMap, True, datatype="displacement")
		
		# Dependent Textures
		for t in libdata.UsedTextures(): self.WriteLibraryTexture(t, p2lmat, file)
		
		# Dependent Volumes
		vol = libdata.InteriorVolume()
		if vol:
		
			# Write out volume defintion
			self.WriteLibraryVolume(vol, p2lmat, file, maxD, maxS)
			
			# Make note of volume data in root material node
			if parent:
				parent.SetInteriorVolume(vol)
			else:
				mAttr.SetInteriorVolume(vol)
		
		# And the same for the exterior volume ...
		vol = libdata.ExteriorVolume()
		if vol:
			self.WriteLibraryVolume(vol, p2lmat, file, maxD, maxS)
			if parent:
				parent.SetExteriorVolume(vol)
			else:
				mAttr.SetExteriorVolume(vol)

		# Deal with the special case of 'mix' materials
		if libdata.Type() == 'mix':
			mix1 = self.WriteMaterial(p2lmat, file, generateName=True, mat=libdata.Mix(1), parent=parent)
			mix2 = self.WriteMaterial(p2lmat, file, generateName=True, mat=libdata.Mix(2), parent=parent)
			mAttr.AddNode("string", "type", "mix")
			mAttr.AddNode("string", "namedmaterial1", mix1.Type())
			mAttr.AddNode("string", "namedmaterial2", mix2.Type())
			
			try:
				x = float(libdata.MixAmount())
				mAttr.AddNode("float", "amount", libdata.MixAmount())
			except:
				mAttr.AddNode("texture", "amount", libdata.MixAmount())
		
		# Write out the material definition
		mAttr.ProcessInputs(p2lmat, maxD, maxS, self)
		mAttr.Write(file)
		
		# Process alpha mask if there is one
		if p2lmat.IsTransparent():
			
			oldname = mAttr.Type()
			transmap = p2lmat.PoserMaterials().TransparencyMap
			if transmap[0]:
				type = "texture"
				value = self.WriteTextureMap(p2lmat.Key(), transmap, True)
			else:
				type = "float"
				value = 1.0 - p2lmat.PoserMaterials().TransparencyValue
			
			mAttr = LuxMaterial()
			mAttr.SetType(oldname + "_alpha")
			mAttr.AddNode("string", "type", "mix")
			mAttr.AddNode("string", "namedmaterial1", "nullMaterial")
			mAttr.AddNode("string", "namedmaterial2", oldname)
			mAttr.AddNode(type, "amount", value)
			mAttr.Write(file)
		
		return mAttr
		
	def WriteMaterials(self):
		
		file = self.GetOutputFile("materials")
		materials = self.data.ListOfMaterials()
		
		if self.LXMfile: print >> self.LXSfile, 'Include "%s"' % (self.gui.GetOutputBasename() + '.lxm')
		
		print >>file, "#### MATERIALS ####\n"
		
		# Default Null material
		attr = LuxMaterial()
		attr.SetType("nullMaterial")
		attr.AddNode("string", "type", "null")
		attr.Write(file)
		
		# Default Volume
		attr = LuxTexture()
		attr.SetType("Pose2LuxExterior")
		attr.SetTextureType('"fresnel" "constant"')
		attr.AddNode("float", "value", 1.000293)
		attr.Write(file)
		
		attr = self.gui.GetWorldVolume()
		if attr == None:		
			attr = LuxVolume()
			attr.SetType("WorldVolume")
			attr.SetVolumeType('"clear"')
			attr.AddNode("texture", "fresnel", "Pose2LuxExterior")
		attr.Write(file)
		
		for m in materials:
			self.matlist[m.Key()] = self.WriteMaterial(m, file)
		
	def WritePixelFilter(self):
		attr = LuxAttribute("PixelFilter")
		self.gui.GetParmsPixelFilter(attr)
		attr.Write(self.LXSfile)

	def WritePortals(self, file, type):
		portals = self.portals.get(type, None)
		if portals:
			for p in portals:
				actor = self.materials[p].Owner()
				geometry = get(actor.actor)
				if not (geometry is None or geometry.is_empty):
					materials = [MakeKey(actor.Name(), mat.ExtName()) for mat in geometry.materials]
					parms = {'compute_normals':True}
					preprocess(geometry, self, parms)
					write(file, geometry, materials, self)
		
	def WriteRenderer(self):
		attr = LuxAttribute("Renderer")
		self.gui.GetParmsRenderer(attr)
		attr.Write(self.LXSfile)
		
	def WriteSampler(self):
		attr = LuxAttribute("Sampler")
		self.gui.GetParmsSampler(attr)
		attr.Write(self.LXSfile)

	def WriteSurfaceIntegrator(self):
		attr = LuxAttribute("SurfaceIntegrator")
		self.gui.GetParmsSurfaceIntegrator(attr)
		attr.Write(self.LXSfile)	
			
	def WriteTextureMap(self, key, texmap, data, colour=(1.0, 1.0, 1.0), value=None, datatype=None):
	
		file = self.GetOutputFile("materials")
		map = texmap[0]
		
		# Adjust colour by value if given
		if value: colour = ColourValue(colour, value)
			
		# We only want to write out a texture map if we haven't already done so
		if map not in self.texture_maps[data]:
			ref = self.AddTextureMap(file, texmap, data)
		else:
			ref = self.texture_maps[data][map]
		
		if not data:
				
			# Write out texture map scaled by 'colour'
			attr = LuxTexture()
			attr.SetType("%s_scaled_%s" % (key, self.UniqueRef()))
			attr.SetTextureType('"color" "scale"')
			attr.AddNode("texture", "tex1", ref)
			attr.AddNode("color", "tex2", str3(colour))
			attr.Write(file)
			
		else:
			if datatype == "bump":
				
				# Convert to inches
				attr2 = LuxTexture()
				attr2.SetType("%s_bump_%s" % (key, self.UniqueRef()))
				attr2.SetTextureType('"float" "mix"')
				attr2.AddNode("texture", "amount", ref)
				attr2.AddNode("float", "tex1", -0.0254)
				attr2.AddNode("float", "tex2", 0.0254)
				attr2.Write(file)
				
				# Scale by bump amount
				amount = self.data.Materials()[key].PoserMaterials().BumpValue
				attr = LuxTexture()
				attr.SetType("%s_bump_%s" % (key, self.UniqueRef()))
				attr.SetTextureType('"float" "scale"')
				attr.AddNode("texture", "tex1", attr2.Type())
				attr.AddNode("float", "tex2", amount)
				attr.Write(file)
				
			elif datatype == "displacement":

				sm = self.materials[key].PoserMaterials()
				
				# Scale to inches
				attr2 = LuxTexture()
				attr2.SetType("%s_dispmix_%s" % (key, self.UniqueRef()))
				attr2.SetTextureType('"float" "mix"')
				attr2.AddNode("texture", "amount", ref)
				attr2.AddNode("float", "tex1", 0.0)
				attr2.AddNode("float", "tex2", sm.DisplacementValue * 0.0254)
				attr2.Write(file)
				
				# Displacement node
				attr = LuxTexture()
				attr.SetType("%s_displace_%s" % (key, self.UniqueRef()))
				attr.SetTextureType('"float" "scale"')
				attr.AddNode("texture", "tex1", attr2.Type())
				attr.AddNode("float", "tex2", sm.DisplacementValue)
				attr.Write(file)
				
			elif datatype == "specular":
			
				# Convert to type = 'color'
				attr2 = LuxTexture()
				attr2.SetType("%s_color_%s" % (key, self.UniqueRef()))
				attr2.SetTextureType('"color" "mix"')
				attr2.AddNode("color", "tex1", "0.0 0.0 0.0")
				attr2.AddNode("color", "tex2", "1.0 1.0 1.0")
				attr2.AddNode("texture", "amount", ref)
				attr2.Write(file)
				
				# Write out texture map scaled by 'colour'
				attr = LuxTexture()
				attr.SetType("%s_specular_%s" % (key, self.UniqueRef()))
				attr.SetTextureType('"color" "scale"')
				attr.AddNode("texture", "tex1", attr2.Type())
				attr.AddNode("color", "tex2", str3(colour))
				attr.Write(file)
			
			else:
				return ref
		
		return attr.Type()
		
	def WriteVolumeIntegrator(self):
		attr = LuxAttribute("VolumeIntegrator")
		self.gui.GetParmsVolumeIntegrator(attr)
		attr.Write(self.LXSfile)
	
# Data classes
class Data:
	def __init__(self, scene, pb=None):
		
		try: scene.ResolvePendingTextures()
		except: pass
			
		self.scene = scene									# The Poser scene
		self.actors = self.GetActors()						# List of Actors
		self.materials = self.GetMaterials()				# Dictionary of Materials, key=Actor/Material names
		self.lightgroups = ["ignore", "default", "Emitters", "Fill", "Main", "Rim", "Sky", "Sun"]
		self.lights = self.GetLights()						# Dictionary of Lights
		self.ResX, self.ResY = self.GetResolution()
		self.camera = Camera(scene, self.ResX, self.ResY)
		self.library = LibraryManager()						# Reference to the library manager
	
	def AllLightSources(self):
		list = self.lights.copy()
		for m in self.ListOfMaterials():
			if m.IsEmitter():
				list[m.Key()] = m
		return list
	
	def AutoAssignMaterials(self, skin, notPreview=True):
		for m in self.materials.values():
			m.AutoAssignMaterial(skin, notPreview)
	
	def GetActors(self):
		actors = []
		
		# Get figures
		for f in self.scene.Figures():
			if f.Visible():
				new = Actor(f)
				
				# Get parameters from added dials, if they exist
				try: new.LUXGEOM_catmullclark = int(f.Actor("BODY").Parameter("LUXGEOM_catmullclark").Value()) or 0
				except: pass
				try: new.LUXGEOM_normals = f.Actor("BODY").Parameter("LUXGEOM_normals").Value() > 0
				except: pass				
				try: new.LUXGEOM_subdivide = int(f.Actor("BODY").Parameter("LUXGEOM_subdivide").Value()) or 0
				except: pass
				try: new.LUXGEOM_sharpbounds = f.Actor("BODY").Parameter("LUXGEOM_sharpbounds").Value() > 0
				except: pass
				try: new.LUXGEOM_normalsplit = f.Actor("BODY").Parameter("LUXGEOM_normalsplit").Value() > 0
				except: pass
				try: new.LUXGEOM_dispbase = f.Actor("BODY").Parameter("LUXGEOM_dispbase").Value() or 0
				except: pass
				
				actors.append(new)
			
		# Get props
		for a in self.scene.Actors():
			if a.Visible() and not a.InternalName().startswith('Grouping'):
				new = Actor(a)			
				if self.ValidActor(a) and not a.IsBodyPart():
				
					# Get parameters from added dials, if they exist
					try: new.LUXGEOM_catmullclark = int(a.Parameter("LUXGEOM_catmullclark").Value()) or 0
					except: pass
					try: new.LUXGEOM_normals = a.Parameter("LUXGEOM_normals").Value() > 0
					except: pass				
					try: new.LUXGEOM_subdivide = int(a.Parameter("LUXGEOM_subdivide").Value()) or 0
					except: pass
					try: new.LUXGEOM_sharpbounds = a.Parameter("LUXGEOM_sharpbounds").Value() > 0
					except: pass
					try: new.LUXGEOM_normalsplit = a.Parameter("LUXGEOM_normalsplit").Value() > 0
					except: pass
					try: new.LUXGEOM_dispbase = a.Parameter("LUXGEOM_dispbase").Value() or 0
					except: pass				
					
					actors.append(new)
			
		return actors
	
	def GetLights(self):
		lights = {}
		for light in self.scene.Lights():
			if light.LightOn() == 1:
				l = Light(light, self.scene)
				lights[l.Name()] = l
				if l.Group() not in self.lightgroups:
					self.lightgroups.append(l.Group())
		return lights
	
	def GetMaterials(self):
		materials = {}
		for a in self.actors:
			mats = a.PoserMaterials()
			if mats:
				for m in mats:
					newmat = Material(m, a)
					materials[MakeKey(a.Name(), newmat.Name())] = newmat
					
					# Update Material with data from parent Actor
					if a.LUXGEOM_catmullclark:
						newmat.cc_levels = a.LUXGEOM_catmullclark
					if a.LUXGEOM_subdivide:
						newmat.subdivide = True
						newmat.lux_levels = a.LUXGEOM_subdivide
					if a.LUXGEOM_sharpbounds:
						newmat.sharp_boundary = a.LUXGEOM_sharpbounds
					if a.LUXGEOM_normalsplit:
						newmat.create_edges = a.LUXGEOM_normalsplit
					if a.LUXGEOM_dispbase:
						newmat.disp_base = a.LUXGEOM_dispbase
						
					# Add Material to Actor's list of its materials
					a.AddMaterial(newmat)
		return materials

	def GetResolution(self):
	
		global POSER_VERSION
		
		if POSER_VERSION >= 8:
			import wx.aui
			def GetOpenGLCanvas():
				l = []
				for child in poser.WxAuiManager().GetManagedWindow().Children:
					GetOpenGLCanvasRec(child, l)
				return l
			def GetOpenGLCanvasRec(win, l):
				for child in win.Children:
					if child.Name.startswith('PoserGLCanvas') or child.Name.startswith('PoserPBCanvas'):
						l.append(child)
					else:
						GetOpenGLCanvasRec(child, l)
			def GetOutputRes():
				s = poser.Scene()
				if POSER_VERSION >= 8 and s.RenderDimAutoscale() == 2:
					for win in GetOpenGLCanvas():
						if win.IsShown(): return win.GetSize()    
				return s.OutputRes()
			
			return GetOutputRes()
		else:
			return poser.Scene().OutputRes()
			
	def GetResX(self):
		return self.ResX
	
	def GetResY(self):
		return self.ResY	
		
	def Inventory(self, *args):
		if "owners" in args:
			print '\nOWNERS:'
			for a in self.actors:
				print "\n", a.Name()
				for m in a.PoserMaterials():
					print "  %s " % m.Name()
		if "lights" in args:
			print '\nLIGHTS:'
			for l in self.lights.values():
				print 'Light name:', l.Name(), ', Group:', l.Group()
				
		print "\n\nInventory complete"

	def LibraryManager(self):
		return self.library
		
	def LightGroups(self):
		return self.lightgroups
	
	def ListOfLights(self):
		return self.lights.values()
		
	def Lights(self):
		return self.lights
	
	def ListOfActors(self):
		return self.actors
		
	def ListOfMaterials(self):
		return self.materials.values()
		
	def LightNamesByType(self, type=None):
		list = []
		for l in self.lights.values():
			if type == None or (l.GetTypeByName() == type):
				list.append(l.Name())
		list.sort()
		return list
	
	def LoadAll(self, root):
		
		# Materials
		element = root.find('materials')
		mats = element.findall('material')
		for m in mats:
			key = m.get("sortByOwner")
			material = self.materials.get(key, None)
			if material: material.Load(m)
			
		# Lights
		element = root.find('lights')
		mats = element.findall('light')
		for m in mats:
			key = m.get("name")
			light = self.lights.get(key, None)
			if light: light.Load(m)		
	
	def Material(self, key):
		return self.materials[key]
	
	def Materials(self):
		return self.materials
	
	def Portals(self):
		d = {"infinite" : [], "sun" : [], "sky" : []}
		for m in self.materials.values():
			if m.IsPortal():
				if m.doInfinite: d["infinite"].append(m.Key())
				if m.doSun: d["sun"].append(m.Key())
				if m.doSky: d["sky"].append(m.Key())
		return d
	
	def SaveAll(self, root):
		
		# Materials
		element = Element('materials')
		root.append(element)
		for m in self.ListOfMaterials(): m.Save(element)
		
		# Lights
		element = Element('lights')
		root.append(element)
		for l in self.ListOfLights(): l.Save(element)
	
	def ValidActor(self, actor):
		geom = actor.Geometry()
		name = actor.Name()
		validGeometry = (actor.IsProp() and geom and geom.NumVertices() > 0 and geom.NumPolygons() > 0)
		if name.startswith("GoalCenterOfMass") or name.startswith("CenterOfMass"):
			valid = False
		elif validGeometry and name not in ("UNIVERSE", "FocusDistanceControl"):
			valid = True
		else:
			valid = False
		return valid
class Actor:
	def __init__(self, actor):
		self.actor = actor
		self.materials = []
		self.options = {}
		
		# Parameters from added Poser dials
		self.LUXGEOM_catmullclark = None
		self.LUXGEOM_normals = None
		self.LUXGEOM_subdivide = None
		self.LUXGEOM_sharpbounds = None
		self.LUXGEOM_normalsplit = None
		self.LUXGEOM_dispbase = None
	
	def Actor(self):
		return self.actor
	
	def AddMaterial(self, mat):
		self.materials.append(mat)
	
	def Materials(self):
		return self.materials
	
	def Name(self):
		return self.actor.Name()

	def ObjFile(self):
		file = self.actor.GeomFileName()
		if file:
			return os.path.basename(self.actor.GeomFileName())
		else:
			return None
	
	def Options(self):
		options = self.options.copy()
		if self.LUXGEOM_normals != None: options['compute_normals'] = self.LUXGEOM_normals
		return options
	
	def PoserMaterials(self):
		return self.actor.Materials()

	def Subdivide(self, levels):
		self.options['subdivisionlevel'] = levels
		for m in self.materials:
			m.cc_levels = levels
class Camera:
	def __init__(self, scene, resX, resY):
		self.camera = scene.CurrentCamera()
		
		# Check for valid camera
		try:
			roll = self.camera.Parameter("roll").Value()
			pitch = self.camera.Parameter("pitch").Value()
			yaw = self.camera.Parameter("yaw").Value()
			self.validcamera = True
		except:
			tkMessageBox.showerror("Pose2Lux", "Camera type is not supported by Pose2Lux. Results may not be what you expect. If so, please use a different camera")
			roll = pitch = yaw = 0
			self.validcamera = False
		
		# Variables
		up = rollPitchYaw((0, 1, 0), roll, pitch, yaw)		# up vector
		la = rollPitchYaw((0, 0, -1), 0, pitch, yaw)		# look-at vector
		pos = self.camera.WorldDisplacement()
		
		self.position = xfm(pos)
		self.up = xfm(up)
		self.lookat = xfm(la)
		self.focalLength = self.camera.ParameterByCode(poser.kParmCodeFOCAL).Value()
		self.fov = self.GetFOV(resX, resY)
		
	def DOF(self, fstop):
		return self.FocalLength() / 1000.0 / (2.0 * float(fstop)) 
		
	def FocalLength(self):
		return self.camera.ParameterByCode(poser.kParmCodeFOCAL).Value()
		
	def FocusDistance(self):
		return self.camera.ParameterByCode(poser.kParmCodeFOCUSDISTANCE).Value()
	
	def FOV(self):
		return self.fov
		
	def GetFOV(self, resX, resY):
		if self.validcamera == True:
			return 360 / pi * atan(12.75 * min(1.0, float(resY) / float(resX)) / self.focalLength)
		else:
			return None
	
	def IsValid(self):
		return self.validcamera
	
	def LookAtString(self):
		return 'LookAt %s %s %s' % (str3(self.position), str3(add3(self.position, self.lookat)), str3(self.up))
	
	def MakeLuxAttribute(self):
		attr = LuxAttribute("Camera")
		attr.SetType("perspective")
		attr.AddNode("float", "fov", self.fov)
		attr.AddNode("float",  "focaldistance", self.FocusDistance())
		return attr
class Light:
	def __init__(self, light, scene):
		
		self.light = light
		
		self.colour = self.GetLightColour()
		self.gain = light.ParameterByCode(poser.kParmCodeKDINTENSITY).Value()
		self.image, self.imagetype = self.GetImage()
		self.name = light.Name()
		self.scale = (light.ParameterByCode(poser.kParmCodeASCALE).Value(), 1.0)[light.LightType() == 0]
		self.type = light.LightType()
		self.yrot = int(light.ParameterByCode(poser.kParmCodeYROT).Value())
		
		# Emitter variables for Point lights
		if self.GetTypeByName() == "Point":
			self.is_emitter =  True
			self.wattage = 150
			self.efficacy = 17
			self.ies = None
			self.use_texture = False
		else:
			self.is_emitter =  False
			self.wattage = 0
			self.efficacy = 0
			self.ies = None
			self.use_texture = False			

		# UI variables
		self.group = self.GetGroup(light.Name())
		self.importance = 1.0
		
		if self.GetTypeByName() == "Spot":
			self.coneangle = light.ParameterByCode(poser.kParmCodeLITEFALLOFFEND).Value()
			self.conedeltaangle = light.ParameterByCode(poser.kParmCodeLITEFALLOFFSTART).Value()		
			if self.coneangle == 0 : self.coneangle = self.conedeltaangle
		
		# Infinite lights (type=0) need to be scaled to 100% or things go awry
		if self.type == 0 and not light.ParameterByCode(poser.kParmCodeASCALE).Value() == 1:
			light.ParameterByCode(poser.kParmCodeASCALE).SetValue(1.0)
			scene.DrawAll()
			scene.ProcessSomeEvents(10)
		
		self.saveload = {
			"name" : "string",
			"gain" : "float",
			"is_emitter" : "bool",
			"wattage" : "integer",
			"gain" : "float",
			"efficacy" : "integer",
			"ies" : "string",
			"use_texture" : "bool",
			"group" : "string",
			"importance" : "float"
			}
			
		if self.GetTypeByName() == "Point":
			self.saveload["colour"] = "tuple"

	def AdjustedGain(self):
		gain = self.gain * (1, 4096, 1, 1000)[self.type]
		if self.type == 3 and not self.image == None: gain *= 25
		return gain
		
	def GetGroup(self, name):
		groupname = name.split(":")
		if len(groupname) > 1:
			return groupname[0]
		else:
			if self.GetTypeByName() == "Point":
				return "Emitters"
			else:
				return "default"
			
	def GetImage(self):
		mat = self.light.Material("Preview")
		node = mat.ShaderTree().Node(0).Input(0).InNode()
		if node and (node.Type() != poser.kNodeTypeCodeIMAGEMAP) : return None, None
        
		try:
			map = node.Input(0).Value().replace("\\", "/")
			idx = int(self.light.Parameter("Shadow").Value())
			type = ("angular", "latlong", "spherical")[idx]
        
		except Exception, err:
			map = type = None
            
		return map, type
		
	def GetLightColour(self):
		colour = (
			self.light.ParameterByCode(poser.kParmCodeKDRED).Value(),
			self.light.ParameterByCode(poser.kParmCodeKDGREEN).Value(),
			self.light.ParameterByCode(poser.kParmCodeKDBLUE).Value())
		return colour
	
	def GetTypeByName(self):
		lightTypes = ["Infinite", "Spot", "Point", "IBL"]
		return lightTypes[self.type]
		
	def Group(self):
		return self.group
		
	def Importance(self):
		return self.importance
	
	def Load(self, data):
		for key, value in data.items():
			try:
				setattr(self, key, Unpack(value, self.saveload[key]))
			except:
				pass
		
	def MakeLuxAttribute(self, gc, antigc=None):
	
		list = []
		
		attr = LuxAttribute("LightSource")
		type = self.GetTypeByName()

		if antigc == None:
			colour = self.colour
		else:
			colour = [x**antigc for x in self.colour]
		
		# Infinite
		if type == "Infinite":
			point = self.PointFromTo()
			attr.SetType("distant")
			attr.AddNode("float", "gain", self.AdjustedGain())
			attr.AddNode("color", "L", "%s %s %s" % tuple(colour))
			attr.AddNode("point", "from", str3(point[0]))
			attr.AddNode("point", "to", str3(point[1]))
			attr.AddNode("float", "importance", self.importance)
			list.append(attr)
			
		# Spot
		elif type == "Spot":
			point = self.PointFromTo()
			attr.SetType("spot")
			attr.AddNode("float", "gain", self.AdjustedGain())
			attr.AddNode("color", "L", "%s %s %s" % tuple(colour))
			attr.AddNode("point", "from", str3(point[0]))
			attr.AddNode("point", "to", str3(point[1]))
			attr.AddNode("float", "coneangle", self.coneangle)
			attr.AddNode("float", "conedeltaangle", self.conedeltaangle)
			attr.AddNode("float", "importance", self.importance)
			list.append(attr)
			
		# Point
		elif type == "Point":
			
			attr = LuxAttribute("AreaLightSource")
			attr.SetType("area")
			attr.AddNode("color", "L", str3(colour))
			attr.AddNode("float", "power", self.wattage)
			attr.AddNode("float", "gain", self.AdjustedGain())
			attr.AddNode("float", "efficacy", self.efficacy)
			if YN(self.ies) == "Yes":
				attr.AddNode("string", "iesname", self.ies)
				attr.AddNode("bool", "flipz", "true")
			list.append(attr)
			
			# Make geometry for point light
			attr2 = LuxAttribute("Shape")
			attr2.SetType("sphere")
			attr2.AddNode("float", "radius", self.scale * (1.35 / 103.2))
			list.append(attr2)
			
		# IBL
		elif type == "IBL":
			attr.SetType("infinite")
			attr.AddNode("float", "gain", self.AdjustedGain())
			if self.image:
				attr.AddNode("string", "mapname", self.image)
				attr.AddNode("float", "gamma", gc)
			else:
				attr.AddNode("color", "L", str3(self.colour))
			list.append(attr)
		
		return list
	
	def Name(self):
		return self.name
	
	def NameAlt(self):
		return self.name
	
	def PointFromTo(self):
		wm = self.light.WorldMatrix()
		pto = (0, 0, -1)
		pto = dotp(pto, mcol(wm, 0)), dotp(pto, mcol(wm, 1)), dotp(pto, mcol(wm, 2))
		if self.type == 0:
			point_from = xfm(mul3(pto, [-1, -1, -1]))
			point_to = [0,0,0]
		else:
			pto = add3(pto, wm[3])
			point_from = xfm(self.light.WorldDisplacement())
			point_to = xfm(pto)
		return (point_from, point_to)
	
	def Save(self, element):
		attribs = {}
		for key, value in self.saveload.items():
			attribs[key] = str(getattr(self, key))
		element.append(Element("light", attribs))
		
	def SetGroup(self, group):
		self.group = group
		
	def SetImportance(self, value):
		self.importance = value
	
	def SunVector(self):
		wm = self.light.WorldMatrix()
		pto = (0, 0, -1)
		pto = dotp(pto, mcol(wm, 0)), dotp(pto, mcol(wm, 1)), dotp(pto, mcol(wm, 2))
		return xfm(mul3(pto, [-1, -1, -1]))
class Material:
	def __init__(self, material, owner):
	
		# Principal variables
		self.luxname = None
		self.material = material
		self.owner = owner
		self.sortByName = MakeKey(self.Name(), self.owner.Name())
		self.sortByOwner = MakeKey(self.owner.Name(), self.Name())
		self.simplemat = SimpleMat(material)

		# Emitter parameters
		self.is_emitter = False
		self.wattage = 0
		self.gain = 0.0
		self.efficacy = 0
		self.ies = None
		self.use_texture = False
		self.group = "Emitters"
		self.colour = (0, 0, 0)
		self.importance = 1.0
		
		# Subdivision parameters
		self.subdivide = False
		self.lux_levels = 0
		self.smooth_normals = True
		self.sharp_boundary = False
		self.create_edges = False
		self.cc_levels = 0
		self.disp_levels = 1
		
		# Displacement parameters
		self.displacement = False
		self.use_micro = True
		self.disp_base = 0
		
		# Backface parameters
		self.double_sided = False
		self.distance = 0.0
		
		# Portal parameters
		self.portal = False
		self.doInfinite = False
		self.doSun = False
		self.doSky = False
		
		# Custom material parameters
		self.use_custom = False
		
		# Update parameters from SimpleMat
		if self.simplemat.AmbientValue > 0:
			self.SetEmitter(True)
			self.colour = self.simplemat.AmbientColour
			self.wattage = 150
			self.efficacy = 17
			self.gain = self.simplemat.AmbientValue
		if self.simplemat.DisplacementValue > 0: self.displacement = True
		
		self.saveload = {
			"sortByOwner" : "string",
			"luxname" : "string",
			"is_emitter" : "bool",
			"wattage" : "integer",
			"gain" : "float",
			"efficacy" : "integer",
			"ies" : "string",
			"use_texture" : "bool",
			"group" : "string",
			"colour" : "tuple",
			"subdivide" : "bool",
			"lux_levels" : "integer",
			"smooth_normals" : "bool",
			"sharp_boundary" : "bool",
			"create_edges" : "bool",
			"cc_levels" : "integer",
			"disp_levels" : "integer",
			"displacement" : "bool",
			"use_micro" : "bool",
			"disp_base" : "integer",
			"double_sided" : "bool",
			"distance" : "float",
			"portal" : "bool",
			"doInfinite" : "bool",
			"doSun" : "bool",
			"doSky" : "bool",
			"use_custom" : "bool"			
			}
		
	def ApplyPortalTo(self):
		return (self.doInfinite, self.doSun, self.doSky)
		
	def AssignMaterial(self, material):
		self.luxname = material
	
	def AutoAssignMaterial(self, skin, notPreview):
	
		if self.IsPreview() and notPreview == True: return
		if self.UseCustomMaterial(): return
	
		if self.luxname == None:
		
			# Special cases
			if self.IsSkin(): self.luxname = skin or "People/ODF skin"; return
			if self.IsEmitter(): self.luxname = "Basic/Matte"
			
			# Transparency, but no transparency map
			if self.IsTransparent() and not self.HasTransparencyMap():
				if self.simplemat.TransparencyValue == 1:
					self.luxname = "Basic/Architectural glass"; return
				else:
					# TO DO: Check for Specular and if true use Glossy translucent
					self.luxname = "Basic/Matte Translucent"; return
			
			# Reflective materials use Basic/Glossy
			if self.IsReflective(): self.luxname = "Basic/Glossy"; return

			# If has Specular, use one of the Glossy variants
			if self.HasSpecular():
				if self.simplemat.SpecularSize > .8: self.luxname = "Basic/Glossy"; return			
				if self.simplemat.SpecularSize > .3: self.luxname = "Basic/Not so glossy"; return			
				if self.simplemat.SpecularSize > .1: self.luxname = "Basic/Smooth glossy"; return			
				self.luxname = "Basic/Rough glossy"; return
				
			# Catch all ...
			self.luxname = "Basic/Matte"		
	
	def GetTypeByName(self):
		return "Material"

	def Group(self):
		return self.group
		
	def HasDiffuseMap(self):
		return self.simplemat.DiffuseMap[0] != None
	
	def HasDisplacementMap(self):
		return self.simplemat.DisplacementMap[0] != None
	
	def HasSpecular(self):
		return (True, False)[(self.simplemat.SpecularValue == 0 or self.simplemat.SpecularColour == (0,0,0))]
	
	def HasTransparencyMap(self):
		return self.simplemat.TransparencyMap[0] != None
	
	def Importance(self):
		return self.importance	
	
	def IsEmitter(self):
		return self.is_emitter
	
	def IsPortal(self):
		return self.portal
	
	def IsPreview(self):
		return self.Name() == "Preview" and not len(self.Owner().PoserMaterials()) == 1

	def IsReflective(self):
		return self.simplemat.ReflectionValue > 0
		
	def IsRefractive(self):
		return self.simplemat.RefractionValue > 0

	def IsSkin(self):
	
		obj = {
			'!!!Kali_And_Kelm.obj':'Nursoda',
			'!!Eepo_Kobee.obj':'Nursoda',
			'3DrmLO_LO.obj':'Little1',
			'Alyson2_HR.obz':'P8',
			'Angela.obj':'Angela',
			'Antonia.obj':'Antonia',
			'Ben.obz':'P6',
			'BenP7.obz':'P6',
			'blAiko3.obj':'Aiko3',
			'blDavid.obj':'DAZgen3',
			'blFREAK.obj':'DAZgen3',
			'blGirl.obj':'TheGirl1',
			'blHiro.obj':'DAZgen3',
			'blMilKid_k4b.obj':'DAZgen4',
			'blMilMan_m3.obj':'DAZgen3',
			'blMilMan_m3GEN.obj':'M3Genital',
			'blMilMan_m3Lo.obj':'DAZgen3',
			'blMilMan_M3_M2.obj':'DAZgen3',
			'blMilMan_m4b.obj':'DAZgen4',
			'blMilMan_m4GEN.obj':'M4Genital',
			'blMilWom.obj':'DAZgen2',
			'blMilWomlo.obj':'DAZgen2',
			'blMilWom_v3.obj':'DAZgen3',
			'blMilWom_v3Lo.obj':'DAZgen3',
			'blMilWom_v4b.obj':'DAZgen4',
			'blNudeMan_Gen.obz':'P4',
			'blYTBoy.obj':'DAZgen3',
			'blYTGirl.obj':'DAZgen3',
			'blStephanie.obj':'DAZgen3',
			'blStephaniePetite.obj':'DAZgen3',
			'blVickiP4.obj':'DAZgen2',
			'Hein.obj':'Hein',
			'JamesHiRes.obz':'P6',
			'JamesLoRes.obz':'P6',
			'JessiHires.obz':'P6',
			'Jessi.obz':'P6',
			'Kate.obz':'P6',
			'KateP7.obz':'P6',
			'Kena.obj':'Nursoda',
			'Malik.obj':'Nursoda',
			'Mavka.obj':'Mavka',
			'MayaDoll01Basic.obj':'MayaDoll01',
			'Miki.obj':'Miki2',
			'Miki2.obj':'Miki2',
			'Miki3.obj':'Miki3',
			'P4NudeWom.obz':'P4',
			'PHFemale.obz':'PHFemale',
			'PHMale.obz':'PHMale',
			'P8Alyson_HR.obz':'P8',
			'P8Alyson_LR.obz':'P8',
			'P8Ryan_HR.obz':'P8',
			'P8Ryan_LR.obz':'P8',
			'PPro_LR_Fem.obz':'P8',
			'PPro_MR_Fem.obz':'P8',
			'Ryan2_HR.obz':'P8',
			'SimonG2.obz':'P6',
			'SydneyG2.obz':'P6'
			}
		skin = {
			'Aiko3':("SkinTorso","Nipples","SkinNeck","SkinHip","SkinFeet","SkinKnee","SkinShin","SkinThigh","SkinForearm","SkinArm","SkinHand","SkinHead","SkinScalp","Lips","Nostrils"),
			'Antonia':("brows", "lips", "skin_ARMS", "skin_BODY", "skin_HEAD", "skin_LEGS", "toeCap" ),
			'DAZgen2':("SkinHead","Eyebrows","Lips","Nostrils","UpperEyebrows","SkinBody","Nipples"),
			'DAZgen3':("SkinTorso","Nipples","SkinNeck","SkinHip","SkinFeet","SkinLeg","SkinForearm","SkinArm","SkinHand","SkinHead","SkinScalp","Lips","Nostrils"),
			'DAZgen4':("2_SkinHead","2_SkinNeck","2_SkinTorso","2_Nipple","2_SkinHip","3_SkinArm","3_SkinFoot","3_SkinForearm","3_SkinHand","3_SkinLeg","1_SkinFace","1_Lip","1_Nostril"),
			'Little1':("lowerTorso", "upperTorso", "neck", "nipples", "head", "upperFace", "lowerFace", "shoulders", "arms", "hands", "legs", "feets"),
			'M3Genital':("SkinTorso","ForeSkin"),
			'M4Genital':("Gen_Skin","Glans"),
			'Miki2':("Head","Body"),
			'Miki3':("Head","Lips","Arms","Legs","Body"),
			'Angela':("Body", "Gen", "Legs", "Arms", "Lips", "Face" ),
			'Hein':("Body", "HandNfeet", "Head", "Mouth" ),
			'Mavka':("bodyMat", "headMat", "lipMat" ),
			'MayaDoll01':("SkinBody", "Gen", "OuterNipples", "SkinHead", "Lips", "Naris" ),
			'Nursoda':("Body", "Head" ),
			'P4':("skin", "nipple", "lips", "eyeBrow", "nipples"),
			'P6':("Body", "Genitals", "Head"),
			'P8':("Body", "Head", "Nostrils", "Lips"),
			'PHFemale':( "Skin", "Nipples", "Genitalia", "Lips"),
			'PHMale':( "Skin", "Lips", "Nipples"),
			'TheGirl1':("SkinScalp", "SkinHead", "Nostrils", "Lips", "SkinTorso", "Nipples", "SkinNeck", "SkinFeet", "SkinHip", "SkinLeg", "SkinForearm", "SkinArm", "SkinHand" ),
			}
		
		result = False
		geom = self.Owner().ObjFile()
		if geom:
			figObj = obj.get(geom, None)
			if figObj:
				skinList = skin.get(figObj, None)
				if skinList:
					result = self.Name() in skinList
		
		return result
		
	def IsTransparent(self):
		return self.simplemat.TransparencyValue > 0
	
	def Key(self):
		return self.sortByOwner
	
	def Load(self, data):
		for key, value in data.items():
			v = Unpack(value, self.saveload[key])
			setattr(self, key, v)
			if key == "cc_levels" and v != 0:
				self.Owner().Subdivide(v)
		
	def Luxname(self):
		return (self.luxname, MakeKey("CUSTOM", self.Key()))[self.use_custom]
		
	def Name(self):
		return self.material.ExtName()

	def NameAlt(self):
		return self.sortByOwner
		
	def Options(self):
		return self.is_emitter, self.subdivide, self.displacement, self.double_sided, self.portal, self.use_custom
		
	def Owner(self):
		return(self.owner)

	def PoserMaterials(self):
		return self.simplemat
	
	def Save(self, element):
		attribs = {}
		for key, value in self.saveload.items():
			attribs[key] = str(getattr(self, key))
		element.append(Element("material", attribs))
	
	def SetEmitter(self, state, parms=None):
		self.is_emitter = state
		if parms:
			self.wattage = parms[0]
			self.gain = parms[1]
			self.efficacy = parms[2]
			self.ies = parms[3]
			self.use_texture = parms[4]
			self.group = parms[5]
	
	def SetGroup(self, group):
		self.group = group
		
	def SetImportance(self, value):
		self.importance = value
		
	def ToBeExported(self):
		return (True, False)[self.IsPreview() and self.Luxname() == None]
		
	def UseCustomMaterial(self):
		return self.use_custom
class SimpleMat:
	def __init__(self, material):
		
		self.mat = material
		self.ps = material.ShaderTree().Node(0)
		
		self.DiffuseColour = mul3by1(self.GetValue("Diffuse_Color"), self.GetValue("Diffuse_Value"))
		self.DiffuseMap = self.GetMap("Diffuse_Color")
		self.SpecularColour = self.GetValue("Highlight_Color")
		self.SpecularValue = self.GetValue("Highlight_Value")
		self.SpecularMap = self.GetMap("Highlight_Value")
		self.SpecularSize = self.GetValue("Roughness")
		self.AmbientColour = self.GetValue("Ambient_Color")
		self.AmbientValue = self.GetValue("Ambient_Value")
		self.TransparencyValue = self.GetValue("Transparency_Max")
		self.TransparencyMap = self.GetMap("Transparency_Max")
		self.ReflectionColour = self.GetValue("Reflection_Color")
		self.ReflectionValue = self.GetValueIfInput("Reflection_Value", "Reflection_Color")
		self.RefractionValue = self.GetValueIfInput("Refraction_Value", "Refraction_Color")
		self.BumpMap = self.GetMap("Bump")
		self.BumpValue = self.GetValue("Bump", True, self.BumpMap)
		self.DisplacementMap = self.GetMap("Displacement")
		self.DisplacementValue = self.GetValue("Displacement", True, self.DisplacementMap)
		
		# Check for nodes plugged in to AltDiffuse and AltSpecular
		self.CheckAlternateNodes()
		
	def CheckAlternateNodes(self):
	
		# Alternate Diffuse
		start = self.ps.InputByInternalName("AlternateDiffuse")
		if start.InNode():				# Then this material *appears* to override the diffuse settings 
			if self.DiffuseMap[0] == None:
				self.DiffuseMap = self.GetMap("AlternateDiffuse")
			if  bw(self.DiffuseColour) == 0:
				self.DiffuseColour = self.GetValue("AlternateDiffuse")
					
		# Alternate Specular
		start = self.ps.InputByInternalName("AlternateSpecular")
		if start.InNode():				# Then this material *appears* to override the specular settings 
			if self.SpecularMap[0] == None:
				self.SpecularMap = self.GetMap("AlternateSpecular")
			if  bw(mul3by1(self.SpecularColour, self.SpecularValue)) == 0:
				self.SpecularColour = self.GetValue("AlternateSpecular")
				
	def FindAKindOfNode(self, node, type):
		if not node:
			return None
		if node.Type() == type:
			return node
		for inp in node.Inputs():
			v = inp.InNode()
			found = self.FindAKindOfNode(v, type)
			if found:
				return found
		return None
		
	def GetMap(self, input):
		start = self.ps.InputByInternalName(input)
		node = start.InNode()
		map = uscale = vscale = None
		if node:
			mapnode = self.FindAKindOfNode(node, poser.kNodeTypeCodeIMAGEMAP)
			if mapnode:
				map = (mapnode.Input(0).Value(), None)[mapnode.Input(0).Value() == "" or mapnode.Input(0).Value() == "NO_MAP"]
				if map:
					map = map.replace("\\", "/", -1)
					uscale = 1.0 / (mapnode.Input(2).Value(), 1)[mapnode.Input(2).Value() == 0]
					vscale = 1.0 / (mapnode.Input(3).Value(), 1)[mapnode.Input(3).Value() == 0] * -1
		return (map, uscale, vscale)
	
	def GetValue(self, input, usemap=False, map=None):
		if usemap == False: return self.ps.InputByInternalName(input).Value()
		
		if map:
			if YN(map[0]) == "Yes":
				return self.ps.InputByInternalName(input).Value()
			else:
				return 0
		else:
			return 0
			
	def GetValueIfInput(self, input1, input2):
		check = self.ps.InputByInternalName(input2)
		if check.InNode():
			start = self.ps.InputByInternalName(input1)
			return start.Value()
		else:
			return 0

# UI classes
class GUIManager:
	def __init__(self, root, notebook):
	
		global NAME, POSER_VERSION, P2L_VERSION, HOME_FOLDER, FORCE_HOME_FOLDER
		self.root = root
		
		# Variables
		self.controls = {}			# All controls affected by the Load/Save Settings menu option
		self.data = None			# All data pertaining to the Poser scene
		self.settingsfile = None	# File location of currently loaded settings file
		
		# Initialise
		self.AnalyseScene()
		self.library = self.data.LibraryManager()
		self.AddPages(notebook)
		self.AddMenu(root)
		
		# Load default settings
		if POSER_VERSION > 6 and not FORCE_HOME_FOLDER:
			folder = os.path.dirname(poser.ScriptLocation())
		else:
			folder = HOME_FOLDER
		if folder:
			defaults = os.path.join(folder,'pose2lux.p2l')
			self.LoadSettings(path=defaults, root=None)		
		
		# Complete initialisation
		self.Finalise()
		
		# Start the main application
		root.geometry(("1200x750", "1125x750", "1150x700", "1200x760", "1200x760")[POSER_VERSION - 6])
		root.title(NAME + " v" + P2L_VERSION)
		root.mainloop()
	
	def AddLightGroup(self, newgroup):
	
		l = self.pageExposure.pageLights
		m = self.pageMaterials
		
		if not newgroup in l.groups:
			l.groups.append(newgroup)
			l.optGroups.addOption(newgroup, l.varGroup)
			m.Group.addOption(newgroup, l.varGroup)
			
			l.varGroup.set(newgroup) 
		
	def AddMenu(self, root):
	
		menubar = Menu(root)
		
		# File menu
		filemenu = Menu(menubar, tearoff=0)
		
		filemenu.add_command(label="Load settings", command=self.LoadSettings)
		filemenu.add_command(label="Save settings", command=self.SaveSettings)
		
		filemenu.add_separator()
		filemenu.add_command(label="Load Scene", command=self.LoadScene)	
		filemenu.add_command(label="Save Scene", command=self.SaveScene)
		
		filemenu.add_separator()
		filemenu.add_command(label="Load Template", command=self.LoadTemplate)	
		filemenu.add_command(label="Save Template", command=self.SaveTemplate)
		
		filemenu.add_separator()
		filemenu.add_command(label="Load material library", command=self.LoadMaterialLibrary)
			
		menubar.add_cascade(label="File", menu=filemenu)
		filemenu.add_separator()
		filemenu.add_command(label="Exit", command=self.Exit)

		root.config(menu=menubar)		
	
	def AddPages(self, notebook):
	
		global DEBUG
		
		# Create pages
		self.pageExport = pageExport(notebook(), self)
		self.pageExposure = pageCameraLights(notebook(), self)
		self.pageMaterials = pageMaterials(notebook(), self)
		self.pageOptions = pageOptions(notebook(), self)
		
		if DEBUG: self.pageDebug = pageDebug(notebook(), self)
		
		# Add pages to notebook
		notebook.AddScreen(self.pageMaterials.mainFrame, "Materials ")
		notebook.AddScreen(self.pageExposure.mainFrame, "Camera and Lights ")
		notebook.AddScreen(self.pageExport.mainFrame, "Export to Lux ")
		notebook.AddScreen(self.pageOptions.mainFrame, "Options ")
		if DEBUG: notebook.AddScreen(self.pageDebug.mainFrame, "** DEBUG ** ")
		
	def AnalyseScene(self):
		self.data = Data(poser.Scene())
	
	def AntiGCLights(self):
		if self.pageOptions.varAntiGC.get() == True:
			return self.pageExport.varFilmGamma.get()
		else:
			return None
	
	def ClayMode(self):
		return self.pageExport.varClay.get()
	
	def ComputeNormals(self):
		return self.pageExport.varComputeNormals.get()
	
	def DefaultFolders(self, folder=None):
		paths = self.pageOptions.GetAllPaths()
		folders = ["materials", "autoload", "scene", "template", "textures", "tabbeddata", "iesdata", "savescene"]
		
		if not folder:
			return paths
		else:
			return paths[folders.index(folder)]
	
	def Exit(self):
		self.root.destroy()	
	
	def Finalise(self):
		m = self.pageMaterials
		m.InitFinalise()
		self.pageExposure.pageLights.InitFinalise()
		self.library.AutoloadMaterials(self.DefaultFolders("autoload"), (m.optLibs, m.varLibs, m.UpdateLuxMats))
	
	def FlipHDRI(self):
		return self.pageExport.varFlipHDRI.get()
	
	def GCforIBL(self):
		return self.pageExport.IBLGC.get()
	
	def GetCurrentFigure(self):
		m = self.pageMaterials.matlist
		selected = m.curselection()
		if selected:
			row = m.get(selected[0])
			key = MakeKey(row[0], row[1])
			mat = self.data.materials[key]
			return mat.Owner()
		else:
			return None
	
	def GetData(self):
		return self.data
	
	def GetInitPath(self, path):
		if path:
		
			output = os.path.split(path)
			
			if output[1] == "": initFile = "Pose2Lux.p2l"
			else: initFile = output[1]
			
			if output[0] == "":
				initDir = os.path.dirname(poser.ScriptLocation())
			else:
				initDir = output[0]
		else:
			initDir = os.path.dirname(poser.ScriptLocation())
			initFile = "Pose2Lux.p2l"
		
		return initDir, initFile
	
	def GetLightGroups(self):
		l = self.pageExposure.pageLights
		return (l.groups, l.varGroup)
	
	def GetMaxColourValues(self):
		if self.pageExport.varScaleColours.get() == True:
			return self.pageExport.varScaleDiffuse.get(), self.pageExport.varScaleSpecular.get()
		else:
			return 1.0, 1.0
	
	def GetOutputBasename(self):
		return os.path.splitext(self.pageExport.ExportLocation())[0]
	
	def GetParmsAccelerator(self, attr):
		attr.SetType(self.pageExport.varAccelerator.get())
	
	def GetParmsCamera(self, attr, camera):
		cam = self.pageExposure.pageCamera
		if cam.UseDOF() == True:
			attr.AddNode("float", "lensradius", camera.DOF(cam.GetfStop()))
	
	def GetParmsFilm(self, attr):
		ex = self.pageExport
		cam = self.pageExposure.pageCamera
		
		attr.SetType("fleximage")
		attr.AddNode("integer", "xresolution", ex.GetResX())
		attr.AddNode("integer", "yresolution", ex.GetResY())
		attr.AddNode("bool", "premultiplyalpha", ("false", "true")[ex.varPreMultiplyAlpha.get()])
		attr.AddNode("bool", "write_png", ("false", "true")[ex.varWritePNG.get()])
		attr.AddNode("bool", "write_png_16bit", ("false", "true")[ex.varWrite16bitPNG.get()])
		attr.AddNode("bool", "write_png_gamutclamp", ("false", "true")[ex.varClamp.get()])
		attr.AddNode("string", "ldr_clamp_method", ex.varClampMethod.get())
		attr.AddNode("string", "write_png_channels", ex.varChannels.get())
		attr.AddNode("integer", "writeinterval", ex.WriteInterval.get())
		attr.AddNode("string", "filename", ex.varLuxName.get() or "luxout")
		attr.AddNode("float", "gamma", ex.varFilmGamma.get())
		attr.AddNode("bool", "write_resume_flm", ("false", "true")[ex.varFilmWriteFLM.get()])
		attr.AddNode("bool", "restart_resume_flm", ("false", "true")[ex.varFilmRestartFLM.get()])
		attr.AddNode("integer", "flmwriteinterval", ex.varFilmFLMWriteInterval.get())
		attr.AddNode("integer", "displayinterval", ex.varFilmDisplayInterval.get())
		attr.AddNode("integer", "outlierrejection_k", ex.varFilmOutlier.get())
		attr.AddNode("string", "tonemapkernel", "linear")
		attr.AddNode("float", "linear_sensitivity", cam.GetISO())
		attr.AddNode("float", "linear_exposure", cam.GetShutter())
		attr.AddNode("float", "linear_fstop", cam.GetfStop())
		attr.AddNode("float", "linear_gamma", cam.GetGamma())
	
	def GetParmsPixelFilter(self, attr):
		ex = self.pageExport
		pf = ex.varPixelFilter.get()
		
		attr.SetType(pf)
		if pf == "box":
			attr.AddNode("float", "xwidth", ex.varPFBoxXWidth.get())
			attr.AddNode("float", "ywidth", ex.varPFBoxYWidth.get())
		elif pf == "triangle":
			attr.AddNode("float", "xwidth", ex.varPFTriangleXWidth.get())
			attr.AddNode("float", "ywidth", ex.varPFTriangleYWidth.get())		
		elif pf == "gaussian":
			attr.AddNode("float", "xwidth", ex.varPFGaussianXWidth.get())
			attr.AddNode("float", "ywidth", ex.varPFGaussianYWidth.get())
			attr.AddNode("float", "alpha", ex.varPFGaussianAlpha.get())
		elif pf == "mitchell":
			attr.AddNode("float", "xwidth", ex.varPFMitchellXWidth.get())
			attr.AddNode("float", "ywidth", ex.varPFMitchellYWidth.get())
			attr.AddNode("float", "B", ex.varPFB.get())
			attr.AddNode("float", "C", ex.varPFC.get())
			attr.AddNode("bool", "supersample", ("false", "true")[ex.varPFSupersample.get()])
		elif pf == "sinc":
			attr.AddNode("float", "xwidth", ex.varPFSincXWidth.get())
			attr.AddNode("float", "ywidth", ex.varPFSincYWidth.get())
			attr.AddNode("float", "tau", ex.varPFSincTau.get())

	def GetParmsRenderer(self, attr):
		attr.SetType(self.GetRenderer())
	
	def GetParmsSampler(self, attr):
		ex = self.pageExport
		s = ex.varSampler.get()
		
		attr.SetType(s)
		if s == "random":
			attr.AddNode("integer", "pixelsamples", ex.varRandomPixelSamples.get())
			attr.AddNode("string", "pixelsampler", ex.varRandomPixelSampler.get())
		elif s == "lowdiscrepancy":
			attr.AddNode("integer", "pixelsamples", ex.varLowdiscrepancyPixelSamples.get())
			attr.AddNode("string", "pixelsampler", ex.varLowdiscrepancyPixelSampler.get())			
		elif s == "metropolis":
			attr.AddNode("integer", "maxconsecrejects", ex.varMetropolisMaxConsecRejects.get())
			attr.AddNode("float", "largemutationprob", ex.varMetropolisLargeMutationProbability.get())
		elif s == "erpt":
			attr.AddNode("integer", "initsamples", ex.varInitSamples.get())
			attr.AddNode("integer", "chainlength", ex.varChainLength.get())
	
	def GetParmsSurfaceIntegrator(self, attr):
		ex = self.pageExport
		
		if self.GetRenderer() == "sppm":
			s = "sppm"
		elif self.GetRenderer() == "hybrid":
			s = "path"
		else:
			s = ex.varSurfaceIntegrator.get()
		
		attr.SetType(s)
		
		if s == "bidirectional":
			attr.AddNode("integer", "eyedepth", ex.bidirEyedepth.get())
			attr.AddNode("integer", "lightdepth", ex.bidirLightdepth.get())
			attr.AddNode("string", "strategy", ex.varStrategy.get())
		elif s == "directlighting":
			attr.AddNode("string", "lightstrategy", ex.varDLLightStrategy.get())
			attr.AddNode("integer", "maxdepth", ex.directMaxDepth.get())
		elif s == "distributedpath":
			attr.AddNode("string", "strategy", ex.varDPStrategy.get())
		elif s == "path":
			attr.AddNode("integer", "maxdepth", ex.pathMaxDepth.get())
			attr.AddNode("string", "lightstrategy", ("one", ex.varPathLightStrategy.get())[self.GetRenderer() == "sampler"])
			attr.AddNode("string", "rrstrategy", ex.varPathRRStrategy.get())
			attr.AddNode("float", "rrcontinueprob", ex.varPathRRContinueProb.get())
		elif s == "exphotonmap":
			attr.AddNode("integer", "maxdepth", ex.exphotonMaxDepth.get())
			attr.AddNode("integer", "maxphotondepth", ex.exphotonMaxPhotonDepth.get())
			attr.AddNode("integer", "directphotons", ex.exphotonDirectphotons.get())
			attr.AddNode("integer", "causticphotons", ex.exphotonCausticphotons.get())
			attr.AddNode("integer", "indirectphotons", ex.exphotonIndirectphotons.get())
			attr.AddNode("integer", "radiancephotons", ex.exphotonRadiancephotons.get())
			attr.AddNode("integer", "nphotonsused", ex.exphotonNphotons.get())
			attr.AddNode("float", "maxphotondist", ex.exphotonMaxDist.get())
			attr.AddNode("bool", "finalgather", ("false", "true")[ex.varFinalGather.get()])
			attr.AddNode("string", "renderingmode", ex.varRenderMode.get())
		elif s == "sppm":
			attr.AddNode("integer", "photonperpass", ex.SPPMppp.get())
			attr.AddNode("integer", "maxeyedepth", ex.SPPMmed.get())
			attr.AddNode("integer", "maxphotondepth", ex.SPPMmpd.get())
			attr.AddNode("integer", "startk", ex.SPPMstartk.get())
			attr.AddNode("float", "startradius", ex.SPPMsr.get())		
			attr.AddNode("float", "alpha", ex.SPPMalpha.get())		
			attr.AddNode("bool", "includeenvironment", ("false", "true")[ex.SPPMincludeEnv.get()])
			attr.AddNode("float", "glossythreshold", ex.SPPMglossy.get())
			attr.AddNode("string", "lookupaccel", ex.varSPPMlookupaccel.get())
			attr.AddNode("string", "pixelsampler", ex.varSPPMpixelsampler.get())
			attr.AddNode("string", "photonsampler", ex.varSPPMphotonsampler.get())
	
	def GetParmsVolumeIntegrator(self, attr):
		attr.SetType(self.pageExport.varVolumeIntegrator.get())
		
	def GetRenderer(self):
		return self.pageExport.varRenderer.get()
	
	def GetSamples(self):
		return self.pageExposure.pageLights.NSamples.get()
	
	def GetSurfaceIntegrator(self):
		return self.pageExport.varSurfaceIntegrator.get()
	
	def GetSky(self):
		lights = self.pageExposure.pageLights
		
		if lights.Sky():
			attr = LuxAttribute("LightSource")
			attr.SetType("sky")
			attr.AddNode("float", "gain", lights.varSkyGain.get())
			attr.AddNode("integer", "nsamples", lights.varSkySamples.get())
			attr.AddNode("float", "turbidity", lights.varSkyTurbidity.get())
			attr.AddNode("float", "aconst", lights.aconst.get())
			attr.AddNode("float", "bconst", lights.bconst.get())
			attr.AddNode("float", "cconst", lights.cconst.get())
			attr.AddNode("float", "dconst", lights.dconst.get())
			attr.AddNode("float", "econst", lights.econst.get())
			if lights.Sun() != "None":
				attr.AddNode("vector", "sundir", str3(self.data.lights[lights.Sun()].SunVector()))	
		else:
			attr = None
		
		return attr
		
	def GetSun(self, justname=False):
		lights = self.pageExposure.pageLights
		
		sun = lights.Sun()
		if sun == "None":
			return sun, None
		else:
			if justname == False:
				attr = LuxAttribute("LightSource")
				attr.SetType("sun")
				attr.AddNode("float", "gain", lights.varSunGain.get())
				attr.AddNode("integer", "nsamples", lights.varSunSamples.get())
				attr.AddNode("float", "turbidity", lights.varSunTurbidity.get())
				attr.AddNode("float", "relsize", lights.varSunRelSize.get())
				attr.AddNode("vector", "sundir", str3(self.data.lights[sun].SunVector()))
			
				return sun, attr
			
			else:
				return sun

	def GetWorldVolume(self):
		o = self.pageOptions
		if o.varUseCustomWV.get() == 1:
			attr = LuxVolume()
			attr.SetType("WorldVolume")
			attr.SetVolumeType('"homogeneous"')
			attr.AddNode("texture", "fresnel", "Pose2LuxExterior")
			
			abs_r = o.Absorption_R.get()/100.0
			abs_g = o.Absorption_G.get()/100.0
			abs_b = o.Absorption_B.get()/100.0
			sca_r = o.Scattering_R.get()/100.0
			sca_g = o.Scattering_G.get()/100.0
			sca_b = o.Scattering_B.get()/100.0
			
			attr.AddNode("color", "sigma_a", str3((abs_r, abs_g, abs_b)))
			attr.AddNode("color", "sigma_s", str3((sca_r, sca_g, sca_b)))
			attr.AddNode("color", "g", str3((o.Asymmetry_R.get(),o.Asymmetry_G.get(),o.Asymmetry_B.get())))
		else:
			attr = None
		return attr
				
	def LibraryManager(self):
		return self.library

	def LoadMaterialLibrary(self):
		m = self.pageMaterials
		self.library.LoadMaterialLibrary(dir=self.DefaultFolders("materials"),menu=(m.optLibs, m.varLibs, m.UpdateLuxMats))
	
	def LoadScene(self):
		
		# Get input file
		initDir = self.DefaultFolders("scene")
		initFile = None
		path = ChooseLoadFile(initDir, initFile, "p2ls", "scene data")
		if path == "":
			return
		else:
			try:
				file = open(path)
			except:
				return
				
		# Make sure we have a valid XML file
		try:
			tree = parse(file)
			root = tree.getroot()
		except IOError, err:
			tkMessageBox.showerror("Pose2Lux", "Pose2Lux could not open the library file:\n%s" % path); return
		except Exception, err:
			tkMessageBox.showerror("Pose2Lux", "Pose2Lux encountered a problem with:\n%s\n%s" % (path, err)); return
			
		# Load generic settings
		self.LoadSettings(path=None, root=root)
		
		# Load camera settings
		element = root.find('camera')
		self.pageExposure.pageCamera.Load(element)
		
		# Load Lights page data
		element = root.find('lightspage')
		self.pageExposure.pageLights.Load(element)
		
		# Load light groups
		element = root.find('lightgroups')
		
		# Keep for compatibility:
		for light in element.getchildren():
			if light.tag != "LightGroup": self.AddLightGroup(light.tag)
			
		groups = element.findall("LightGroup")
		for g in groups:
			self.AddLightGroup(g.get("name"))
		
		# Load custom library
		self.library.LoadCustomLibrary(root)
		
		# Load scene data
		self.data.LoadAll(root)
		
		# Refresh GUI elements
		self.Refresh()
				
	def LoadSettings(self, path=None, root=None):
	
		if root == None:
		
			if path == None:			
				initDir, initFile = self.GetInitPath(self.settingsfile)
				path = ChooseLoadFile(initDir, initFile, "p2l", "settings")
			
			if path == "":
				return
			else:
				try:
					file = open(path)
					self.settingsfile = path
				except:
					return
				
			# Make sure we have a valid XML file
			try:
				tree = parse(file)
				root = tree.getroot()
			except IOError, err:
				tkMessageBox.showerror("Pose2Lux", "Pose2Lux could not open the library file:\n%s" % path); return
			except Exception, err:
				tkMessageBox.showerror("Pose2Lux", "Pose2Lux encountered a problem with:\n%s\n%s" % (path, err)); return

		# Get the settings Element
		element = root.find('UIParameters')
		for key, value in element.items():
			type, target = self.controls[key]
			if type == 'string': target.set(value)
			elif type == 'float': target.set(float(value))
			elif type == 'int': target.set(int(value))			
	
	def LoadTemplate(self):
		currentFig = self.GetCurrentFigure()
		if currentFig:
			name = currentFig.Name()
			initDir = self.DefaultFolders("template")
			initFile = None
			path = ChooseLoadFile(initDir, initFile, "p2lm", "template")
			if path:
				try:
					input = open(path)
				except:
					tkMessageBox.showerror("Pose2Lux", "Unable to open template file"); return
					
				try:
					for i in input:
						key, parm = i.split('=', 1)
						
						o, m = key.split('/')
						key = MakeKey(name,m)
						
						mat = self.data.materials.get(key, None)
						if mat:
							parm = parm.rstrip()
							if self.library.ValidMaterial(parm):
								mat.luxname = parm
				finally:
					input.close()
					self.pageMaterials.PopulateMaterialList()
				
		else:
			tkMessageBox.showwarning("Load template", "No figure selected")
	
	def Progress(self, done, max=None, text=None, colour=None):
		x = self.pageExport
		x.pb.updateProgress(done, newMax=max, newCol=colour)
		if text:
			x.varProgressText.set(text)
			x.ProgressText.update()
	
	def Refresh(self):
		self.pageMaterials.PopulateMaterialList()
		self.pageExposure.pageLights.lights = self.data.AllLightSources()
		self.pageExposure.pageLights.PopulateList()
	
	def RegisterControl(self, id, type, control):
		self.controls[id] = (type, control)
	
	def SaveScene(self):
		
		# Create the root XML
		root = Element('P2LSCENE')
		tree = ElementTree(root)
		
		# Get the output file, and exit if we don't get one
		initDir = self.DefaultFolders("scene")
		initFile = None
		file = ChooseSaveFile(initDir, initFile, "p2ls", "scene data")	
		if not file: return
		
		# Save generic settings
		self.SaveSettings(root=root)
		
		# Save camera settings
		attribs = self.pageExposure.pageCamera.Save()
		root.append(Element("camera", attribs))
		
		# Save Lights page data
		attribs = self.pageExposure.pageLights.Save()
		root.append(Element("lightspage", attribs))
		
		# Save light groups
		groups, var = self.GetLightGroups()
		element = Element("lightgroups")
		root.append(element)
		for g in groups:
			attribs = {"name":g}
			e = Element("LightGroup", attribs)
			element.append(e)
		
		# Save data
		self.data.SaveAll(root)
		
		# Save custom materials
		self.library.SaveCustomLibrary(root)
		
		# Write data to disk
		tree.write(file)
		try: file.close()
		except: pass
	
	def SaveSettings(self, root=None):
	
		# Create a dictionary to hold all the settings
		settings = {}
		for key, value in self.controls.items():
			settings[key] = str(value[1].get())
		
		# Get the root element
		if root == None:
			root = Element('P2LSETTINGS')
			tree = ElementTree(root)
			writefile = True
		else:
			writefile = False
		
		# Add the settings to the XML
		element = Element('UIParameters', settings)
		root.append(element)
		
		# If we weren't given a root then write the file to disk
		# (If we were, then the responsibility for writing to disk is with the calling routine)
		
		if writefile:
		
			initDir, initFile = self.GetInitPath(self.settingsfile)
			file = ChooseSaveFile(initDir, initFile, "p2l", "settings")

			# Save to disk
			if file:
				tree.write(file)
				try: file.close()
				except: pass

	def SaveTemplate(self):
		currentFig = self.GetCurrentFigure()
		if currentFig:
			initDir = self.DefaultFolders("template")
			initFile = None
			file = ChooseSaveFile(initDir, initFile, "p2lm", "template")
			if file:
				data = []
				for m in currentFig.Materials():
					if m.Luxname() is not None:
						key = 'PL2_GENERIC/%s' % m.Name()
						data.append(key + '=' + m.Luxname() + '\n')
				try:
					file.writelines(data)
					tkMessageBox.showinfo("Pose2Lux", "Template saved")
				finally:
					file.close()
		else:
			tkMessageBox.showwarning("Save template", "No figure selected")
class pageCamera:
	def __init__(self, gui, parent):
		self.mainFrame = parent
		
		self.frmTonemapping = Frame(self.mainFrame)
		self.frmTonemapping.pack(side=TOP, padx=10, pady=10)
		
		self.frmDOF = Frame(self.mainFrame)
		self.frmDOF.pack(side=TOP, padx=10, fill=X)
		
		self.fStops = ["32", "22", "16", "11", "8", "5.6", "4", "2.8", "2", "1.4", "1"]
		self.shutterSpeeds = ["1/4000","1/2000","1/1000","1/500","1/250","1/125","1/60","1/30","1/15","1/8","1/4","1/2","1","2","4","8","15","30"]
		
		self.InitTonemapping()
		self.ResetDefaults()
		self.InitDOF()
		
		self.saveload = {
			"ISO" : "integer",
			"Shutter" : "integer",
			"fStop" : "integer",
			"Gamma" : "float",
			"varUseDOF" : "integer"
			}

	def GetEV(self):
		isoEV = self.GetISOEV(self.ISO.get())
		fStopEV = self.fStop.get()
		shutterEV = self.Shutter.get()
		return isoEV + shutterEV + fStopEV + 6
		
	def GetfStop(self):
		return(self.fStops[self.fStop.get()])
		
	def GetGamma(self):
		return(self.Gamma.get())
		
	def GetISO(self):
		return(self.ISO.get())
		
	def GetISOEV(self, iso):
		if iso >= 3200: return 7
		if iso >= 1600: return 6
		if iso >= 800: return 5
		if iso >= 400: return 4
		if iso >= 200: return 3
		if iso >= 100: return 2
		return 1
		
	def GetShutter(self):
		return(eval(self.shutterSpeeds[self.Shutter.get()]+".0"))
	
	def InitDOF(self):
		Label(self.frmDOF, text="DEPTH OF FIELD", bg="black", fg="white").pack(side=TOP, fill=X)
		self.varUseDOF = IntVar()
		Checkbutton(self.frmDOF, text="Use Depth of Field", anchor=W, variable=self.varUseDOF).pack(side=TOP, fill=X, pady=10)
		Message(self.frmDOF, aspect=100, text = "To use DOF in your scene:\n\nUse Poser's Focus Distance Guide to set the camera's focus_Distance parameter.\n\nRun Pose2Lux and set the fStop value to modify the DOF. \n\nSwitch on DOF by selecting the Use Depth of Field checkbox control.\n\nNB: While you can subsequently adjust the fStop value in the Lux GUI it will only affect the exposure, not the DOF. The DOF is calculated at the time of export and cannot be changed during the render.").pack(side=TOP)
		
	def InitTonemapping(self):
		
		Label(self.frmTonemapping, text="EXPOSURE", bg="black", fg="white").pack(side=TOP, fill=X)
		
		self.ISO = Scale(self.frmTonemapping, orient=HORIZONTAL, length=200, from_=50, to=3200, resolution=50, label="ISO", showvalue=0, command=self.UpdateISO)
		self.ISO.pack(side=TOP)
		
		self.Shutter = Scale(self.frmTonemapping, orient=HORIZONTAL, length=200, from_=0, to=17, label="Shutter", showvalue=0, command=self.UpdateShutter)
		self.Shutter.pack(side=TOP)
		
		self.fStop = Scale(self.frmTonemapping, orient=HORIZONTAL, length=200, from_=0, to=10, label="fStop", showvalue=0, command=self.UpdatefStop)
		self.fStop.pack(side=TOP)
		
		self.Gamma = Scale(self.frmTonemapping, orient=HORIZONTAL, length=200, from_=1, to=3, resolution=0.1, label="Gamma = x.x", showvalue=0, command=self.UpdateGamma)
		self.Gamma.pack(side=TOP)
		
		self.varEV = StringVar()
		self.varDifference = StringVar()
		Label(self.frmTonemapping, textvariable=self.varEV, anchor=W).pack(side=TOP, fill=X)
		Label(self.frmTonemapping, text="Presets:", anchor=W).pack(side=TOP, fill=X)
		self.frmPresets = Frame(self.frmTonemapping)
		self.frmPresets.pack(side=TOP, fill=X)
		Button(self.frmPresets, text="SUN", command=self.SetSun).pack(side=LEFT)
		Button(self.frmPresets, text="SKY", command=self.SetSky).pack(side=LEFT)
		Button(self.frmPresets, text="SUN & SKY", command=self.SetSunSky).pack(side=LEFT)
		Button(self.frmPresets, text="SKYDOME", command=self.SetSkydome).pack(side=LEFT)		
		Button(self.frmPresets, text="Reset", command=self.ResetDefaults).pack(side=LEFT)

	def Load(self, data):
		for key, value in data.items():
			getattr(self, key).set(Unpack(value, self.saveload[key]))
		
	def ResetDefaults(self):
		self.ISO.set(200)
		self.Shutter.set(4)
		self.fStop.set(4)
		self.Gamma.set(1.0)
		self.UpdateExposure()

	def Save(self):
		attribs = {}
		for key, value in self.saveload.items():
			attribs[key] = str(getattr(self, key).get())
		return attribs
	
	def SetSky(self):
		self.ISO.set(200)
		self.Shutter.set(4)
		self.fStop.set(8)
		self.Gamma.set(1.0)
		self.UpdateExposure()

	def SetSun(self):
		self.ISO.set(150)
		self.Shutter.set(4)
		self.fStop.set(2)
		self.Gamma.set(1.0)
		self.UpdateExposure()

	def SetSunSky(self):
		self.ISO.set(100)
		self.Shutter.set(3)
		self.fStop.set(2)
		self.Gamma.set(1.0)
		self.UpdateExposure()

	def SetSkydome(self):
		self.ISO.set(100)
		self.Shutter.set(3)
		self.fStop.set(2)
		self.Gamma.set(1.0)
		self.UpdateExposure()
		
	def UpdateExposure(self):
		ev = self.GetEV()
		self.varEV.set('Exposure value (EV) = %s' % ev)
		
	def UpdatefStop(self, value):
		self.fStop.config(label = "fStop = f" + self.fStops[int(value)])
		self.UpdateExposure()
		
	def UpdateGamma(self, value):
		self.Gamma.config(label = "Gamma = " + str(value))
		self.UpdateExposure()
	
	def UpdateISO(self, value):
		self.ISO.config(label = "ISO = " + str(value))
		self.UpdateExposure()
	
	def UpdateShutter(self, value):
		self.Shutter.config(label = "Shutter speed = " + self.shutterSpeeds[int(value)] + " second(s)")
		self.UpdateExposure()
	
	def UpdateTarget(self, value):
		self.UpdateExposure()
	
	def UseDOF(self):
		return (False, True)[self.varUseDOF.get()]
class pageCameraLights:
	def __init__(self, parent, gui):
		
		self.mainFrame = Frame(parent)
		
		self.frmCamera = Frame(self.mainFrame, bd=2, relief=GROOVE)
		self.frmCamera.pack(side=LEFT, fill=Y)
		
		self.spacer = Frame(self.mainFrame, width=40)
		self.spacer.pack_propagate(False)
		self.spacer.pack(side=LEFT)
		
		self.frmLights = Frame(self.mainFrame, bd=2, relief=GROOVE)
		self.frmLights.pack(side=LEFT, fill=BOTH, expand=1)
		
		Label(self.frmCamera, text="CAMERA settings", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)
		Label(self.frmLights, text="LIGHTS", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)
		
		self.pageCamera = pageCamera(gui, self.frmCamera)
		self.pageLights = pageLights(gui, self.frmLights)
class pageDebug:
	def __init__(self, parent, gui):
		
		self.gui = gui
		
		self.mainFrame = Frame(parent)
		Label(self.mainFrame, text="DEBUG", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)

		self.notebook = Notebook(self.mainFrame, TOP)
		self.pageDebug = debug_MATView(self.notebook(), self.gui)
		self.notebook.AddScreen(self.pageDebug.mainFrame, "MAT View")			
class pageExport:
	def __init__(self, parent, gui):
	
		global titleFont
		
		self.gui = gui
		self.currentPanel = None
		self.panels = {}
		
		# Frame definitions
		self.mainFrame = Frame(parent)
		
		Label(self.mainFrame, text="EXPORT settings", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)
		
		self.frmParameters = Frame(self.mainFrame)
		self.frmParameters.pack(side=LEFT, fill=Y)
		
		self.frmConfig = Frame(self.mainFrame, width=250, bd=2, relief=RIDGE)
		self.frmConfig.pack(side=LEFT, fill=BOTH, padx=20, pady=20)
		self.frmConfig.pack_propagate(False)
		
		Frame(self.frmParameters, height=20).pack(side=TOP, fill=X)
		
		self.frmExport = Frame(self.frmParameters, bd=2, relief=RIDGE)
		self.frmExport.pack(side=TOP, fill=X)
		
		self.frmProgress = Frame(self.frmParameters)
		self.frmProgress.pack(side=TOP, fill=X, pady=5)
		
		self.frmColumnA = Frame(self.frmParameters)
		self.frmColumnA.pack(side=LEFT, fill=Y)
		self.frmColumnB = Frame(self.frmParameters)
		self.frmColumnB.pack(side=LEFT, fill=Y, padx=50)
		self.frmColumnC = Frame(self.frmParameters)
		self.frmColumnC.pack(side=LEFT, fill=Y)
		self.frmColumnD = Frame(self.frmParameters)
		self.frmColumnD.pack(side=RIGHT, fill=Y)
		
		# Column A
		self.frmExportFiles = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmExportFiles.pack(side=TOP, fill=X)
		
		self.frmResolution = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmResolution.pack(side=TOP, fill=X, pady=20)
		
		self.frmImageOutput = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmImageOutput.pack(side=TOP, fill=X)
		
		# Column B
		self.frmRenderer = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmRenderer.pack(side=TOP, fill=X)
		
		self.frmSampler = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmSampler.pack(side=TOP, fill=X, pady=20)
		
		self.frmPixelFilter = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmPixelFilter.pack(side=TOP, fill=X)

		# Column C
		self.frmAccelerator = Frame(self.frmColumnC, bd=2, relief=RIDGE)
		self.frmAccelerator.pack(side=TOP, fill=X)
		
		self.frmSurfaceIntegrator = Frame(self.frmColumnC, bd=2, relief=RIDGE)
		self.frmSurfaceIntegrator.pack(side=TOP, fill=X, pady=20)

		self.frmVolumeIntegrator = Frame(self.frmColumnC, bd=2, relief=RIDGE)
		self.frmVolumeIntegrator.pack(side=TOP, fill=X)
		
		self.frmMiscellaneous = Frame(self.frmColumnC, bd=2, relief=RIDGE)
		self.frmMiscellaneous.pack(side=TOP, fill=X, pady=20)
		
		# Column D
		self.frmFilm = Frame(self.frmColumnD, bd=2, relief=RIDGE)
		self.frmFilm.pack(side=TOP, fill=X)
		
		self.frmScaleColours = Frame(self.frmColumnD, bd=2, relief=RIDGE)
		self.frmScaleColours.pack(side=TOP, fill=X, pady=20)
		
		# Panel definitions
		self.InitConfigPanel()
		self.InitExportLocation()
		self.InitProgressBar()
		self.InitExportFiles()
		self.InitOutputResolution()
		self.InitImageOutput()
		self.InitRenderer()
		self.InitSampler()
		self.InitPixelFilter()
		self.InitAccelerator()
		self.InitSurfaceIntegrator()
		self.InitVolumeIntegrator()
		self.InitFilm()
		self.InitScaleColours()
		self.InitMiscellaneous()

	def AddPanel(self, panel, name):
		self.panels[name] = panel
		
	def ChooseOutputFile(self):
	
		output = os.path.split(self.varExportLocation.get())
		if output[0] == "": initDir = self.gui.DefaultFolders("savescene")
		else: initDir = output[0]
		
		if output[1] == "": initFile = "MyLuxScene.lxs"
		else: initFile = output[1]
		
		result = tkFileDialog.asksaveasfilename(
			defaultextension=[("*.lxs")],
			filetypes=[("LuxRender scene", "*.lxs")],
			initialdir=initDir,
			initialfile=initFile,
			title="Select output file for your Luxrender scene")
			
		if result != "": self.varExportLocation.set(result)
	
	def ConfigPixelFilterBox(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "box")
		
		Label(frame, text="Pixel Filter : Box").pack(side=TOP, pady=10, fill=X)
		
		self.varPFBoxXWidth = DoubleVar(); self.varPFBoxXWidth.set(0.5)
		self.varPFBoxYWidth = DoubleVar(); self.varPFBoxYWidth.set(0.5)
		Scale(frame, label="X Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFBoxXWidth).pack(side=TOP, fill=X)
		Scale(frame, label="Y Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFBoxYWidth).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("PFBoxXWidth", "float", self.varPFBoxXWidth)
		self.gui.RegisterControl("PFBoxYWidth", "float", self.varPFBoxYWidth)
	
	def ConfigPixelFilterTriangle(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "triangle")
		
		Label(frame, text="Pixel Filter : Triangle").pack(side=TOP, pady=10, fill=X)	

		self.varPFTriangleXWidth = DoubleVar(); self.varPFTriangleXWidth.set(2.0)
		self.varPFTriangleYWidth = DoubleVar(); self.varPFTriangleYWidth.set(2.0)
		Scale(frame, label="X Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFTriangleXWidth).pack(side=TOP, fill=X)
		Scale(frame, label="Y Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFTriangleYWidth).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("PFTriangleXWidth", "float", self.varPFTriangleXWidth)
		self.gui.RegisterControl("PFTriangleYWidth", "float", self.varPFTriangleYWidth)
	
	def ConfigPixelFilterGaussian(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "gaussian")
		
		Label(frame, text="Pixel Filter : Gaussian").pack(side=TOP, pady=10, fill=X)	

		self.varPFGaussianXWidth = DoubleVar(); self.varPFGaussianXWidth.set(2.0)
		self.varPFGaussianYWidth = DoubleVar(); self.varPFGaussianYWidth.set(2.0)
		Scale(frame, label="X Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFGaussianXWidth).pack(side=TOP, fill=X)
		Scale(frame, label="Y Width", orient=HORIZONTAL, from_= 0.1, to=2.0, resolution=0.1, variable=self.varPFGaussianYWidth).pack(side=TOP, fill=X)
		
		self.varPFGaussianAlpha = DoubleVar(); self.varPFGaussianAlpha.set(2.0)
		Scale(frame, label="Alpha", orient=HORIZONTAL, from_=0.1, to=10.0, resolution=0.1, variable=self.varPFGaussianAlpha).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("PFGaussianXWidth", "float", self.varPFGaussianXWidth)
		self.gui.RegisterControl("PFGaussianYWidth", "float", self.varPFGaussianYWidth)
		self.gui.RegisterControl("PFGaussianAlpha", "float", self.varPFGaussianAlpha)
		
	def ConfigPixelFilterMitchell(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "mitchell")
		
		Label(frame, text="Pixel Filter : Mitchell").pack(side=TOP, pady=10, fill=X)	

		self.varPFMitchellXWidth = DoubleVar(); self.varPFMitchellXWidth.set(2.0)
		self.varPFMitchellYWidth = DoubleVar(); self.varPFMitchellYWidth.set(2.0)
		Scale(frame, label="X Width", orient=HORIZONTAL, from_= 0.1, to=3.0, resolution=0.1, variable=self.varPFMitchellXWidth).pack(side=TOP, fill=X)
		Scale(frame, label="Y Width", orient=HORIZONTAL, from_= 0.1, to=3.0, resolution=0.1, variable=self.varPFMitchellYWidth).pack(side=TOP, fill=X)
		
		self.varPFB = DoubleVar(); self.varPFB.set(0.35)
		self.varPFC = DoubleVar(); self.varPFC.set(0.35)
		Scale(frame, label="B", orient=HORIZONTAL, from_=0.0, to=1.0, resolution=0.05, variable=self.varPFB).pack(side=TOP, fill=X)
		Scale(frame, label="C", orient=HORIZONTAL, from_=0.0, to=1.0, resolution=0.05, variable=self.varPFC).pack(side=TOP, fill=X)
		
		self.varPFSupersample = IntVar()
		Checkbutton(frame, text="Supersample", variable=self.varPFSupersample, anchor=W).pack(side=TOP, fill=X, pady=5)
		
		self.gui.RegisterControl("PFMitchellXWidth", "float", self.varPFMitchellXWidth)
		self.gui.RegisterControl("PFMitchellYWidth", "float", self.varPFMitchellYWidth)
		self.gui.RegisterControl("PFB", "float", self.varPFB)
		self.gui.RegisterControl("PFC", "float", self.varPFC)
		self.gui.RegisterControl("PFSuperSample", "int", self.varPFSupersample)
		
	def ConfigPixelFilterSinc(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "sinc")
		
		Label(frame, text="Pixel Filter : Sinc").pack(side=TOP, pady=10, fill=X)	

		self.varPFSincXWidth = DoubleVar(); self.varPFSincXWidth.set(4.0)
		self.varPFSincYWidth = DoubleVar(); self.varPFSincYWidth.set(4.0)
		Scale(frame, label="X Width", orient=HORIZONTAL, from_= 0.1, to=5.0, resolution=0.1, variable=self.varPFSincXWidth).pack(side=TOP, fill=X)
		Scale(frame, label="Y Width", orient=HORIZONTAL, from_= 0.1, to=5.0, resolution=0.1, variable=self.varPFSincYWidth).pack(side=TOP, fill=X)
		
		self.varPFSincTau = DoubleVar(); self.varPFSincTau.set(3.0)
		Scale(frame, label="Tau", orient=HORIZONTAL, from_=0.1, to=10.0, resolution=0.1, variable=self.varPFSincTau).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("PFSincXWidth", "float", self.varPFSincXWidth)
		self.gui.RegisterControl("PFSincYWidth", "float", self.varPFSincYWidth)
		self.gui.RegisterControl("PFSincTau", "float", self.varPFSincTau)
		
	def ConfigSamplerERPT(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "erpt")
		
		Label(frame, text="Sampler : ERPT").pack(side=TOP, pady=10, fill=X)

		self.varInitSamples = IntVar(); self.varInitSamples.set(100000)
		Scale(frame, label="Initial samples", orient=HORIZONTAL, from_=10000, to=1000000, resolution=10000, variable=self.varInitSamples).pack(side=TOP, pady=5, fill=X)

		self.varChainLength = IntVar(); self.varChainLength.set(2000)
		Scale(frame, label="Chain length", orient=HORIZONTAL, from_=1, to=32768, resolution=500, variable=self.varChainLength).pack(side=TOP, pady=5, fill=X)
		
		self.gui.RegisterControl("initsamples", "int", self.varInitSamples)
		self.gui.RegisterControl("chainlength", "int", self.varChainLength)
		
	def ConfigSamplerLowdiscrepancy(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "lowdiscrepancy")
		
		Label(frame, text="Sampler : Low discrepancy").pack(side=TOP, pady=10, fill=X)

		self.varLowdiscrepancyPixelSamples = IntVar(); self.varLowdiscrepancyPixelSamples.set(4)
		Scale(frame, label="Pixel samples", orient=HORIZONTAL, from_=1, to=512, resolution=1, variable=self.varLowdiscrepancyPixelSamples).pack(side=TOP, pady=5, fill=X)
		
		self.varLowdiscrepancyPixelSampler = StringVar(); self.varLowdiscrepancyPixelSampler.set("vegas")
		Label(frame, text="Pixel sampler:", anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Hilbert", value="hilbert", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Linear", value="linear", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Vegas", value="vegas", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Low discrepancy", value="lowdiscrepancy", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Tile", value="tile", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Random", value="random", variable=self.varLowdiscrepancyPixelSampler, anchor=W).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("LowDPixelSamples", "int", self.varLowdiscrepancyPixelSamples)
		self.gui.RegisterControl("LowDPixelSampler", "string", self.varLowdiscrepancyPixelSampler)
		
	def ConfigSamplerMetropolis(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "metropolis")
		
		Label(frame, text="Sampler : Metropolis").pack(side=TOP, pady=10, fill=X)

		self.varMetropolisMaxConsecRejects = IntVar(); self.varMetropolisMaxConsecRejects.set(512)
		Scale(frame, label="Maximum consecutive rejects", orient=HORIZONTAL, from_=0, to=32768, resolution=128, variable=self.varMetropolisMaxConsecRejects).pack(side=TOP, pady=5, fill=X)
		
		self.varMetropolisLargeMutationProbability = DoubleVar(); self.varMetropolisLargeMutationProbability.set(0.4)
		Scale(frame, label="Large mutation probability", orient=HORIZONTAL, from_=0, to=1, resolution=0.05, variable=self.varMetropolisLargeMutationProbability).pack(side=TOP, pady=5, fill=X)
		
		self.gui.RegisterControl("MetroMCR", "int", self.varMetropolisMaxConsecRejects)
		self.gui.RegisterControl("MetroLMP", "float", self.varMetropolisLargeMutationProbability)
		
	def ConfigSamplerRandom(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "random")
		
		Label(frame, text="Sampler : Random").pack(side=TOP, pady=10, fill=X)
		
		self.varRandomPixelSamples = IntVar(); self.varRandomPixelSamples.set(4)
		Scale(frame, label="Pixel samples", orient=HORIZONTAL, from_=1, to=512, resolution=1, variable=self.varRandomPixelSamples).pack(side=TOP, pady=5, fill=X)
		
		self.varRandomPixelSampler = StringVar(); self.varRandomPixelSampler.set("vegas")
		Label(frame, text="Pixel sampler:", anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Hilbert", value="hilbert", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Linear", value="linear", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Vegas", value="vegas", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Low discrepancy", value="lowdiscrepancy", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Tile", value="tile", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(frame, text="Random", value="random", variable=self.varRandomPixelSampler, anchor=W).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("RandomPixelSamples", "int", self.varRandomPixelSamples)
		self.gui.RegisterControl("RandomPixelSampler", "string", self.varRandomPixelSampler)
	
	def ConfigSurfaceBidirectional(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "bidirectional")
		
		Label(frame, text="Surface Integrator : Bidirectional").pack(side=TOP, pady=10, fill=X)
		
		# eyedepth
		self.bidirEyedepth = Scale(frame, orient=HORIZONTAL, length=400, from_=1, to=50, resolution=1, label="Eye depth")
		self.bidirEyedepth.set(16)
		self.bidirEyedepth.pack(side=TOP)
		
		# lightdepth
		self.bidirLightdepth = Scale(frame, orient=HORIZONTAL, length=400, from_=1, to=50, resolution=1, label="Light depth")
		self.bidirLightdepth.set(16)
		self.bidirLightdepth.pack(side=TOP)

		# strategy
		strategylist = ("one", "all", "auto")
		self.varStrategy = StringVar()
		
		frmStrategy = Frame(frame)
		frmStrategy.pack(side=TOP, fill=X)
		Label(frmStrategy, text="Strategy:", anchor=W).pack(side=LEFT, fill=X)
		self.varStrategy.set(strategylist[2])
		self.bidirStrategy = OptionMenu(frmStrategy, self.varStrategy, *strategylist)
		self.bidirStrategy.pack(side=LEFT, pady=5)

		self.gui.RegisterControl("bidirEyedepth", "int", self.bidirEyedepth)
		self.gui.RegisterControl("bidirLightdepth", "int", self.bidirLightdepth)
		self.gui.RegisterControl("bidirStrategy", "string", self.varStrategy)
		
	def ConfigSurfacePath(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "path")
		
		Label(frame, text="Surface Integrator : Path").pack(side=TOP, pady=10, fill=X)	

		# maxdepth
		self.pathMaxDepth = Scale(frame, orient=HORIZONTAL, length=400, from_=1, to=40, resolution=1, label="Max depth")
		self.pathMaxDepth.set(16)
		self.pathMaxDepth.pack(side=TOP)
		
		# lightstrategy
		lightstrategylist = ("one", "all", "auto", "importance", "powerimp", "allpowerimp", "logpowerimp")
		frmPathLightStrategy = Frame(frame)
		frmPathLightStrategy.pack(side=TOP, fill=X)
		self.varPathLightStrategy = StringVar()
		self.varPathLightStrategy.set(lightstrategylist[2])
		Label(frmPathLightStrategy, text="Light strategy:", anchor=W).pack(side=LEFT, fill=X)
		OptionMenu(frmPathLightStrategy, self.varPathLightStrategy, *lightstrategylist).pack(side=LEFT, pady=5)
		
		# rrstrategy
		frmRRStrategy = Frame(frame)
		frmRRStrategy.pack(side=TOP, fill=X, pady=5)
		self.varPathRRStrategy = StringVar(); self.varPathRRStrategy.set("efficiency")
		Label(frmRRStrategy, text="RR Strategy", anchor=W).pack(side=LEFT)
		OptionMenu(frmRRStrategy, self.varPathRRStrategy, "none", "probability", "efficiency").pack(side=LEFT)
		
		#rrcontinueprob
		self.varPathRRContinueProb = DoubleVar(); self.varPathRRContinueProb.set(0.65)
		Scale(frame, label="RR continue probability", orient=HORIZONTAL, from_=0.0, to=1.0, resolution=0.05, variable=self.varPathRRContinueProb).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("pathMaxdepth", "int", self.pathMaxDepth)
		self.gui.RegisterControl("pathLightStrategy", "string", self.varPathLightStrategy)
		self.gui.RegisterControl("pathRRStrategy", "string", self.varPathRRStrategy)
		self.gui.RegisterControl("pathRRContProb", "float", self.varPathRRContinueProb)
		
	def ConfigSurfaceExphotonmap(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "exphotonmap")
		
		Label(frame, text="Surface Integrator : Exphotonmap").pack(side=TOP, pady=10, fill=X)
		
		# maxdepth
		self.exphotonMaxDepth = Scale(frame, orient=HORIZONTAL, from_=1, to=20, resolution=1, label="Max depth")
		self.exphotonMaxDepth.set(8)
		self.exphotonMaxDepth.pack(side=TOP, fill=X)		

		# maxphotondepth
		self.exphotonMaxPhotonDepth = Scale(frame, orient=HORIZONTAL, from_=1, to=20, resolution=1, label="Max photon depth")
		self.exphotonMaxPhotonDepth.set(10)
		self.exphotonMaxPhotonDepth.pack(side=TOP, fill=X)

		# directphotons
		self.exphotonDirectphotons = Scale(frame, orient=HORIZONTAL, from_=50000, to=2000000, resolution=50000, label="Direct photonmap size")
		self.exphotonDirectphotons.set(1000000)
		self.exphotonDirectphotons.pack(side=TOP, fill=X)

		# causticphotons
		self.exphotonCausticphotons = Scale(frame, orient=HORIZONTAL, from_=5000, to=100000, resolution=5000, label="Caustic photonmap size")
		self.exphotonCausticphotons.set(20000)
		self.exphotonCausticphotons.pack(side=TOP, fill=X)		

		# indirectphotons
		self.exphotonIndirectphotons = Scale(frame, orient=HORIZONTAL, from_=50000, to=2000000, resolution=50000, label="Indirect photonmap size")
		self.exphotonIndirectphotons.set(200000)
		self.exphotonIndirectphotons.pack(side=TOP, fill=X)

		# radiancephotons
		self.exphotonRadiancephotons = Scale(frame, orient=HORIZONTAL, from_=50000, to=2000000, resolution=50000, label="Radiance photonmap size")
		self.exphotonRadiancephotons.set(200000)
		self.exphotonRadiancephotons.pack(side=TOP, fill=X)

		# nphotonsused
		self.exphotonNphotons = Scale(frame, orient=HORIZONTAL, from_=5, to=100, resolution=5, label="Max photons used")
		self.exphotonNphotons.set(50)
		self.exphotonNphotons.pack(side=TOP, fill=X)

		# maxphotondist
		self.exphotonMaxDist = Scale(frame, orient=HORIZONTAL, from_=0.05, to=1, resolution=0.05, label="Maximum distance")
		self.exphotonMaxDist.set(0.1)
		self.exphotonMaxDist.pack(side=TOP, fill=X)
		
		# finalgather
		self.varFinalGather = IntVar()
		self.varFinalGather.set(1)
		self.FinalGather = Checkbutton(frame, anchor=W, text="Final Gather", variable=self.varFinalGather)
		self.FinalGather.pack(side=TOP, fill=X)
		
		# renderingmode
		self.varRenderMode = StringVar()
		frmRenderStrategy = Frame(frame)
		frmRenderStrategy.pack(side=TOP, fill=X)
		Label(frmRenderStrategy, text="Render strategy:").pack(side=LEFT)
		self.varRenderMode.set("directlighting")
		OptionMenu(frmRenderStrategy, self.varRenderMode, "directlighting", "path").pack(side=LEFT)
		
		self.gui.RegisterControl("ExphotonMaxdepth", "int", self.exphotonMaxDepth)
		self.gui.RegisterControl("ExphotonMaxPhotondepth", "int", self.exphotonMaxPhotonDepth)
		self.gui.RegisterControl("ExphotonDirectphotons", "int", self.exphotonDirectphotons)
		self.gui.RegisterControl("ExphotonCausticphotons", "int", self.exphotonCausticphotons)
		self.gui.RegisterControl("ExphotonIndirectphotons", "int", self.exphotonIndirectphotons)
		self.gui.RegisterControl("ExphotonRadiancephotons", "int", self.exphotonRadiancephotons)
		self.gui.RegisterControl("ExphotonNUsed", "int", self.exphotonNphotons)	
		self.gui.RegisterControl("ExphotonNUsed", "int", self.exphotonNphotons)	
		self.gui.RegisterControl("ExphotonMaxdist", "float", self.exphotonMaxDist)	
		self.gui.RegisterControl("ExphotonFinalgather", "int", self.varFinalGather)	
		self.gui.RegisterControl("ExphotonRenderingmode", "string", self.varRenderMode)	
		
	def ConfigSurfaceDirectLighting(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "directlighting")
		
		Label(frame, text="Surface Integrator : Direct lighting").pack(side=TOP, pady=10, fill=X)

		# maxdepth
		self.directMaxDepth = Scale(frame, orient=HORIZONTAL, length=400, from_=1, to=20, resolution=1, label="Max depth")
		self.directMaxDepth.set(5)
		self.directMaxDepth.pack(side=TOP)
		
		# lightstrategy
		lightstrategylist = ("one", "all", "auto", "importance", "powerimp", "allpowerimp", "logpowerimp")
		self.varDLLightStrategy = StringVar()
		
		frmLightStrategy = Frame(frame)
		frmLightStrategy.pack(side=TOP, fill=X, pady=10)
		Label(frmLightStrategy, text="Light strategy:", anchor=W).pack(side=LEFT, fill=X)
		self.varDLLightStrategy.set(lightstrategylist[2])
		OptionMenu(frmLightStrategy, self.varDLLightStrategy, *lightstrategylist).pack(side=LEFT, pady=5)
		
		self.gui.RegisterControl("DLMaxdepth", "int", self.directMaxDepth)
		self.gui.RegisterControl("DLLightstrategy", "string", self.varDLLightStrategy)
		
	def ConfigSurfaceDistributedPath(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, "distributedpath")
		
		Label(frame, text="Surface Integrator : Distributed path").pack(side=TOP, pady=10, fill=X)

		# strategy
		strategylist = ("one", "all", "auto")
		self.varDPStrategy = StringVar()
		frmDPStrategy = Frame(frame)
		frmDPStrategy.pack(side=TOP, fill=X)
		Label(frmDPStrategy, text="Strategy:").pack(side=LEFT)
		self.varDPStrategy.set(strategylist[2])
		OptionMenu(frmDPStrategy, self.varDPStrategy, *strategylist).pack(side=LEFT)
		
		self.gui.RegisterControl("DPStrategy", "string", self.varDPStrategy)
	
	def ConfigSurfaceSPPM(self):
		frame = Frame(self.frmConfig)
		self.AddPanel(frame, 'sppm')
		
		Label(frame, text="Surface Integrator : SPPM").pack(side=TOP, pady=10, fill=X)
		
		# photonperpass
		self.SPPMppp = PickScale(frame, type="int", from_=0, to=1000000000, resolution=1000000, label='photonperpass')
		self.SPPMppp.set(1000000)
		
		# maxeyedepth
		self.SPPMmed = PickScale(frame, type="int", from_=0, to=2048, resolution=1, label='maxeyedepth')
		self.SPPMmed.set(16)
		
		# maxphotondepth
		self.SPPMmpd = PickScale(frame, type="int", from_=0, to=2048, resolution=1, label='maxphotondepth')
		self.SPPMmpd.set(16)
		
		# glossythreshold
		self.SPPMglossy = PickScale(frame, type="float", from_=0.0, to=100000.0, resolution=10.0, label="glossythreshold")
		self.SPPMglossy.set(100.0)
		
		# startk
		self.SPPMstartk = PickScale(frame, type="int", from_=0, to=2048, resolution=1, label='startk')
		self.SPPMstartk.set(15)
		
		# startradius
		self.SPPMsr = LabelEntry(frame, initValue=2.0, expand=1, text="startradius", type="float", minVal=0.0, maxVal=100.0, formatstring="%.3f")
		
		# alpha
		self.SPPMalpha = LabelEntry(frame, initValue=0.700, expand=1, text="alpha", type="float", minVal=0.001, maxVal=1.0, formatstring="%.3f")
		
		# includeenvironment
		self.SPPMincludeEnv = Boolean(frame, label="Include environment", text=None, initValue=True)
		
		# lookupaccel
		self.varSPPMlookupaccel = StringVar()
		self.varSPPMlookupaccel.set("hybridhashgrid")
		self.SPPMlookupaccel = LabelledMenu(frame, self.varSPPMlookupaccel, ["hybridhasgrid", "kdtree", "grid"], "lookupaccel")
		
		# pixelsampler
		self.varSPPMpixelsampler = StringVar()
		self.varSPPMpixelsampler.set("hilbert")
		self.SPPMpixelsampler = LabelledMenu(frame, self.varSPPMpixelsampler, ["hilbert", "linear", "tile", "vegas"], "pixelsampler")
		
		# photonsampler
		self.varSPPMphotonsampler = StringVar()
		self.varSPPMphotonsampler.set("halton")
		self.SPPMphotonsampler = LabelledMenu(frame, self.varSPPMphotonsampler, ["halton", "amc"], "photonsampler")
		
		self.gui.RegisterControl("SPPMppp", "int", self.SPPMppp)
		self.gui.RegisterControl("SPPMmed", "int", self.SPPMmed)
		self.gui.RegisterControl("SPPMmpd", "int", self.SPPMmpd)
		self.gui.RegisterControl("SPPMglossy", "float", self.SPPMglossy)
		self.gui.RegisterControl("SPPMstartk", "int", self.SPPMstartk)
		self.gui.RegisterControl("SPPMstartradius", "float", self.SPPMsr)
		self.gui.RegisterControl("SPPMalpha", "float", self.SPPMalpha)
		self.gui.RegisterControl("SPPMinclenv", "bool", self.SPPMincludeEnv)
		self.gui.RegisterControl("SPPMlookup", "string", self.varSPPMlookupaccel)
		self.gui.RegisterControl("SPPMpixelsampler", "string", self.varSPPMpixelsampler)
		self.gui.RegisterControl("SPPMphotonsampler", "string", self.varSPPMphotonsampler)
		
	def ConfigurePixelFilter(self):
		self.SetPanel(self.varPixelFilter.get())
		
	def ConfigureSampler(self):
		self.SetPanel(self.varSampler.get())
	
	def ConfigureSurfaceIntegrator(self):
		self.SetPanel(self.varSurfaceIntegrator.get())
		
	def CopyOutputName(self):
		copyfrom = os.path.basename(self.varExportLocation.get())
		if copyfrom != "":
			if copyfrom.endswith('.lxs'): copyfrom = copyfrom[:-4]
		self.varLuxName.set(copyfrom)

	def Export(self):
		
		# Perform basic validation
		if self.ValidateResolution() == False: return
		files = self.ValidateOutputFile()
		if files == None: return
		
		# Export scene
		self.ExportButton.config(state=DISABLED, bg="black", text="Exporting to Lux ..")
		exporter = Exporter(self.gui)
		exporter.SetVerbosity(0)
		exporter.SetOutputFiles(files)
		exporter.Export()
		self.ExportButton.config(state=NORMAL, bg="red", text="EXPORT TO LUX")
		
		# Try to close all files
		for f in files:
			if f:
				try:
					close(f)
				except:
					continue
					
	def ExportLocation(self):
		return self.varExportLocation.get()

	def GetResX(self):
		if self.varOverrideResolution.get() == 1:
			return int(self.varXRes.get())
		else:
			return self.gui.data.GetResX()
	
	def GetResY(self):
		if self.varOverrideResolution.get() == 1:
			return int(self.varYRes.get())
		else:
			return self.gui.data.GetResY()
	
	def InitAccelerator(self):
		
		Label(self.frmAccelerator, text="ACCELERATOR", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varAccelerator = StringVar(); self.varAccelerator.set("kdtree")
		Radiobutton(self.frmAccelerator, text="Kdtree", value="kdtree", variable=self.varAccelerator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmAccelerator, text="QBVH", value="qbvh", variable=self.varAccelerator, anchor=W).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("accelerator", "string", self.varAccelerator)

	def InitConfigPanel(self):
		Label(self.frmConfig, text="CONFIGURE", bg="black", fg="white").pack(side=TOP, fill=X)
		self.AddPanel(Frame(self.frmConfig), "none")
		self.SetPanel("none")
	
	def InitExportFiles(self):
		# Export files
		Label(self.frmExportFiles, text="EXPORT FILES", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		Label(self.frmExportFiles, text="Export as separate file:", anchor=W).pack(side=TOP, fill=X)
		self.varFileGeometry = IntVar()
		Checkbutton(self.frmExportFiles, text="Geometry", anchor=W, variable=self.varFileGeometry).pack(side=TOP, fill=X)
		self.varFileMaterials = IntVar()
		Checkbutton(self.frmExportFiles, text="Materials", anchor=W, variable=self.varFileMaterials).pack(side=TOP, fill=X)	
		self.varFileLights = IntVar()
		Checkbutton(self.frmExportFiles, text="Lights", anchor=W, variable=self.varFileLights).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("export_geometry", "int", self.varFileGeometry)
		self.gui.RegisterControl("export_materials", "int", self.varFileMaterials)
		self.gui.RegisterControl("export_lights", "int", self.varFileLights)
	
	def InitExportLocation(self):
		# Export location
		Label(self.frmExport, text="EXPORT LOCATION", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		Label(self.frmExport, text="Export to:").pack(side=LEFT, pady=10)
		self.varExportLocation = StringVar()
		Entry(self.frmExport, width=50, textvariable=self.varExportLocation).pack(side=LEFT)
		Button(self.frmExport, text="Select", command=self.ChooseOutputFile).pack(side=LEFT, padx=5)
		Button(self.frmExport, text=">>", command=self.CopyOutputName).pack(side=LEFT, padx=5)
		Label(self.frmExport, text="Lux name:").pack(side=LEFT)
		self.varLuxName = StringVar()
		Entry(self.frmExport, width=20, textvariable=self.varLuxName).pack(side=LEFT, padx=5)
		self.ExportButton = Button(self.frmExport, text="EXPORT TO LUX", bg="red", fg="white", command=self.Export)
		self.ExportButton.pack(side=RIGHT, padx=5)
		# Register
		self.gui.RegisterControl("export_location", "string", self.varExportLocation)
		self.gui.RegisterControl("luxname", "string", self.varLuxName)

	def InitFilm(self):
		Label(self.frmFilm, text="FILM", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		# gamma
		self.varFilmGamma = DoubleVar(); self.varFilmGamma.set(2.2)
		Scale(self.frmFilm, label="Gamma correction", orient=HORIZONTAL, from_=0.0, to=5.0, resolution=0.1, variable=self.varFilmGamma).pack(side=TOP, fill=X)
		
		# write_resume_flm and restart_resume_flm
		self.varFilmWriteFLM = IntVar()
		Checkbutton(self.frmFilm, text="Write FLM file", variable=self.varFilmWriteFLM, anchor=W).pack(side=TOP, fill=X)
		self.varFilmRestartFLM = IntVar()
		Checkbutton(self.frmFilm, text="Ignore existing FLM file", variable=self.varFilmRestartFLM, anchor=W).pack(side=TOP, fill=X)
		
		# flmwriteinterval
		self.varFilmFLMWriteInterval = IntVar(); self.varFilmFLMWriteInterval.set(60)
		Scale(self.frmFilm, label="FLM write interval", orient=HORIZONTAL, from_=1, to=3600, resolution=10, variable=self.varFilmFLMWriteInterval).pack(side=TOP, fill=X)

		# displayinterval
		self.varFilmDisplayInterval = IntVar(); self.varFilmDisplayInterval.set(12)
		Scale(self.frmFilm, label="Display update interval", orient=HORIZONTAL, from_=1, to=60, resolution=1, variable=self.varFilmDisplayInterval).pack(side=TOP, fill=X)

		# outlierrejection_k
		self.varFilmOutlier = IntVar(); self.varFilmOutlier.set(0)
		Scale(self.frmFilm, label="Firefly killer", orient=HORIZONTAL, from_=0, to=10, resolution=1, variable=self.varFilmOutlier).pack(side=TOP, fill=X) 
		
		self.gui.RegisterControl("film_gc", "float", self.varFilmGamma)
		self.gui.RegisterControl("write_flm", "int", self.varFilmWriteFLM)
		self.gui.RegisterControl("restart_flm", "int", self.varFilmRestartFLM)
		self.gui.RegisterControl("flm_write_interval", "int", self.varFilmFLMWriteInterval)
		self.gui.RegisterControl("display_interval", "int", self.varFilmDisplayInterval)
		self.gui.RegisterControl("firefly_killer", "int", self.varFilmOutlier)
		
	def InitImageOutput(self):
		# Image output
		Label(self.frmImageOutput, text="IMAGE OUTPUT", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varPreMultiplyAlpha = IntVar()
		Checkbutton(self.frmImageOutput, text="Write alpha channel", anchor=W, variable=self.varPreMultiplyAlpha).pack(side=TOP, fill=X)
		self.varWritePNG = IntVar(); self.varWritePNG.set(1)
		Checkbutton(self.frmImageOutput, text="Write PNG image file", anchor=W, variable=self.varWritePNG).pack(side=TOP, fill=X)
		self.varWrite16bitPNG = IntVar()
		Checkbutton(self.frmImageOutput, text="Write 16 bit PNG", anchor=W, variable=self.varWrite16bitPNG).pack(side=TOP, fill=X)
		self.varClamp = IntVar(); self.varClamp.set(1)
		Checkbutton(self.frmImageOutput, text="Clamp out of gamut colours", anchor=W, variable=self.varClamp).pack(side=TOP, fill=X)

		self.frmClampMethod = Frame(self.frmImageOutput)
		self.frmClampMethod.pack(side=TOP, fill=X)
		Label(self.frmClampMethod, text="Clamp method: ").pack(side=LEFT)
		self.varClampMethod = StringVar(); self.varClampMethod.set("lum")
		OptionMenu(self.frmClampMethod, self.varClampMethod, "lum", "hue", "cut").pack(side=LEFT)
		
		self.frmChannels = Frame(self.frmImageOutput)
		self.frmChannels.pack(side=TOP, fill=X)
		Label(self.frmChannels, text="Write channels: ").pack(side=LEFT)
		self.varChannels = StringVar(); self.varChannels.set("RGB")
		OptionMenu(self.frmChannels, self.varChannels, "RGB", "RGBA", "Y", "YA").pack(side=LEFT)
		
		self.WriteInterval = Scale(self.frmImageOutput, orient=HORIZONTAL, from_=5, to=300, resolution=5, label="Image write interval (in seconds)")
		self.WriteInterval.set(60)
		self.WriteInterval.pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("premultiply_alpha", "int", self.varPreMultiplyAlpha)
		self.gui.RegisterControl("write_png", "int", self.varWritePNG)
		self.gui.RegisterControl("write_16bit_png", "int", self.varWrite16bitPNG)
		self.gui.RegisterControl("clamp", "int", self.varClamp)
		self.gui.RegisterControl("clamp_method", "string", self.varClampMethod)
		self.gui.RegisterControl("write_channels", "string", self.varChannels)
		self.gui.RegisterControl("write_interval", "int", self.WriteInterval)

	def InitMiscellaneous(self):
		Label(self.frmMiscellaneous, text="OTHER OPTIONS", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		# Flip HDRI
		self.varFlipHDRI = IntVar(); self.varFlipHDRI.set(1)
		Checkbutton(self.frmMiscellaneous, anchor=W, text="Flip HDRI maps horizontally", variable=self.varFlipHDRI).pack(side=TOP, fill=X, pady=5)
		self.gui.RegisterControl("flip_hdri", "int", self.varFlipHDRI)
		
		# Compute normals
		self.varComputeNormals = IntVar()
		self.varComputeNormals.set(1)
		Checkbutton(self.frmMiscellaneous, anchor=W, text="Compute normals", variable=self.varComputeNormals).pack(side=TOP, fill=X)
		self.gui.RegisterControl("compute_normals", "int", self.varComputeNormals)
		
		# 'Clay' render mode
		self.varClay = IntVar()
		Checkbutton(self.frmMiscellaneous, anchor=W, text="Clay mode render", variable=self.varClay).pack(side=TOP, fill=X)
		self.gui.RegisterControl("clay_mode", "int", self.varClay)
		
		# GC for skydomes and HDRI
		self.IBLGC = LabelEntry(self.frmMiscellaneous, initValue=1.0, text="GC for IBL", type="float", formatstring="%.1f", expand=0)
		self.gui.RegisterControl("ibl_gc", "float", self.IBLGC)
		
	def InitOutputResolution(self):
		# Output resolution
		Label(self.frmResolution, text="OUTPUT RESOLUTION", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varOverrideResolution = IntVar()
		Checkbutton(self.frmResolution, text="Override resolution", anchor=W, variable=self.varOverrideResolution).pack(side=TOP, fill=X)
		
		self.frmResolutionXYRes = Frame(self.frmResolution)
		self.frmResolutionXYRes.pack(side=TOP, fill=X)
		self.varXRes = StringVar();	self.varXRes.set(str(self.gui.data.GetResX()))
		self.varYRes = StringVar(); self.varYRes.set(str(self.gui.data.GetResY()))
		Label(self.frmResolutionXYRes, text="X Res:").pack(side=LEFT)
		Entry(self.frmResolutionXYRes, width=5, textvariable=self.varXRes).pack(side=LEFT, padx=5)
		Label(self.frmResolutionXYRes, text="Y Res:").pack(side=LEFT)
		Entry(self.frmResolutionXYRes, width=5, textvariable=self.varYRes).pack(side=LEFT, padx=5)
		
		self.gui.RegisterControl("override_res", "int", self.varOverrideResolution)
		self.gui.RegisterControl("export_xres", "string", self.varXRes)
		self.gui.RegisterControl("export_yres", "string", self.varYRes)
		
		self.frmResolutionScale = Frame(self.frmResolution)
		self.frmResolutionScale.pack(side=TOP, fill=X, padx=5, pady=5)
		Button(self.frmResolutionScale, text="1/2", command=self.MulHalf).pack(side=LEFT)
		Button(self.frmResolutionScale, text="1/3", command=self.MulThird).pack(side=LEFT)
		Button(self.frmResolutionScale, text="1/4", command=self.MulQuarter).pack(side=LEFT)
		Button(self.frmResolutionScale, text="2x", command=self.MulTwo).pack(side=LEFT)
		Button(self.frmResolutionScale, text="3x", command=self.MulThree).pack(side=LEFT)
		Button(self.frmResolutionScale, text="Reset", command=self.ResetResolution).pack(side=LEFT, padx=5)

	def InitPixelFilter(self):
	
		self.ConfigPixelFilterBox()
		self.ConfigPixelFilterTriangle()
		self.ConfigPixelFilterGaussian()
		self.ConfigPixelFilterMitchell()
		self.ConfigPixelFilterSinc()
	
		Label(self.frmPixelFilter, text="PIXEL FILTER", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varPixelFilter = StringVar(); self.varPixelFilter.set("mitchell")
		Radiobutton(self.frmPixelFilter, text="Box", value="box", variable=self.varPixelFilter, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmPixelFilter, text="Triangle", value="triangle", variable=self.varPixelFilter, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmPixelFilter, text="Gaussian", value="gaussian", variable=self.varPixelFilter, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmPixelFilter, text="Mitchell", value="mitchell", variable=self.varPixelFilter, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmPixelFilter, text="Sinc", value="sinc", variable=self.varPixelFilter, anchor=W).pack(side=TOP, fill=X)
		Button(self.frmPixelFilter, text="Configure", relief=GROOVE, command=self.ConfigurePixelFilter).pack(side=TOP, fill=X, padx=3)
		
		self.gui.RegisterControl("pixel_filter", "string", self.varPixelFilter)
	
	def InitProgressBar(self):
		Label(self.frmProgress, text="Progress:").pack(side=LEFT)
		self.pb = ProgressBarView(self.frmProgress, height=10, width=300, doLabel=0, background='white', fillColor="red", min=0, max=100, value=0)
		self.pb.pack(side=LEFT)
		self.varProgressText = StringVar()
		self.ProgressText = Label(self.frmProgress, textvariable=self.varProgressText, anchor=W)
		self.ProgressText.pack(side=TOP, fill=X, expand=1)
	
	def InitRenderer(self):
		# Renderer
		Label(self.frmRenderer, text="RENDERER", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varRenderer = StringVar(); self.varRenderer.set("sampler")
		Radiobutton(self.frmRenderer, text="Sampler", value="sampler", variable=self.varRenderer, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmRenderer, text="Hybrid", value="hybrid", variable=self.varRenderer, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmRenderer, text="SPPM", value="sppm", variable=self.varRenderer, anchor=W).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("renderer", "string", self.varRenderer)

	def InitSampler(self):
	
		self.ConfigSamplerERPT()
		self.ConfigSamplerLowdiscrepancy()
		self.ConfigSamplerMetropolis()
		self.ConfigSamplerRandom()
	
		Label(self.frmSampler, text="SAMPLER", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varSampler = StringVar(); self.varSampler.set("metropolis")
		Radiobutton(self.frmSampler, text="Random", value="random", variable=self.varSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSampler, text="Low discrepancy", value="lowdiscrepancy", variable=self.varSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSampler, text="Metropolis", value="metropolis", variable=self.varSampler, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSampler, text="ERPT", value="erpt", variable=self.varSampler, anchor=W).pack(side=TOP, fill=X)
		Button(self.frmSampler, text="Configure", relief=GROOVE, command=self.ConfigureSampler).pack(side=TOP, fill=X, padx=3)
		
		self.gui.RegisterControl("sampler", "string", self.varSampler)
	
	def InitScaleColours(self):
		Label(self.frmScaleColours, text="SCALE COLOURS", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varScaleColours = IntVar(); self.varScaleColours.set(1)
		Checkbutton(self.frmScaleColours, text="Scale colours", variable=self.varScaleColours, anchor=W).pack(side=TOP, fill=X)
		
		self.varScaleDiffuse = DoubleVar(); self.varScaleDiffuse.set(0.8)
		Scale(self.frmScaleColours, label="Scale Diffuse", orient=HORIZONTAL, from_=0.1, to=1.0, resolution=0.05, variable=self.varScaleDiffuse).pack(side=TOP, fill=X)
		
		self.varScaleSpecular = DoubleVar(); self.varScaleSpecular.set(0.2)
		Scale(self.frmScaleColours, label="Scale Specular", orient=HORIZONTAL, from_=0.1, to=1.0, resolution=0.05, variable=self.varScaleSpecular).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("scale_colours", "int", self.varScaleColours)
		self.gui.RegisterControl("scale_diffuse", "float", self.varScaleDiffuse)
		self.gui.RegisterControl("scale_specular", "float", self.varScaleSpecular)
	
	def InitSurfaceIntegrator(self):
	
		self.ConfigSurfaceBidirectional()
		self.ConfigSurfacePath()
		self.ConfigSurfaceExphotonmap()
		self.ConfigSurfaceDirectLighting()
		self.ConfigSurfaceDistributedPath()
		self.ConfigSurfaceSPPM()
	
		Label(self.frmSurfaceIntegrator, text="SURFACE INTEGRATOR", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varSurfaceIntegrator = StringVar(); self.varSurfaceIntegrator.set("bidirectional")
		Radiobutton(self.frmSurfaceIntegrator, text="Bidirectional", value="bidirectional", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSurfaceIntegrator, text="Path", value="path", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSurfaceIntegrator, text="Exphotonmap", value="exphotonmap", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSurfaceIntegrator, text="Direct lighting", value="directlighting", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSurfaceIntegrator, text="Distributed path", value="distributedpath", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmSurfaceIntegrator, text="SPPM", value="sppm", variable=self.varSurfaceIntegrator, anchor=W).pack(side=TOP, fill=X)
		Button(self.frmSurfaceIntegrator, text="Configure", relief=GROOVE, command=self.ConfigureSurfaceIntegrator).pack(side=TOP, fill=X, padx=3)
		
		self.gui.RegisterControl("surface_integrator", "string", self.varSurfaceIntegrator)

	def InitVolumeIntegrator(self):
		Label(self.frmVolumeIntegrator, text="VOLUME INTEGRATOR", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varVolumeIntegrator = StringVar(); self.varVolumeIntegrator.set("multi")
		Radiobutton(self.frmVolumeIntegrator, text="Emission", value="emission", variable=self.varVolumeIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmVolumeIntegrator, text="Single", value="single", variable=self.varVolumeIntegrator, anchor=W).pack(side=TOP, fill=X)
		Radiobutton(self.frmVolumeIntegrator, text="Multi", value="multi", variable=self.varVolumeIntegrator, anchor=W).pack(side=TOP, fill=X)
		
		self.gui.RegisterControl("volume_integrator", "string", self.varVolumeIntegrator)
		
	def MulQuarter(self):
		x = self.varXRes.get()
		if x.isdigit():
			x = int(x)/4
			self.varXRes.set(str(x))
			
		y = self.varYRes.get()
		if y.isdigit():
			y = int(y)/4
			self.varYRes.set(str(y))
			self.varOverrideResolution.set(1)

	def MulThird(self):
		x = self.varXRes.get()
		if x.isdigit():
			x = int(x)/3
			self.varXRes.set(str(x))
			
		y = self.varYRes.get()
		if y.isdigit():
			y = int(y)/3
			self.varYRes.set(str(y))
			self.varOverrideResolution.set(1)

	def MulHalf(self):
		x = self.varXRes.get()
		if x.isdigit():
			x = int(x)/2
			self.varXRes.set(str(x))
			
		y = self.varYRes.get()
		if y.isdigit():
			y = int(y)/2
			self.varYRes.set(str(y))
			self.varOverrideResolution.set(1)

	def MulTwo(self):
		x = self.varXRes.get()
		if x.isdigit():
			x = int(x)*2
			self.varXRes.set(str(x))
			
		y = self.varYRes.get()
		if y.isdigit():
			y = int(y)*2
			self.varYRes.set(str(y))
			self.varOverrideResolution.set(1)

	def MulThree(self):
		x = self.varXRes.get()
		if x.isdigit():
			x = int(x)*3
			self.varXRes.set(str(x))
			
		y = self.varYRes.get()
		if y.isdigit():
			y = int(y)*3
			self.varYRes.set(str(y))
			self.varOverrideResolution.set(1)
	
	def ResetResolution(self):
		self.varXRes.set(str(self.gui.data.GetResX()))
		self.varYRes.set(str(self.gui.data.GetResY()))
		self.varOverrideResolution.set(0)
	
	def SetPanel(self, name):
		if self.currentPanel != name:
			if self.currentPanel: self.panels[self.currentPanel].forget()
			self.panels[name].pack(side=TOP, fill=X)
			self.currentPanel = name
	
	def ValidateOutputFile(self):
		files = [None, None, None, None]		# LXS, LXO, LXM, LXL output files
		try:
			s = self.varExportLocation.get()
			files[0] = open(s, "w")
		except:
			tkMessageBox.showwarning("Oops!", "Unable to open output file")
			files = None
			return
		
		if self.varFileGeometry.get() == 1:
			s = os.path.splitext(self.varExportLocation.get())[0] + '.lxo'
			try:
				files[1] = open(s, "w")
			except:
				files[1] = None
		
		if self.varFileMaterials.get() == 1:
			s = os.path.splitext(self.varExportLocation.get())[0] + '.lxm'
			try:
				files[2] = open(s, "w")
			except:
				files[2] = None
		
		if self.varFileLights.get() == 1:
			s = os.path.splitext(self.varExportLocation.get())[0] + '.lxl'
			try:
				files[3] = open(s, "w")
			except:
				files[3] = None
				
		return files
		
	def ValidateResolution(self):
		if self.varOverrideResolution.get() == 1:
			try:
				x = int(self.varXRes.get())
				y = int(self.varYRes.get())
			except:
				tkMessageBox.showwarning("Oops!", "Output resolution is not valid\nUncheck 'Override resolution' option\nOr use valid values")
				return False
		return True
class pageLights:
	def __init__(self, gui, parent):
	
		global titleFont
	
		self.gui = gui
		self.data = gui.GetData()
		self.groups = self.data.LightGroups()
		
		self.views = (
			("Light", "Type", "Group", "Importance"),
			("Light", "Watts", "Gain", "Efficacy", "IES", "Tex"))
		
		self.mainFrame = parent
		self.frmLights = Frame(self.mainFrame)
		self.frmLights.pack(side=TOP, fill=BOTH, expand=YES, padx=20, pady=10)
		
		self.frmColumnA = Frame(self.frmLights)
		self.frmColumnA.pack(side=LEFT, fill=Y)
		self.frmColumnB = Frame(self.frmLights)
		self.frmColumnB.pack(side=LEFT, fill=Y, padx=20)
		
		self.frmLightList = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmLightList.pack(side=TOP, fill=X)
		
		self.frmView = Frame(self.frmColumnA)
		self.frmView.pack(side=TOP, fill=X)
		
		self.frmGroups = Frame(self.frmColumnB)
		self.frmGroups.pack(side=TOP, fill=X, padx=10)
		
		self.frmLightParms = Frame(self.frmColumnB)
		self.frmLightParms.pack(side=TOP, fill=X, pady=10)
		
		self.frmEmitterParams = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmEmitterParams.pack(side=TOP, fill=X)
		
		self.frmMisc = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmMisc.pack(side=TOP, fill=X, pady=10)
		
		self.frmSunSky = Frame(self.frmColumnA)
		self.frmSunSky.pack(side=TOP, fill=X, pady=5)
		
		self.frmSun = Frame(self.frmSunSky, bd=2, relief=RIDGE)
		self.frmSun.pack(side=LEFT, fill=Y)

		self.frmSky = Frame(self.frmSunSky, bd=2, relief=RIDGE)
		self.frmSky.pack(side=RIGHT, fill=Y)
		
		self.InitList()
		self.InitGroups()
		self.InitParameters()
		self.InitSun()
		self.InitSky()
		self.InitMiscellaneous()
		
		self.saveload = {
			"varSunLight" : "string",
			"varSunGain" : "float",
			"varSunSamples" : "integer",
			"varSunTurbidity" : "float",
			"varSunRelSize" : "float",
			"varUseSky" : "integer",
			"varSkyGain" : "float",
			"varSkySamples" : "integer",
			"varSkyTurbidity" : "float",
			"aconst" : "float",
			"bconst" : "float",
			"cconst" : "float",
			"dconst" : "float",
			"econst" : "float"
			}
	
	def AddGroup(self):
		newgroup = self.newgroup.get()
		if newgroup and newgroup not in self.groups:
			self.gui.AddLightGroup(newgroup)
	
	def Assign(self):
		selected = self.LightList.curselection()
		if selected:
			data = self.LightList.get(selected[0])
			key = data[0]
			light = self.lights[key]
			if light.GetTypeByName() in ("Material", "Point"):
				light.colour = self.EmitterCol.get()
				light.wattage = self.Watts.get()
				light.efficacy = self.Efficacy.get()
				light.gain = self.Gain.get()
				light.ies = self.IES.get()
				light.use_texture = self.varUseTexture.get()
				self.gui.Refresh()
				self.btnAssign.config(bg=defaultColour, fg="black")
			else:
				tkMessageBox.showerror("Oops!", "Selected light is not an emitter")			
	
	def ChangeView(self, *dummy):
		self.PopulateList()
		view = self.varView.get()
		self.LightList.Display(self.views[view])		
	
	def Grab(self):
		selected = self.LightList.curselection()
		if selected:
			data = self.LightList.get(selected[0])
			key = data[0]
			light = self.lights[key]
			if light.GetTypeByName() in ("Material", "Point"):
				self.EmitterCol.set(light.colour)
				self.Watts.set(light.wattage)
				self.Efficacy.set(light.efficacy)
				self.Gain.set(light.gain)
				self.IES.set(light.ies or "")
				self.varUseTexture.set(light.use_texture)
				self.btnAssign.config(bg="red", fg="white")
			else:
				tkMessageBox.showerror("Oops!", "Selected light is not an emitter")
	
	def InitEmitterParams(self):
	
		Label(self.frmEmitterParams, text="EMITTER PARAMETERS", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		# Power
		self.Watts = PickScale(self.frmEmitterParams, orient=HORIZONTAL, label="Wattage", from_=1, to=10000, resolution=10)
		
		# Gain
		self.Gain = PickScale(self.frmEmitterParams, type="float", orient=HORIZONTAL, label="Gain", from_=0, to=100, resolution=0.1)
		
		# Efficacy
		self.Efficacy = PickScale(self.frmEmitterParams, orient=HORIZONTAL, label="Efficacy", from_=1, to=10000, resolution=1)
		
		# IES
		self.IES = PickFile(self.frmEmitterParams, label="IES", pady=10, initDir=self.gui.DefaultFolders("iesdata"))
		
		# Use texture map
		self.varUseTexture = IntVar()
		Checkbutton(self.frmEmitterParams, text="Use texture instead of colour", anchor=W, variable=self.varUseTexture).pack(side=TOP, fill=X)

		# Colour
		self.EmitterCol = PickColour(self.frmEmitterParams, width=10, label="Colour")
		self.EmitterCol.pack(side=TOP, fill=X, expand=0, pady=10)
		
		# GRAB and ASSIGN Buttons
		self.frmButtons = Frame(self.frmEmitterParams)
		self.frmButtons.pack(side=TOP, fill=X)
		self.btnGrab = Button(self.frmButtons, text="GRAB", command=self.Grab)
		self.btnGrab.pack(side=LEFT)
		self.btnAssign = Button(self.frmButtons, text="ASSIGN", command=self.Assign)
		self.btnAssign.pack(side=LEFT)

	def InitFinalise(self):
		self.InitEmitterParams()
	
	def InitGroups(self):
	
		global titleFont
	
		Label(self.frmGroups, text="Light groups:", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		
		# Add
		self.frmAdd = Frame(self.frmGroups)
		self.frmAdd.pack(side=TOP, fill=X, pady=10)
		Label(self.frmAdd, text="New group:").pack(side=LEFT)
		self.newgroup = Entry(self.frmAdd); self.newgroup.pack(side=LEFT, padx=5)
		Button(self.frmAdd, text="ADD", command=self.AddGroup).pack(side=LEFT)
		
		# Change
		self.frmChange = Frame(self.frmGroups)
		self.frmChange.pack(side=TOP, fill=X)
		self.frmSelect = Frame(self.frmChange)
		self.frmSelect.pack(side=TOP, fill=X)
		Label(self.frmSelect, text="Select group:").pack(side=LEFT)
		self.varGroup = StringVar(); self.varGroup.set("default")
		self.optGroups = OptionMenu(self.frmSelect, self.varGroup, *self.groups)
		self.optGroups.pack(side=LEFT, padx=5)
		Button(self.frmSelect, text="SET", command=self.SetGroup).pack(side=LEFT)
		
	def InitList(self):
	
		self.lights = self.data.AllLightSources()
		
		Label(self.frmLightList, text="LIGHTS IN SCENE", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		self.LightList = MultiListbox(self.frmLightList, 12, (
			('Light', 20, False), 
			('Type', 15, False), 
			('Group', 15, False), 
			('Importance', 5, False),
			('Watts', 10, None),
			('Gain', 10, None),
			('Efficacy', 8, None),
			('IES', 5, None),
			('Tex', 5, None),
			('Dummy', 1, None)
			))
		self.LightList.pack(side=TOP)
			
		# View options
		self.varView = IntVar()
		self.varView.trace('w', self.ChangeView)
		Label(self.frmView, text="View:").pack(side=LEFT)
		Radiobutton(self.frmView, text="Standard", value=0, variable=self.varView).pack(side=LEFT)
		Radiobutton(self.frmView, text="Emitters", value=1, variable=self.varView).pack(side=LEFT)
		
		self.PopulateList()
		self.ChangeView()
	
	def InitMiscellaneous(self):
		Label(self.frmMisc, text="Other parameters", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.NSamples = PickScale(self.frmMisc,	type="int",	orient=HORIZONTAL, length=None, from_=1, to=5, resolution=1, variable=None, label=None, prefix="NSamples")
		self.NSamples.set(1)
		self.gui.RegisterControl("nsamples", "int", self.NSamples)
	
	def InitParameters(self):
	
		# Importance
		frmImportance = Frame(self.frmLightParms)
		frmImportance.pack(side=TOP, fill=X)
		self.varImportance = DoubleVar(); self.varImportance.set(1.0)
		Scale(frmImportance, label="Importance", orient=HORIZONTAL, from_=0, to=10, resolution=0.1, variable=self.varImportance).pack(side=LEFT, fill=X, expand=1)
		
		# Set
		Button(frmImportance, text="SET", command=self.SetImportance).pack(side=BOTTOM, pady=5)
	
	def InitSky(self):
		Label(self.frmSky, text="SKY", bg="black", fg="white").pack(side=TOP, fill=X)
		
		self.varUseSky = IntVar()
		Checkbutton(self.frmSky, text="Export sky", variable=self.varUseSky, anchor=W).pack(side=TOP, fill=X)
		
		frmCol1 = Frame(self.frmSky)
		frmCol1.pack(side=LEFT, fill=Y)
		frmCol2 = Frame(self.frmSky)
		frmCol2.pack(side=LEFT, fill=Y)
		
		self.varSkyGain = DoubleVar(); self.varSkyGain.set(1.0)
		Scale(frmCol1, label="Gain", orient=HORIZONTAL, from_=0, to=10, resolution=0.1, variable=self.varSkyGain).pack(side=TOP, fill=X)
		
		self.varSkySamples = IntVar(); self.varSkySamples.set(1)
		Scale(frmCol1, label="Shadow samples", orient=HORIZONTAL, from_=1, to=10, resolution=1, variable=self.varSkySamples).pack(side=TOP, fill=X)

		self.varSkyTurbidity = DoubleVar(); self.varSkyTurbidity.set(2.0)
		Scale(frmCol1, label="Sky turbidity", orient=HORIZONTAL, from_=1.0, to=30.0, resolution=0.1, variable=self.varSkyTurbidity).pack(side=TOP, fill=X)	

		self.aconst = Scale(frmCol2, label="Horizon brightness", orient=HORIZONTAL, from_=0, to=2, resolution=0.1, length=120)
		self.aconst.set(0.5)
		self.aconst.pack(side=TOP)
		
		self.bconst = Scale(frmCol2, label="Horizon size", orient=HORIZONTAL, from_=0, to=2, resolution=0.1, length=120)
		self.bconst.set(0.5)
		self.bconst.pack(side=TOP)
		
		self.cconst = Scale(frmCol2, label="Sun halo brightness", orient=HORIZONTAL, from_=0, to=1, resolution=0.1, length=120)
		self.cconst.set(1.0)
		self.cconst.pack(side=TOP)
		
		self.dconst = Scale(frmCol2, label="Sun halo size", orient=HORIZONTAL, from_=0, to=1, resolution=0.1, length=120)
		self.dconst.set(1.0)
		self.dconst.pack(side=TOP)
		
		self.econst = Scale(frmCol2, label="Backscattering", orient=HORIZONTAL, from_=0, to=5, resolution=0.1, length=120)
		self.econst.set(1.0)
		self.econst.pack(side=TOP)		
			
	def InitSun(self):
		Label(self.frmSun, text="SUN", bg="black", fg="white").pack(side=TOP, fill=X)
		
		Label(self.frmSun, anchor=W, text="Set Sun to:").pack(side=TOP, fill=X)
		list = self.data.LightNamesByType("Infinite")
		list.insert(0, "None")
		self.varSunLight = StringVar(); self.varSunLight.set(list[0])
		self.optSun = OptionMenu(self.frmSun, self.varSunLight, *list)
		self.optSun.configure(width=15, anchor=W)
		self.optSun.pack(side=TOP, fill=X)
		
		self.varSunGain = DoubleVar(); self.varSunGain.set(1.0)
		Scale(self.frmSun, label="Gain", orient=HORIZONTAL, from_=0, to=10, resolution=0.1, variable=self.varSunGain).pack(side=TOP, fill=X)
		
		self.varSunSamples = IntVar(); self.varSunSamples.set(1)
		Scale(self.frmSun, label="Shadow samples", orient=HORIZONTAL, from_=1, to=10, resolution=1, variable=self.varSunSamples).pack(side=TOP, fill=X)

		self.varSunTurbidity = DoubleVar(); self.varSunTurbidity.set(2.0)
		Scale(self.frmSun, label="Sun turbidity", orient=HORIZONTAL, from_=1.0, to=30.0, resolution=0.1, variable=self.varSunTurbidity).pack(side=TOP, fill=X)		

		self.varSunRelSize = DoubleVar(); self.varSunRelSize.set(1.0)
		Scale(self.frmSun, label="Sun relative size", orient=HORIZONTAL, from_=0.1, to=10.0, resolution=0.1, variable=self.varSunRelSize).pack(side=TOP, fill=X)	
	
	def Load(self, data):
		for key, value in data.items():
			getattr(self, key).set(Unpack(value, self.saveload[key]))
	
	def PopulateList(self):
	
		selected_lines =  self.LightList.curselection()
		target = self.LightList.nearest(0)
		self.LightList.delete(0, END)
	
		for v in self.lights.values():
		
			if self.varView.get() == 1 and not v.is_emitter: continue
		
			self.LightList.insert(END, 
				(v.NameAlt(), 
				v.GetTypeByName(), 
				v.group, 
				v.Importance(),
				v.wattage,
				v.gain,
				v.efficacy,
				YN(v.ies),
				YN(v.use_texture),
				"dummy"))		

		for line in selected_lines:
			self.LightList.selection_set(line)
			
		self.LightList.yview_scroll(target, UNITS)
	
	def Save(self):
		attribs = {}
		for key, value in self.saveload.items():
			attribs[key] = str(getattr(self, key).get())
		return attribs
		
	def SetGroup(self):
		widget = self.LightList
		old = widget.curselection()
		if old:
			data = widget.get(old[0])
			newdata = data[0:2] + [self.varGroup.get()] + data[3:]
			widget.delete(old[0])
			widget.insert(old[0],newdata)
			widget.selection_toggle(old[0])
			light = self.lights[data[0]]
			light.SetGroup(self.varGroup.get())
		
	def SetImportance(self):
		widget = self.LightList
		old = widget.curselection()
		if old:
			data = widget.get(old[0])
			light = self.lights[data[0]]
			light.SetImportance(self.varImportance.get())
			self.PopulateList()

	def Sky(self):
		return (False, True)[self.varUseSky.get()]
			
	def Sun(self):
		return self.varSunLight.get()
class pageMaterials:
	def __init__(self, parent, gui):
		
		self.current_grab = None
		self.gui = gui
		self.data = self.gui.GetData()
		self.libs = self.gui.LibraryManager()
		self.materials = self.data.ListOfMaterials()
		
		self.views = (
			("Owner", "Material", "Lux material", "Emit"),
			("Owner", "Material", "Watts", "Gain", "Efficacy", "IES", "Tex", "Group"),
			("Owner", "Material", "SubD", "Lux SubD", "CC SubD", "Disp SubD", "Micro"),
			("Owner", "Material", "Micro", "Base", "2-Sided", "Distance"))
		
		# Frame definitions
		self.mainFrame = Frame(parent)
		Label(self.mainFrame, text="MATERIALS", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)
		
		self.frmColumnA = Frame(self.mainFrame)
		self.frmColumnA.pack(side=LEFT, fill=Y)
		self.frmColumnB = Frame(self.mainFrame)
		self.frmColumnB.pack(side=LEFT, fill=Y, padx=20)
		self.frmColumnC = Frame(self.mainFrame)
		self.frmColumnC.pack(side=LEFT, fill=Y)
		
		# Poser materials
		self.frmMatList = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmMatList.pack(side=TOP, fill=X)
		self.frmView = Frame(self.frmColumnA)
		self.frmView.pack(side=TOP, fill=X)
		self.frmFilter = Frame(self.frmColumnA)
		self.frmFilter.pack(side=TOP, fill=X, pady=10)
		
		# Lux materials
		self.frmLuxLibs = Frame(self.frmColumnB, bd=2, relief=RIDGE)
		self.frmLuxLibs.pack(side=TOP, fill=X)
		
		# Filters
		Label(self.frmFilter, text="FILTERS", bg="black", fg="white").pack(fill=X)	
		self.frmFilterA = Frame(self.frmFilter)
		self.frmFilterA.pack(side=LEFT, fill=Y)
		self.frmFilterB = Frame(self.frmFilter)
		self.frmFilterB.pack(side=LEFT, fill=Y)
		self.frmFilterC = Frame(self.frmFilter)
		self.frmFilterC.pack(side=LEFT, fill=Y)
		
		# More Lux features
		self.frmMore = Frame(self.frmColumnB)
		self.frmMore.pack(side=TOP, fill=X, pady=10)
		self.frmEmitter = Frame(self.frmMore)
		self.frmEmitter.pack(side=TOP, fill=X)
		self.frmSubD = Frame(self.frmMore)
		self.frmSubD.pack(side=TOP, fill=X)
		self.frmDisplacement = Frame(self.frmMore)
		self.frmDisplacement.pack(side=TOP, fill=X)
		self.frmBackface = Frame(self.frmMore)
		self.frmBackface.pack(side=TOP, fill=X)
		self.frmPortal = Frame(self.frmMore)
		self.frmPortal.pack(side=TOP, fill=X)
		self.frmCustom = Frame(self.frmMore)
		self.frmCustom.pack(side=TOP, fill=X)
		self.frmButtons = Frame(self.frmMore)
		self.frmButtons.pack(side=TOP, fill=X, pady=5)
		self.frmAuto = Frame(self.frmMore)
		self.frmAuto.pack(side=TOP, fill=X)
		
		# Configuration panel
		self.frmConfigure = Frame(self.frmColumnC, bd=2, relief=RIDGE)
		#self.frmConfigure = Frame(self.frmColumnC, bd=2, relief=RIDGE, width=250)
		self.frmConfigure.pack(side=LEFT, fill=BOTH)
		#self.frmConfigure.pack_propagate(False)
		Label(self.frmConfigure, text="CONFIGURE", bg="black", fg="white").pack(side=TOP, fill=X)
		
		# Initialise all panels
		self.InitFilters()
		self.InitMaterialList()
		self.InitLuxLibraries()
		self.InitMoreLuxFeatures()
		
		# Initialise configuration panel
		self.notebook = Notebook(self.frmConfigure, TOP)
	
	def Assign(self):
		
		# Ensure there is a Lux material selected
		lux_selected = self.luxlist.curselection()
		if lux_selected:
		
			mat_selected = self.matlist.curselection()
			materials = self.data.Materials()

			for s in mat_selected:
			
				# Get the Material object
				row = self.matlist.get(s)
				key = MakeKey(row[0], row[1])
				mat = materials[key]
				
				# ... and update with the Lux material
				luxname = self.luxlist.get(lux_selected)
				mat.AssignMaterial(MakeKey(self.varLibs.get(),luxname))
				
			# Update table
			self.PopulateMaterialList()
	
	def AssignMoreStuff(self, key=None):
		
		selected = self.matlist.curselection()

		for s in selected:
		
			# Get key
			data = self.matlist.get(s)
			key = MakeKey(data[0], data[1])
			m = self.GetMaterial(key)
			
			# Emitters
			if self.varUseEmitter.get():
				m.is_emitter = self.varMakeEmitter.get()
				m.wattage = self.Watts.get()
				m.gain = self.Gain.get()
				m.efficacy = self.Efficacy.get()
				m.ies = self.IES.get()
				m.use_texture = self.varUseTexture.get()
				m.group = self.varGroup.get()
				m.colour = self.EmitterCol.get()
			
			# Subdivision
			if self.varUseSubD.get():
				m.subdivide = self.varMakeSubD.get()
				m.lux_levels = self.LuxSubD.get()
				m.smooth_normals = self.varSmoothNormals.get()
				m.sharp_boundary = self.varSharpBoundary.get()
				m.create_edges = self.varCreateEdges.get()
				m.disp_levels = self.DisplacementSubD.get()
				
				# Catmull-Clark subdivision is applied at the actor level ...
				m.owner.Subdivide(self.CCSubD.get())
			
			# Displacement
			if self.varUseDisplacement.get():
				m.displacement = self.varMakeDisplacement.get()
				m.use_micro = self.varUseMicroSubD.get()
				m.disp_base = self.DispBase.get()
			
			# Backface
			if self.varUseBackface.get():
				m.double_sided = self.varMakeBackface.get()
				m.distance = float(self.Backface.get())
			
			# Portal
			if self.varUsePortal.get():
				m.portal = self.varMakePortal.get()
				m.doInfinite = self.varPortalInfinite.get()
				m.doSun = self.varPortalSun.get()
				m.doSky = self.varPortalSky.get()
				
			# Custom material
			if self.varUseCustom.get():
				m.use_custom = self.varMakeCustom.get()
				if self.varMakeCustom.get():
					newdef = self.CustomMat.AddToLibrary(m.Key())
					self.libs.AddMaterial(newdef)
					
		self.UpdateAssign(False)
		self.gui.Refresh()
	
	def AutoAssign(self):
		self.data.AutoAssignMaterials(self.varDefaultSkin.get(), self.varPreview.get())
		self.PopulateMaterialList()
	
	def CallbackPreselect(self, row):
		newgrab = self.matlist.get(row)
		newkey = MakeKey(newgrab[0], newgrab[1])
		oldkey = self.current_grab
		
		if (newkey != oldkey) and self.varAutoAssign.get() == 1:
			self.AssignMoreStuff()
		
	def CallbackPostselect(self, row):
		if self.varAutoGrab.get() == 1: self.Grab()
	
	def ChangeView(self, *dummy):
		
		view = self.varView.get()
		self.matlist.Display(self.views[view])
		
	def ConfigBackface(self):
		self.notebook.DisplayByName("2-sided")
	
	def ConfigCustom(self):
		self.notebook.DisplayByName("Custom")
	
	def ConfigDisplacement(self):
		self.notebook.DisplayByName("Displace")
	
	def ConfigEmitter(self):
		self.notebook.DisplayByName("Emitter")
		
	def ConfigSubD(self):
		self.notebook.DisplayByName("SubD")
	
	def GetMaterial(self, key):
		return self.data.Material(key)
	
	def ConfigPortal(self):
		self.notebook.DisplayByName("Portal")
	
	def Grab(self):
		selected = self.matlist.curselection()
		if selected:
		
			# Get key
			data = self.matlist.get(selected[0])
			key = MakeKey(data[0], data[1])
			self.current_grab = key
			m = self.GetMaterial(key)
			
			# Emitters
			if self.varUseEmitter.get():
				self.varMakeEmitter.set(m.is_emitter)
				self.Watts.set(m.wattage)
				self.Gain.set(m.gain)
				self.Efficacy.set(m.efficacy)
				self.IES.set(m.ies or "")
				self.varUseTexture.set(m.use_texture)
				self.varGroup.set(m.group)
				self.EmitterCol.set(m.colour)
			
			# Subdivision
			if self.varUseSubD.get():
				self.varMakeSubD.set(m.subdivide)
				self.LuxSubD.set(m.lux_levels)
				self.varSmoothNormals.set(m.smooth_normals)
				self.varSharpBoundary.set(m.sharp_boundary)
				self.varCreateEdges.set(m.create_edges)
				self.CCSubD.set(m.cc_levels)
				self.DisplacementSubD.set(m.disp_levels)
			
			# Displacement
			if self.varUseDisplacement.get():
				self.varMakeDisplacement.set(m.displacement)
				self.varUseMicroSubD.set(m.use_micro)
				self.DispBase.set(m.disp_base)
			
			# Backface
			if self.varUseBackface.get():
				self.varMakeBackface.set(m.double_sided)
				self.Backface.set(m.distance)
			
			# Portal
			if self.varUsePortal.get():
				self.varMakePortal.set(m.portal)
				self.varPortalInfinite.set(m.doInfinite)
				self.varPortalSun.set(m.doSun)
				self.varPortalSky.set(m.doSky)
				
			# Custom material
			if self.varUseCustom.get():
				self.varMakeCustom.set(m.use_custom)
				self.CustomMat.Reset()
				if m.use_custom:
					matdef = self.libs.GetMaterial(MakeKey("CUSTOM", m.Key()))
					self.CustomMat.LoadData(matdef)
			
			self.UpdateAssign(False)
	
	def InitConfigBackface(self, nb):
	
		self.frmCfgBackface = Frame(nb())
		
		self.Backface = LabelEntry(self.frmCfgBackface, text="Distance", initValue="0.0", type="float", callback=self.UpdateAssignTrue)
		
		nb.AddScreen(self.frmCfgBackface, "2-sided")
	
	def InitConfigCustom(self, nb):
	
		self.frmCfgCustom = Frame(nb())
		self.CustomMat = CustomMaterial(self.frmCfgCustom, self.libs, self.UpdateAssignTrue)
		self.UpdateAssign(False)
		nb.AddScreen(self.frmCfgCustom, "Custom")
	
	def InitConfigDisplacement(self, nb):
	
		self.frmCfgDisp = Frame(nb())
		
		self.varUseMicroSubD = IntVar()
		self.varUseMicroSubD.set(1)
		self.varUseMicroSubD.trace('w', self.UpdateAssignTrue)
		Checkbutton(self.frmCfgDisp, text="Use microdisplacement subdivision", variable=self.varUseMicroSubD, anchor=W).pack(side=TOP, fill=X)
		
		self.DispBase = PickScale(self.frmCfgDisp, from_=0, to=1, resolution=0.01, type="float", label="Displacement base", callback=self.UpdateAssignTrue)
		
		nb.AddScreen(self.frmCfgDisp, "Displace")
	
	def InitConfigEmitter(self, nb):
		
		self.frmCfgEmitter = Frame(nb())
		
		# Power
		self.Watts = PickScale(self.frmCfgEmitter, orient=HORIZONTAL, label="Wattage", from_=1, to=10000, resolution=10, callback=self.UpdateAssignTrue)
		
		# Gain
		self.Gain = PickScale(self.frmCfgEmitter, type="float", orient=HORIZONTAL, label="Gain", from_=0, to=100, resolution=0.1, callback=self.UpdateAssignTrue)
		
		# Efficacy
		self.Efficacy = PickScale(self.frmCfgEmitter, orient=HORIZONTAL, label="Efficacy", from_=0, to=10000, resolution=1, callback=self.UpdateAssignTrue)
		
		# IES
		self.IES = PickFile(self.frmCfgEmitter, label="IES", pady=10, initDir=self.gui.DefaultFolders("iesdata"), callback=self.UpdateAssignTrue)
		
		# Use texture map
		self.varUseTexture = IntVar()
		self.varUseTexture.trace('w', self.UpdateAssignTrue)
		Checkbutton(self.frmCfgEmitter, text="Use texture instead of colour", anchor=W, variable=self.varUseTexture).pack(side=TOP, fill=X)
		
		# Group
		l = self.gui.pageExposure.pageLights
		self.varGroup = StringVar(); self.varGroup.set("Emitters")
		self.Group = LabelledMenu(self.frmCfgEmitter, self.varGroup, l.groups, label="Assign to group", callback=self.UpdateAssignTrue)
		
		# Colour
		self.EmitterCol = PickColour(self.frmCfgEmitter, width=10, label="Colour", callback=self.UpdateAssignTrue)
		self.EmitterCol.pack(side=TOP, fill=X, expand=0, pady=10)
		
		nb.AddScreen(self.frmCfgEmitter, "Emitter")
		
	def InitConfigPortal(self, nb):
		
		self.frmCfgPortal = Frame(nb())
		
		self.varPortalInfinite = IntVar()
		self.varPortalInfinite.trace('w', self.UpdateAssignTrue)
		self.varPortalSun = IntVar()
		self.varPortalSun.trace('w', self.UpdateAssignTrue)
		self.varPortalSky = IntVar()
		self.varPortalSky.trace('w', self.UpdateAssignTrue)
		
		Label(self.frmCfgPortal, text="Add portal to:", anchor=W).pack(side=TOP, fill=X)
		
		Checkbutton(self.frmCfgPortal, anchor=W, text="All Diffuse IBL lights", variable=self.varPortalInfinite).pack(side=TOP, fill=X)
		Checkbutton(self.frmCfgPortal, anchor=W, text="The Sun", variable=self.varPortalSun).pack(side=TOP, fill=X)
		Checkbutton(self.frmCfgPortal, anchor=W, text="The Sky", variable=self.varPortalSky).pack(side=TOP, fill=X)
		
		nb.AddScreen(self.frmCfgPortal, "Portal")
	
	def InitConfigSubD(self, nb):
	
		global titleFont
	
		self.frmCfgSubD = Frame(nb())
		
		Label(self.frmCfgSubD, text="Luxrender subdivision:", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		
		# Luxrender subdivision levels
		self.LuxSubD = PickScale(self.frmCfgSubD, orient=HORIZONTAL, label="Luxrender subdivision levels", from_=0, to=5, resolution=1, callback=self.UpdateAssignTrue)
		
		# Smooth normals
		self.varSmoothNormals = IntVar()
		self.varSmoothNormals.set(1)
		self.varSmoothNormals.trace('w', self.UpdateAssignTrue)
		Checkbutton(self.frmCfgSubD, text="Smooth normals", variable=self.varSmoothNormals, anchor=W).pack(side=TOP, fill=X)
		
		# Sharp boundary
		self.varSharpBoundary = IntVar()
		self.varSharpBoundary.trace('w', self.UpdateAssignTrue)
		Checkbutton(self.frmCfgSubD, text="Keep sharp boundaries", variable=self.varSharpBoundary, anchor=W).pack(side=TOP, fill=X)
		
		# Create edges at shading normal splits
		self.varCreateEdges = IntVar()
		self.varCreateEdges.trace('w', self.UpdateAssignTrue)
		Checkbutton(self.frmCfgSubD, text="Create edges at shading normal splits", variable=self.varCreateEdges, anchor=W).pack(side=TOP, fill=X)
		
		Label(self.frmCfgSubD, text="Catmull-Clark subdivision:", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		Message(self.frmCfgSubD, text="Catmull-Clark subdivision is applied at the Owner level rather than the Material level, so if you select this option for a material it will affect all Materials with the same owner. Also, to switch it off you must the CC levels to ZERO - unticking the subdivide option is not enough!", aspect=300).pack(side=TOP)
		
		# Catmull-Clark subdivision levels
		self.CCSubD = PickScale(self.frmCfgSubD, orient=HORIZONTAL, label="Catmull-Clark subdivision levels", from_=0, to=5, resolution=1, callback=self.UpdateAssignTrue)
		
		Label(self.frmCfgSubD, text="NB:", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		Message(self.frmCfgSubD, text="If a displacement map is present then the 'Displacement SubD levels' setting overrides all other subdivision settings", aspect=300).pack(side=TOP)
		
		# Displacement subdivision
		self.DisplacementSubD = PickScale(self.frmCfgSubD, orient=HORIZONTAL, label="Displacement SubD levels", from_=0, to=20, resolution=1, initval=1, callback=self.UpdateAssignTrue)
	
		nb.AddScreen(self.frmCfgSubD, "SubD")
		
	def InitFilters(self):
				
		self.varSelOwnerOnly = IntVar()
		self.chkSelOwnerOnly = Checkbutton(self.frmFilterA, anchor=W, text="Show only selected owner", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varSelOwnerOnly)
		self.chkSelOwnerOnly.pack(fill=X)
		self.varTransparent = IntVar()
		self.chkTransparent = Checkbutton(self.frmFilterA, anchor=W, text="Has transparency", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varTransparent)
		self.chkTransparent.pack(fill=X)
		self.varReflective = IntVar()
		self.chkReflective = Checkbutton(self.frmFilterA, anchor=W, text="Is reflective", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varReflective)
		self.chkReflective.pack(fill=X)
		self.varRefractive = IntVar()
		self.chkRefractive = Checkbutton(self.frmFilterA, anchor=W, text="Is refractive", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varRefractive)
		self.chkRefractive.pack(fill=X)
		self.varSpecular = IntVar()
		self.chkSpecular = Checkbutton(self.frmFilterA, anchor=W, text="Has Specular", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varSpecular)
		self.chkSpecular.pack(fill=X)
		self.varLuxmat = IntVar()
		self.chkLuxmat = Checkbutton(self.frmFilterB, anchor=W, text="No Lux material", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varLuxmat)
		self.chkLuxmat.pack(fill=X)
		self.varSkin = IntVar()
		self.chkSkin = Checkbutton(self.frmFilterB, anchor=W, text="Is skin", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varSkin)
		self.chkSkin.pack(fill=X)
		self.varPreview = IntVar()
		self.varPreview.set(1)
		self.chkPreview = Checkbutton(self.frmFilterB, anchor=W, text="Ignore Preview material", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varPreview)
		self.chkPreview.pack(fill=X)
		self.varIsEmitter = IntVar()
		self.chkIsEmitter = Checkbutton(self.frmFilterB, anchor=W, text="Is an emitter", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varIsEmitter)
		self.chkIsEmitter.pack(fill=X)
		self.varSubdivided = IntVar()
		self.chkSubdivided = Checkbutton(self.frmFilterB, anchor=W, text="Is subdivided", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varSubdivided)
		self.chkSubdivided.pack(fill=X)
		self.varIsPortal = IntVar()
		self.chkIsPortal = Checkbutton(self.frmFilterC, anchor=W, text="Is a portal", onvalue=1, offvalue=0, command=self.PopulateMaterialList, variable=self.varIsPortal)
		self.chkIsPortal.pack(fill=X)
	
	def InitFinalise(self):
		self.InitConfigEmitter(self.notebook)
		self.InitConfigSubD(self.notebook)
		self.InitConfigDisplacement(self.notebook)
		self.InitConfigBackface(self.notebook)
		self.InitConfigPortal(self.notebook)
		self.InitConfigCustom(self.notebook)
	
	def InitLuxLibraries(self):
	
		global POSER_VERSION
	
		Label(self.frmLuxLibs, text="LUX LIBRARIES", bg="black", fg="white").pack(side=TOP, fill=X)
		frmSelect = Frame(self.frmLuxLibs)
		frmSelect.pack(side=TOP, fill=X)
		Label(frmSelect, text="Select:").pack(side=LEFT)
		
		libraries = self.libs.GetLibraries()
		
		# The library menu
		self.varLibs = StringVar()
		self.varLibs.set(libraries[0])
		self.varLibs.trace('w', self.UpdateLuxMats)	
		self.optLibs = OptionMenu(frmSelect, self.varLibs, *libraries)
		self.optLibs.pack(side=LEFT, padx=5)

		# The Listbox
		self.lb = Frame(self.frmLuxLibs)
		self.lb.pack(side=TOP, fill=X)
		self.scroll2 = Scrollbar(self.lb, orient=VERTICAL)
		if POSER_VERSION < 7:
			self.luxlist = Listbox(self.lb, height=20, width=30, yscrollcommand=self.scroll2.set)
		else:
			self.luxlist = Listbox(self.lb, height=23, width=30, yscrollcommand=self.scroll2.set)
		
		# Populate with materials
		self.UpdateLuxMats(None, None, None)
		
		self.scroll2["command"] = self.luxlist.yview
		self.luxlist.pack(side=LEFT)
		self.scroll2.pack(side=LEFT, fill=Y)
		
		# Buttons
		frmButtons = Frame(self.frmLuxLibs)
		frmButtons.pack(side=TOP, fill=X, padx=5, pady=5)
		Button(frmButtons, text="Assign", command=self.Assign).pack(side=LEFT)
		Button(frmButtons, text="Reset", command=self.Reset).pack(side=LEFT)
		Button(frmButtons, text="Auto assign", command=self.AutoAssign).pack(side=LEFT)
		Button(frmButtons, text="Skin", command=self.SetSkin).pack(side=LEFT)
		
		# Set default skin
		frmSkin = Frame(self.frmLuxLibs)
		frmSkin.pack(side=TOP, fill=X, padx=5)
		self.varDefaultSkin = StringVar()
		self.varDefaultSkin.set("People/Basic Skin")
		Label(frmSkin, textvariable=self.varDefaultSkin).pack(side=TOP, fill=X)
		
	def InitMaterialList(self):
		
		global POSER_VERSION
		Label(self.frmMatList, text="LIST OF MATERIALS", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		padding = (3, 6)[POSER_VERSION == 6]
		
		# The table
		self.matlist = MultiListbox(self.frmMatList, 28, (
			('Owner', 20, self.SortByOwner),
			('Material', 20, self.SortByMaterial),
			('Lux material', 30, self.SortByLuxmat),
			('Emit', 4, None),
			('Watts', 6, None),
			('Gain', 6, None),
			('Efficacy', 6, None),
			('IES', 3, None),
			('Tex', 3, None),
			('Group', 10, None),
			('SubD', 4, None),
			('Lux SubD', 4, None),
			('CC SubD', 4, None),
			('Disp SubD', 5, None),
			('Micro', 5, None),
			('Base', 5, None),
			('2-Sided', 5, None),
			('Distance', 10, None),
			('Dummy', 1, None)
			), multiselect=True, labelpad=padding, preselect=self.CallbackPreselect, postselect=self.CallbackPostselect)
		self.matlist.pack(side=TOP)
	
		# View options
		self.varView = IntVar()
		self.varView.trace('w', self.ChangeView)
		Label(self.frmView, text="View:").pack(side=LEFT)
		Radiobutton(self.frmView, text="Standard", value=0, variable=self.varView).pack(side=LEFT)
		Radiobutton(self.frmView, text="Emitters", value=1, variable=self.varView).pack(side=LEFT)
		Radiobutton(self.frmView, text="Subdivision", value=2, variable=self.varView).pack(side=LEFT)
		Radiobutton(self.frmView, text="Other", value=3, variable=self.varView).pack(side=LEFT)
	
		# Populate the list
		self.SortByOwner()	
		self.ChangeView()

	def InitMoreLuxFeatures(self):
	
		# Emitter
		self.varUseEmitter = IntVar()
		Checkbutton(self.frmEmitter, text="USE", indicatoron=0, selectcolor="red", command=self.UseEmitter, variable=self.varUseEmitter).pack(side=LEFT)
		self.varMakeEmitter = IntVar()
		self.varMakeEmitter.trace('w', self.UpdateAssignTrue)
		self.chkUseEmitter = Checkbutton(self.frmEmitter, text="Is an emitter", variable=self.varMakeEmitter, state=DISABLED)
		self.chkUseEmitter.pack(side=LEFT, fill=Y)
		self.btnUseEmitter = Button(self.frmEmitter, text="Config", state=DISABLED, command=self.ConfigEmitter)
		self.btnUseEmitter.pack(side=RIGHT)
		
		# Subdivision
		self.varUseSubD = IntVar()
		Checkbutton(self.frmSubD, text="USE", indicatoron=0, selectcolor="red", command=self.UseSubD, variable=self.varUseSubD).pack(side=LEFT)
		self.varMakeSubD = IntVar()
		self.varMakeSubD.trace('w', self.UpdateAssignTrue)
		self.chkUseSubD = Checkbutton(self.frmSubD, text="Subdivide", variable=self.varMakeSubD, state=DISABLED)
		self.chkUseSubD.pack(side=LEFT, fill=Y)
		self.btnUseSubD = Button(self.frmSubD, text="Config", state=DISABLED, command=self.ConfigSubD)
		self.btnUseSubD.pack(side=RIGHT)

		# Displacement
		self.varUseDisplacement = IntVar()
		Checkbutton(self.frmDisplacement, text="USE", indicatoron=0, selectcolor="red", command=self.UseDisplacement, variable=self.varUseDisplacement).pack(side=LEFT)
		self.varMakeDisplacement = IntVar()
		self.varMakeDisplacement.trace('w', self.UpdateAssignTrue)
		self.chkUseDisplacement = Checkbutton(self.frmDisplacement, text="Uses displacement", variable=self.varMakeDisplacement, state=DISABLED)
		self.chkUseDisplacement.pack(side=LEFT, fill=Y)
		self.btnUseDisplacement = Button(self.frmDisplacement, text="Config", state=DISABLED, command=self.ConfigDisplacement)
		self.btnUseDisplacement.pack(side=RIGHT)		
		
		# Backface geometry
		self.varUseBackface = IntVar()
		Checkbutton(self.frmBackface, text="USE", indicatoron=0, selectcolor="red", command=self.UseBackface, variable=self.varUseBackface).pack(side=LEFT)
		self.varMakeBackface = IntVar()
		self.varMakeBackface.trace('w', self.UpdateAssignTrue)
		self.chkUseBackface = Checkbutton(self.frmBackface, text="Make double-sided", variable=self.varMakeBackface, state=DISABLED)
		self.chkUseBackface.pack(side=LEFT, fill=Y)
		self.btnUseBackface = Button(self.frmBackface, text="Config", state=DISABLED, command=self.ConfigBackface)
		self.btnUseBackface.pack(side=RIGHT)
		
		# Portal
		self.varUsePortal = IntVar()
		Checkbutton(self.frmPortal, text="USE", indicatoron=0, selectcolor="red", command=self.UsePortal, variable=self.varUsePortal).pack(side=LEFT)
		self.varMakePortal = IntVar()
		self.varMakePortal.trace('w', self.UpdateAssignTrue)
		self.chkUsePortal = Checkbutton(self.frmPortal, text="Is a portal", variable=self.varMakePortal, state=DISABLED)
		self.chkUsePortal.pack(side=LEFT, fill=Y)
		self.btnUsePortal = Button(self.frmPortal, text="Config", state=DISABLED, command=self.ConfigPortal)
		self.btnUsePortal.pack(side=RIGHT)
		
		# Custom material
		self.varUseCustom = IntVar()
		Checkbutton(self.frmCustom, text="USE", indicatoron=0, selectcolor="red", command=self.UseCustom, variable=self.varUseCustom).pack(side=LEFT)
		self.varMakeCustom = IntVar()
		self.varMakeCustom.trace('w', self.UpdateAssignTrue)
		self.chkUseCustom = Checkbutton(self.frmCustom, text="Use custom material", variable=self.varMakeCustom, state=DISABLED)
		self.chkUseCustom.pack(side=LEFT, fill=Y)
		self.btnUseCustom = Button(self.frmCustom, text="Config", state=DISABLED, command=self.ConfigCustom)
		self.btnUseCustom.pack(side=RIGHT)
		
		# ASSIGN and GRAB buttons
		Button(self.frmButtons, text="Grab", command=self.Grab).pack(side=LEFT)
		self.AssignOpts = Button(self.frmButtons, text="Assign", command=self.AssignMoreStuff)
		self.AssignOpts.pack(side=LEFT)
		Button(self.frmButtons, text="Use all", command=self.UseAll).pack(side=LEFT)
		Button(self.frmButtons, text="Use none", command=self.UseNone).pack(side=LEFT)
		Button(self.frmButtons, text="Get used", command=self.UseTicked).pack(side=LEFT)
		
		# Auto Assign and grab
		self.varAutoGrab = IntVar()
		Checkbutton(self.frmAuto, text="Auto grab", variable=self.varAutoGrab).pack(side=LEFT)
		self.gui.RegisterControl("autograb", "int", self.varAutoGrab)
		self.varAutoAssign = IntVar()
		Checkbutton(self.frmAuto, text="Auto assign", variable=self.varAutoAssign).pack(side=LEFT)
		self.gui.RegisterControl("autoassign", "int", self.varAutoAssign)
		
	def PopulateMaterialList(self):
	
		selected_lines =  self.matlist.curselection()
		selected = self.matlist.getCurrent()
		target = self.matlist.nearest(0)
			
		self.matlist.delete(0, END)
		for m in self.materials:
		
			if self.varSelOwnerOnly.get() == 1 and (m.Owner().Name() != selected[0]): continue
			if self.varTransparent.get() == 1 and not m.IsTransparent(): continue
			if self.varReflective.get() == 1 and not m.IsReflective(): continue
			if self.varRefractive.get() == 1 and not m.IsRefractive(): continue
			if self.varSpecular.get() == 1 and not m.HasSpecular(): continue
			if self.varLuxmat.get() == 1 and not m.luxname == None: continue
			if self.varSkin.get() == 1 and m.IsSkin() == False: continue
			if self.varPreview.get() == 1 and m.IsPreview(): continue
			if self.varIsEmitter.get() == 1 and not m.IsEmitter(): continue
			if self.varSubdivided.get() == 1 and not m.subdivide: continue
			if self.varIsPortal.get() == 1 and not m.IsPortal(): continue

			if m.use_custom:
				luxname = "** Custom **"
			else:
				luxname = m.Luxname() or "None"
		
			self.matlist.insert(END,( 
				m.Owner().Name(),
				m.Name(),
				luxname,
				YN(m.IsEmitter()),
				m.wattage, 
				m.gain,
				m.efficacy,
				YN(m.ies),
				YN(m.use_texture),
				m.group,
				YN(m.subdivide),
				m.lux_levels,
				m.cc_levels, 
				m.disp_levels,
				YN(m.use_micro),
				m.disp_base,
				YN(m.double_sided),
				m.distance,
				"Dummy"
				))
				
		for line in selected_lines:
			self.matlist.selection_set(line)
			
		self.matlist.yview_scroll(target, UNITS)
	
	def Reset(self):
	
		mat_selected = self.matlist.curselection()
		materials = self.data.Materials()
		
		for s in mat_selected:
			row = self.matlist.get(s)
			key = MakeKey(row[0], row[1])
			mat = materials[key]
			
			mat.AssignMaterial(None)

		self.PopulateMaterialList()
		
	def SetSkin(self):
		# Set the default skin used by the Auto assign routine
		m = self.luxlist.get(ACTIVE)
		if m: self.varDefaultSkin.set('/'.join((self.varLibs.get(), m)))		
	
	def SortByLuxmat(self):
		global POSER_VERSION
		if POSER_VERSION == 6:
			decorated = [(material.luxname, i, material) for i, material in enumerate(self.materials)]
			decorated.sort()
			self.materials = [material for luxname, i, material in decorated]
		else:
			self.materials.sort(key=operator.attrgetter("luxname"))
		self.PopulateMaterialList()
		
	def SortByMaterial(self):
		global POSER_VERSION
		if POSER_VERSION == 6:
			decorated = [(material.sortByName, i, material) for i, material in enumerate(self.materials)]
			decorated.sort()
			self.materials = [material for sortByName, i, material in decorated]
		else:
			self.materials.sort(key=operator.attrgetter("sortByName"))
		self.PopulateMaterialList()
		
	def SortByOwner(self):
		global POSER_VERSION
		if POSER_VERSION == 6:
			decorated = [(material.sortByOwner, i, material) for i, material in enumerate(self.materials)]
			decorated.sort()
			self.materials = [material for sortByOwner, i, material in decorated]
		else:
			self.materials.sort(key=operator.attrgetter("sortByOwner"))
		self.PopulateMaterialList()
 
	def UpdateAssignTrue(self, *dummy):
		self.UpdateAssign(True)
		
	def UpdateAssign(self, changed):
		global defaultColour
		if changed == True:
			self.AssignOpts.config(bg="red", fg="white")
		else:
			self.AssignOpts.config(bg=defaultColour, fg="black")			
		
	def UpdateLuxMats(self, a, b, c):
		self.luxlist.delete(0, END)
		mats = self.libs.GetMaterials(self.varLibs.get()) or []
		for x in mats: self.luxlist.insert(END, x)

	def UseAll(self):
		self.varUseEmitter.set(1); self.UseEmitter()
		self.varUseSubD.set(1); self.UseSubD()
		self.varUseDisplacement.set(1); self.UseDisplacement()
		self.varUseBackface.set(1); self.UseBackface()
		self.varUsePortal.set(1); self.UsePortal()
		self.varUseCustom.set(1); self.UseCustom()
		
	def UseBackface(self):
		if self.varUseBackface.get() == 1:
			self.chkUseBackface.config(state=ACTIVE)
			self.btnUseBackface.config(state=ACTIVE)
		else:
			self.chkUseBackface.config(state=DISABLED)
			self.btnUseBackface.config(state=DISABLED)

	def UseCustom(self):
		if self.varUseCustom.get() == 1:
			self.chkUseCustom.config(state=ACTIVE)
			self.btnUseCustom.config(state=ACTIVE)
		else:
			self.chkUseCustom.config(state=DISABLED)
			self.btnUseCustom.config(state=DISABLED)
			
	def UseDisplacement(self):
		if self.varUseDisplacement.get() == 1:
			self.chkUseDisplacement.config(state=ACTIVE)
			self.btnUseDisplacement.config(state=ACTIVE)
		else:
			self.chkUseDisplacement.config(state=DISABLED)
			self.btnUseDisplacement.config(state=DISABLED)
			
	def UseEmitter(self):
		if self.varUseEmitter.get() == 1:
			self.chkUseEmitter.config(state=ACTIVE)
			self.btnUseEmitter.config(state=ACTIVE)
		else:
			self.chkUseEmitter.config(state=DISABLED)
			self.btnUseEmitter.config(state=DISABLED)

	def UseNone(self):
		self.varUseEmitter.set(0); self.UseEmitter()
		self.varUseSubD.set(0); self.UseSubD()
		self.varUseDisplacement.set(0); self.UseDisplacement()
		self.varUseBackface.set(0); self.UseBackface()
		self.varUsePortal.set(0); self.UsePortal()
		self.varUseCustom.set(0); self.UseCustom()		
			
	def UsePortal(self):
		if self.varUsePortal.get() == 1:
			self.chkUsePortal.config(state=ACTIVE)
			self.btnUsePortal.config(state=ACTIVE)
		else:
			self.chkUsePortal.config(state=DISABLED)
			self.btnUsePortal.config(state=DISABLED)
			
	def UseSubD(self):
		if self.varUseSubD.get() == 1:
			self.chkUseSubD.config(state=ACTIVE)
			self.btnUseSubD.config(state=ACTIVE)
		else:
			self.chkUseSubD.config(state=DISABLED)
			self.btnUseSubD.config(state=DISABLED)

	def UseTicked(self):
		selected = self.matlist.curselection()
		if selected:
			# Get key
			data = self.matlist.get(selected[0])
			key = MakeKey(data[0], data[1])
			m = self.GetMaterial(key)
			
			opt = m.Options()
			
			self.varUseEmitter.set(opt[0]); self.UseEmitter()
			self.varUseSubD.set(opt[1]); self.UseSubD()
			self.varUseDisplacement.set(opt[2]); self.UseDisplacement()
			self.varUseBackface.set(opt[3]); self.UseBackface()
			self.varUsePortal.set(opt[4]); self.UsePortal()
			self.varUseCustom.set(opt[5]); self.UseCustom()
			
			self.Grab()
class pageOptions:
	def __init__(self, parent, gui):
	
		self.FIGURESTACK={}
		self.gui = gui
		self.parameters2add = {
			'LUXGEOM_normals':("LUX Comp Normals", 0, 1, 1),
			'LUXGEOM_subdivide':("LUX Subdivide", 0, 5, 0),
			'LUXGEOM_sharpbounds':("LUX Sharp boundaries", 0, 1, 1),
			'LUXGEOM_normalsplit':("LUX Normal split", 0, 1, 0),
			'LUXGEOM_catmullclark':("LUX Catmull-Clark", 0, 5, 0),
			'LUXGEOM_dispbase':("LUX Displacement base", 0, 1, 0)
			}
		
		# Frame definitions
		self.mainFrame = Frame(parent)
		
		Label(self.mainFrame, text="OPTIONS", anchor=W, font=titleFont, fg="#888888").pack(side=TOP, fill=X)

		self.frmColumnA = Frame(self.mainFrame)
		self.frmColumnA.pack(side=LEFT, fill=Y)
		self.frmColumnB = Frame(self.mainFrame)
		self.frmColumnB.pack(side=LEFT, fill=Y, padx=20)
		
		# Paths
		self.frmPaths = Frame(self.frmColumnA, bd=2, relief=RIDGE)
		self.frmPaths.pack(side=TOP, fill=X, pady=10)
		
		self.frmColumnA_1 = Frame(self.frmColumnA)
		self.frmColumnA_1.pack(side=LEFT, fill=Y, pady=10)
		self.frmColumnA_2 = Frame(self.frmColumnA)
		self.frmColumnA_2.pack(side=LEFT, fill=Y, padx=20, pady=10)
		
		# World volume
		self.frmWV = Frame(self.frmColumnA_1, bd=2, relief=RIDGE)
		self.frmWV.pack(side=TOP)
		
		# Dials
		self.frmDials = Frame(self.frmColumnA_2, bd=2, relief=RIDGE)
		self.frmDials.pack(side=TOP, fill=X)
		
		# Experimental
		self.frmExperimental = Frame(self.frmColumnA_2, bd=2, relief=RIDGE)
		self.frmExperimental.pack(side=TOP, fill=X, pady=10)
		
		self.InitPaths()
		self.InitWorldVolume()
		self.InitDials()
		self.InitExperimental()

	def delCurrent(self):
		self.deleteSingle(poser.Scene().CurrentActor(), True)

	def deleteAll(self):
		for ac in poser.Scene().Actors() :        
			self.deleteSingle(ac, True)
			poser.Scene().DrawAll()
			
	def deleteSingle(self, ac, force=False):
		if self.goodActor(ac) :     
			fig = ac.ItsFigure() 
			if (not force) and fig :
				if fig.InternalName() in self.FIGURESTACK.keys():
					return
				self.FIGURESTACK[fig.InternalName()] = True
				ac=fig.Actor("BODY")
				
			for internalname, params in self.parameters2add.iteritems():
				try:
					ac.RemoveValueParameter(params[0])
				except: pass 
				
		poser.Scene(1).ProcessSomeEvents(10)
				
	def GetAllPaths(self):
		controls = ("varMatLoc", "varAutoMatLoc", "varSceneLoc", "varTemplateLoc", "varTexLoc", "varDataLoc", "varIESLoc", "varSaveLoc")
		return [self.GetPath(p) for p in controls]

	def GetAutoMatLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varAutoMatLoc.set(folder)

	def GetDataLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varDataLoc.set(folder)

	def GetIESLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varIESLoc.set(folder)
		
	def GetMatLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varMatLoc.set(folder)
		
	def GetPath(self, attr):
		p = getattr(self, attr, None)
		if p:
			path = p.get()
		else:
			path = None
		if path == "": path = None
		return path	

	def GetSaveLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varSaveLoc.set(folder)
		
	def GetSceneLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varSceneLoc.set(folder)

	def GetTemplateLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varTemplateLoc.set(folder)
		
	def GetTexLoc(self):
		folder = tkFileDialog.askdirectory()
		if folder: self.varTexLoc.set(folder)
	
	def goodActor(self, ac):
		EXCLUDEDACTORS=("UNIVERSE", "FocusDistanceControl")
		if not (ac.Visible() or ac.IsProp() or ac.IsBodyPart()) : return False
		if (ac.IsBase() or ac.IsCamera() or ac.IsLight() or ac.IsHairProp() or ac.IsZone() or ac.IsDeformer()) : return False
		return ac.Name() not in EXCLUDEDACTORS
	
	def InitDials(self):
		Label(self.frmDials, text="ADD PARAMETER DIALS", bg="black", fg="white").pack(side=TOP, fill=X)
		Button(self.frmDials, text="Add parameters to ALL actors in scene", command=self.setAll).pack(fill=X, expand=1)
		Button(self.frmDials, text="Add parameters to CURRENT actor only", command=self.setCurrent).pack(fill=X, expand=1)
		Button(self.frmDials, text="Remove parameters from ALL actors", command=self.deleteAll).pack(fill=X, expand=1)
		Button(self.frmDials, text="Remove parameters from CURRENT actor", command=self.delCurrent).pack(fill=X, expand=1)
		Message(self.frmDials, text="These buttons add or remove the following parameters:\n\nLUX Comp Normals\nLUX Subdivide\nLUX Sharp boundaries\nLUX Normal split\nLUX Catmul-Clark\nLUX Displacement base").pack(fill=X, expand=1)		
	
	def InitExperimental(self):
		Label(self.frmExperimental, text="Experimental", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		self.varAntiGC = IntVar()
		self.varAntiGC.set(1)
		Checkbutton(self.frmExperimental, text="Anti GC lights", variable=self.varAntiGC).pack(side=LEFT)
		
		self.gui.RegisterControl("options_antigc", "int", self.varAntiGC)
	
	def InitPaths(self):
		Label(self.frmPaths, text="DEFAULT FOLDER LOCATIONS", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		# Default location for materials
		self.frmMatLoc = Frame(self.frmPaths)
		self.frmMatLoc.pack(side=TOP, fill=X)
		Label(self.frmMatLoc, text='Material libraries: ').pack(side=LEFT)
		Button(self.frmMatLoc, text='SELECT', command=self.GetMatLoc).pack(side=RIGHT)
		self.varMatLoc = StringVar()
		self.MatLoc = Entry(self.frmMatLoc, width=60, textvariable=self.varMatLoc)
		self.MatLoc.pack(side=RIGHT, padx=5)

		# Auto load of materials location
		self.frmAutoMatLoc = Frame(self.frmPaths)
		self.frmAutoMatLoc.pack(side=TOP, fill=X)
		Label(self.frmAutoMatLoc, text='Automatically load materials from: ').pack(side=LEFT)
		Button(self.frmAutoMatLoc, text='SELECT', command=self.GetAutoMatLoc).pack(side=RIGHT)
		self.varAutoMatLoc = StringVar()
		self.AutoMatLoc = Entry(self.frmAutoMatLoc, width=60, textvariable=self.varAutoMatLoc)
		self.AutoMatLoc.pack(side=RIGHT, padx=5)
		
		# Default location for scene materials
		self.frmSceneLoc = Frame(self.frmPaths)
		self.frmSceneLoc.pack(side=TOP, fill=X)
		Label(self.frmSceneLoc, text='Scene materials: ').pack(side=LEFT)
		Button(self.frmSceneLoc, text='SELECT', command=self.GetSceneLoc).pack(side=RIGHT)
		self.varSceneLoc = StringVar()
		self.SceneLoc = Entry(self.frmSceneLoc, width=60, textvariable=self.varSceneLoc)
		self.SceneLoc.pack(side=RIGHT, padx=5)

		# Default location for templates
		self.frmTemplateLoc = Frame(self.frmPaths)
		self.frmTemplateLoc.pack(side=TOP, fill=X)
		Label(self.frmTemplateLoc, text='Templates: ').pack(side=LEFT)
		Button(self.frmTemplateLoc, text='SELECT', command=self.GetTemplateLoc).pack(side=RIGHT)
		self.varTemplateLoc = StringVar()
		self.TemplateLoc = Entry(self.frmTemplateLoc, width=60, textvariable=self.varTemplateLoc)
		self.TemplateLoc.pack(side=RIGHT, padx=5)
		
		# Default location for textures
		self.frmTexLoc = Frame(self.frmPaths)
		self.frmTexLoc.pack(side=TOP, fill=X)
		Label(self.frmTexLoc, text='Textures: ').pack(side=LEFT)
		Button(self.frmTexLoc, text='SELECT', command=self.GetTexLoc).pack(side=RIGHT)
		self.varTexLoc = StringVar()
		self.TexLoc = Entry(self.frmTexLoc, width=60, textvariable=self.varTexLoc)
		self.TexLoc.pack(side=RIGHT, padx=5)
		
		# Default location for tabbeddata
		self.frmDataLoc = Frame(self.frmPaths)
		self.frmDataLoc.pack(side=TOP, fill=X)
		Label(self.frmDataLoc, text='Tabbed data files: ').pack(side=LEFT)
		Button(self.frmDataLoc, text='SELECT', command=self.GetDataLoc).pack(side=RIGHT)
		self.varDataLoc = StringVar()
		self.DataLoc = Entry(self.frmDataLoc, width=60, textvariable=self.varDataLoc)
		self.DataLoc.pack(side=RIGHT, padx=5)
		
		# Default location for IES data
		self.frmIESLoc = Frame(self.frmPaths)
		self.frmIESLoc.pack(side=TOP, fill=X)
		self.varIESLoc = StringVar()
		Label(self.frmIESLoc, text='IES files: ').pack(side=LEFT)
		Button(self.frmIESLoc, text='SELECT', command=self.GetIESLoc).pack(side=RIGHT)
		self.IESLoc = Entry(self.frmIESLoc, width=60, textvariable=self.varIESLoc)
		self.IESLoc.pack(side=RIGHT, padx=5)
		
		# Default location for saved scenes
		self.frmSaveLoc = Frame(self.frmPaths)
		self.frmSaveLoc.pack(side=TOP, fill=X)
		self.varSaveLoc = StringVar()
		Label(self.frmSaveLoc, text='Scene files: ').pack(side=LEFT)
		Button(self.frmSaveLoc, text='SELECT', command=self.GetSaveLoc).pack(side=RIGHT)
		self.SaveLoc = Entry(self.frmSaveLoc, width=60, textvariable=self.varSaveLoc)
		self.SaveLoc.pack(side=RIGHT, padx=5)
		
		self.gui.RegisterControl("path_materials", "string", self.varMatLoc)
		self.gui.RegisterControl("path_autoload", "string", self.varAutoMatLoc)
		self.gui.RegisterControl("path_scene", "string", self.varSceneLoc)
		self.gui.RegisterControl("path_template", "string", self.varTemplateLoc)
		self.gui.RegisterControl("path_textures", "string", self.varTexLoc)
		self.gui.RegisterControl("path_data", "string", self.varDataLoc)
		self.gui.RegisterControl("path_ies", "string", self.varIESLoc)
		self.gui.RegisterControl("path_save", "string", self.varSaveLoc)

	def InitWorldVolume(self):
	
		global  titleFont
		
		Label(self.frmWV, text="WORLD VOLUME", bg="black", fg="white", anchor=W).pack(side=TOP, fill=X)
		
		# Use custom World Volume checkbox
		self.varUseCustomWV = IntVar()
		Checkbutton(self.frmWV, text="Use custom world volume", variable=self.varUseCustomWV, anchor=W).pack(side=TOP, fill=X, pady=5)
		
		# Absorption
		self.frmAbsorb = Frame(self.frmWV)
		self.frmAbsorb.pack(side=TOP, fill=X)
		Label(self.frmAbsorb, text="Absorption", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		self.Absorption_R = LabelEntry(self.frmAbsorb, initValue=0.0, text="R", type="float")
		self.Absorption_G = LabelEntry(self.frmAbsorb, initValue=0.0, text="G", type="float")
		self.Absorption_B = LabelEntry(self.frmAbsorb, initValue=0.0, text="B", type="float")
		
		# Scattering
		self.frmScatter = Frame(self.frmWV)
		self.frmScatter.pack(side=TOP, fill=X)
		Label(self.frmScatter, text="Scattering", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		self.Scattering_R = LabelEntry(self.frmScatter, initValue=0.0, text="R", type="float")
		self.Scattering_G = LabelEntry(self.frmScatter, initValue=0.0, text="G", type="float")
		self.Scattering_B = LabelEntry(self.frmScatter, initValue=0.0, text="B", type="float")
		
		# Asymmetry
		self.frmAsymmetry = Frame(self.frmWV)
		self.frmAsymmetry.pack(side=TOP, fill=X)
		Label(self.frmAsymmetry, text="Asymmetry", font=titleFont, anchor=W).pack(side=TOP, fill=X)
		self.Asymmetry_R = PickScale(self.frmAsymmetry, orient=HORIZONTAL, from_=-1.0, to=1.0, resolution=0.1, prefix="R")
		self.Asymmetry_G = PickScale(self.frmAsymmetry, orient=HORIZONTAL, from_=-1.0, to=1.0, resolution=0.1, prefix="G")
		self.Asymmetry_B = PickScale(self.frmAsymmetry, orient=HORIZONTAL, from_=-1.0, to=1.0, resolution=0.1, prefix="B")
		
		self.gui.RegisterControl("use_custom_wv", "int", self.varUseCustomWV)
		self.gui.RegisterControl("abs_r", "float", self.Absorption_R)
		self.gui.RegisterControl("abs_g", "float", self.Absorption_G)
		self.gui.RegisterControl("abs_b", "float", self.Absorption_B)
		self.gui.RegisterControl("sca_r", "float", self.Scattering_R)
		self.gui.RegisterControl("sca_g", "float", self.Scattering_G)
		self.gui.RegisterControl("sca_b", "float", self.Scattering_B)
		self.gui.RegisterControl("asy_r", "float", self.Asymmetry_R)
		self.gui.RegisterControl("asy_g", "float", self.Asymmetry_G)
		self.gui.RegisterControl("asy_b", "float", self.Asymmetry_B)
		pass
	
	def setAll(self):
		for ac in poser.Scene().Actors() :     
			self.setSingle(ac)
			poser.Scene().DrawAll()  
			
	def setCurrent(self):
		self.setSingle(poser.Scene().CurrentActor(), True)

	def setparams(self, param, name, min, max, value):
		param.SetName(name)
		param.SetMinValue(min)
		param.SetMaxValue(max)
		param.SetValue(value)
		param.SetSensitivity(1)
		param.ApplyLimits()	
		
	def setSingle(self, ac, force=False):
		if self.goodActor(ac) :
			fig = ac.ItsFigure() 
			if (not force) and fig :
				if fig.InternalName() in self.FIGURESTACK.keys(): 
					return
				self.FIGURESTACK[fig.InternalName()] = True
				ac=fig.Actor("BODY")
				
			for internalname, params in self.parameters2add.iteritems():
				try:
					ac.CreateValueParameter(internalname)
				except: 
					pass

				p = ac.ValueParameter(internalname)
				self.setparams(p, *params)
				p.ApplyLimits()

			poser.Scene().ProcessSomeEvents(10)
		
# Debug classes
class debug_MATView:
	def __init__ (self, parent, gui):
		self.mainFrame = Frame(parent)
		self.gui = gui
		
		self.frmA = Frame(self.mainFrame)
		self.frmA.pack(side=LEFT, fill=Y, pady=20)
		self.frmB = Frame(self.mainFrame)
		self.frmB.pack(side=LEFT, fill=BOTH, pady=20)

		# Frame for Listbox
		self.lb = Frame(self.frmA)
		self.lb.pack(side=TOP)
		self.scrolly = Scrollbar(self.lb, orient=VERTICAL)		
 
		# The Listbox itself
		self.selected = StringVar()
		self.mats = Listbox(self.lb, width=40, height=30, listvariable=self.selected, yscrollcommand=self.scrolly.set)
		self.scrolly["command"] = self.mats.yview
		self.mats.pack(side=LEFT)
		self.scrolly.pack(side=LEFT, fill=Y)
		Button(self.frmA, text="Show info for selected material", command=self.ShowInfo).pack(fill=X, pady=5) 
		
		# Populate listbox
		d = gui.data.Materials().keys()
		d.sort()
		self.mats.insert(END, *d)
		
		# Output message
		self.msgtxt = StringVar()
		self.msg = Message(self.frmB, textvariable=self.msgtxt, width=750)
		self.msg.pack(side=TOP, fill=X, padx=10)
 
	def ShowInfo(self):
	
		d = self.gui.GetData().Materials()
		m = d[self.mats.get(ACTIVE)]
		
		txt = "Name: %s" % m.Name() + "\n"
		txt += "Lux material: %s" % m.Luxname() + "\n\n"
		
		txt += "Diffuse colour: %.3f, %.3f, %.3f" % m.simplemat.DiffuseColour + "\n"
		txt += "Diffuse map: %s" % m.simplemat.DiffuseMap[0] + "\n"
		txt += "Specular colour: %.3f, %.3f, %.3f" % m.simplemat.SpecularColour + "\n"
		txt += "Specular value: %.3f" % m.simplemat.SpecularValue + "\n"
		txt += "Specular map: %s" % m.simplemat.SpecularMap[0] + "\n"
		txt += "Specular size: %.3f" % m.simplemat.SpecularSize + "\n"
		txt += "Ambient colour: %.3f, %.3f, %.3f" % m.simplemat.AmbientColour + "\n"
		txt += "Ambient value: %.3f" % m.simplemat.AmbientValue + "\n"
		txt += "Transparency value: %.3f" % m.simplemat.TransparencyValue + "\n"
		txt += "Transparency map: %s" % m.simplemat.TransparencyMap[0] + "\n"
		txt += "Reflection colour: %.3f, %.3f, %.3f" % m.simplemat.ReflectionColour + "\n"
		txt += "Reflection value: %.3f" % m.simplemat.ReflectionValue + "\n"
		txt += "Refraction value: %.3f" % m.simplemat.RefractionValue + "\n"
		txt += "Bump map: %s" % m.simplemat.BumpMap[0] + "\n"
		txt += "Bump value: %.3f" % m.simplemat.BumpValue + "\n"
		txt += "Displacement map: %s" % m.simplemat.DisplacementMap[0] + "\n"
		txt += "Displacement value: %.3f" % m.simplemat.DisplacementValue + "\n"
		
		self.msgtxt.set(txt)
 
# Lux classes
class LuxAttribute:
	def __init__(self, name):
		self.data = None
		self.name = name
		self.nodes = {}
		self.type = ""

	def AdditionalData(self):
		return self.data
	
	def AddNode(self, type, id, value):
		self.nodes[id] = [type, value]
	
	def Name(self):
		return self.name
	
	def SetAdditionalData(self, data):
		self.data = data
	
	def ProcessInputs(self, p2lmat, maxD, maxS, exporter):
	
		sm = p2lmat.PoserMaterials()
		
		for k, i in self.nodes.items():
			if isinstance(i[1], str):
								
				# USE_bmpTEX
				if not i[1].find("USE_bmpTEX") == -1:
					if sm.BumpMap[0]:
						i[1] = i[1].replace("USE_bmpTEX", sm.BumpMap[0])
					else:
						del self.nodes[k]
			
				# USE_BUMPMAP
				if not i[1].find("USE_BUMPMAP") == -1:
					if sm.BumpMap[0]:
						i[1] = i[1].replace("USE_BUMPMAP", exporter.WriteTextureMap(p2lmat.Key(), sm.BumpMap, True, datatype="bump"))
					else:
						del self.nodes[k]
						
				# USE_DATAPATH
				path = exporter.GetDataPath()
				if path: i[1] = i[1].replace("USE_DATAPATH", path)
			
				# USE_HS
				i[1] = i[1].replace("USE_HS", str(sm.SpecularSize))
				
				# USE_HV
				i[1] = i[1].replace("USE_HV", str(sm.SpecularValue))
				
				# USE_InvHV
				i[1] = i[1].replace("USE_InvHV", str(1.0 - sm.SpecularValue))
				
				# Use_InvRV
				i[1] = i[1].replace("USE_InvRV", str(1.0 - sm.ReflectionValue))
				
				# USE_KD
				i[1] = i[1].replace("USE_KD", LuxColour(sm.DiffuseColour, maxD))
				
				# USE_KdMAP
				if not i[1].find("USE_KdMAP") == -1:
					if sm.DiffuseMap[0]:
						i[1] = i[1].replace("USE_KdMAP", exporter.WriteTextureMap(p2lmat.Key(), sm.DiffuseMap, False, sm.DiffuseColour, maxD))
					else:
						self.nodes[k] = ("color", LuxColour(sm.DiffuseColour, maxD))
						
				# USE_KdTEX
				if not i[1].find("USE_KdTEX") == -1:
					if sm.DiffuseMap[0]:
						i[1] = i[1].replace("USE_KdTEX", sm.DiffuseMap[0])
					else:
						del self.nodes[k]

				# USE_KEY
				i[1] = i[1].replace("USE_KEY", p2lmat.Key())
				
				# USE_KR
				i[1] = i[1].replace("USE_KR", LuxColour(ColourValue(sm.ReflectionColour, sm.ReflectionValue)))
				
				# USE_KS
				if sm.SpecularMap[0]:
					if i[1].find("USE_KS") != -1:
						i[0] = i[0].replace("color", "texture")
						i[1] = i[1].replace("USE_KS", exporter.WriteTextureMap(p2lmat.Key(), sm.SpecularMap, True, sm.SpecularColour, maxS, datatype="specular"))
				else:
					i[1] = i[1].replace("USE_KS", LuxColour(sm.SpecularColour, maxS))

				# USE_KsTEX
				if not i[1].find("USE_KsTEX") == -1:
					if sm.SpecularMap[0]:
						i[1] = i[1].replace("USE_KsTEX", sm.SpecularMap[0])
					else:
						del self.nodes[k]
				
				# USE_LIB
				if p2lmat.Luxname(): i[1] = i[1].replace("USE_LIB", p2lmat.Luxname().split('/')[0])
				
				# USE_RV
				i[1] = i[1].replace("USE_RV", str(sm.ReflectionValue))
				
				# USE_TEXTUREPATH
				path = exporter.GetTexturePath()
				if path: i[1] = i[1].replace("USE_TEXTUREPATH", path)

				# USE_TransTEX
				if not i[1].find("USE_TransTEX") == -1:
					if sm.TransparencyMap[0]:
						i[1] = i[1].replace("USE_TransTEX", sm.TransparencyMap[0])
					else:
						del self.nodes[k]
	
	def SetType(self, type, key=None):
		if key:
			self.type = type.replace("USE_KEY", key)
		else:
			self.type = type
	
	def Type(self):
		return self.type
	
	def Write(self, file, nodesonly=False):
		if nodesonly == False: print >> file, '\n%s "%s"' % (self.name, self.type)
		self.WriteIndices(file)
				
	def WriteIndices(self, file):
		for key, item in self.nodes.items():
			if item[0] in ("color", "float", "integer", "point", "vector"):
				print >> file, '"%s %s" [%s]' % (item[0], key, item[1])
			else:
				print >> file, '"%s %s" ["%s"]' % (item[0], key, item[1])	
class LuxMaterial(LuxAttribute):
	def __init__(self):
		LuxAttribute.__init__(self, "MakeNamedMaterial")
		self.exterior_volume = None
		self.interior_volume = None
		self.displacement = None

	def ExteriorVolume(self):
		return self.exterior_volume
		
	def InteriorVolume(self):
		return self.interior_volume
	
	def SetDisplacement(self, parms):
		self.displacement = parms
	
	def SetExteriorVolume(self, volume):
		self.exterior_volume = volume
		
	def SetInteriorVolume(self, volume):
		self.interior_volume = volume
class LuxTexture(LuxAttribute):
	def __init__(self):
		LuxAttribute.__init__(self, "Texture")	
		self.texture_type = ""
		
	def SetTextureType(self, type):
		self.texture_type = type
		
	def Write(self, file):
		print >> file, '\n%s "%s" %s' % (self.name, self.type, self.texture_type)
		self.WriteIndices(file)
class LuxVolume(LuxAttribute):
	def __init__(self):
		LuxAttribute.__init__(self, "MakeNamedVolume")
		self.volume_type = ""
		
	def SetVolumeType(self, type):
		self.volume_type = type
		
	def Write(self, file):
		print >> file, '\n%s "%s" %s' % (self.name, self.type, self.volume_type)
		self.WriteIndices(file)
		
# Analyser classes
class Analyser:
	def __init__(self, libmat, sm):
		self.libmat = libmat
		self.simplemat = sm
		
		self.inputs = []
		
		if libmat.use_kdmap: self.DoKdMap()
		if libmat.use_kd: self.DoKd()
		if libmat.use_kr: self.DoKr()
		if libmat.use_ks: self.DoKs()
		if libmat.use_bump: self.DoBumpMap()
		self.DoExtra()
		
	def DoKdMap(self):
		if self.simplemat.DiffuseMap[0]: self.inputs.append(('texture', 'Kd', 'USE_KdMAP'))
		
	def DoKd(self):
		if not self.simplemat.DiffuseMap[0]: self.inputs.append(('color', 'Kd', 'USE_KD'))
		
	def DoKr(self):
		if self.simplemat.ReflectionValue > 0: self.inputs.append(('color', 'Kr', 'USE_KR'))
		
	def DoKs(self):
		if self.simplemat.SpecularValue > 0: self.inputs.append(('color', 'Ks', 'USE_KS'))
	
	def DoBumpMap(self):
		if self.simplemat.BumpMap[0]: self.inputs.append(('texture', 'bumpmap', 'USE_BUMPMAP'))
		
	def DoExtra(self):
		pass
	
	def Inputs(self):
		return self.inputs
class AnalyseMatte(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "matte"
	def DoKr(self):	pass
	def DoKs(self): pass
class AnalyseMatteTranslucent(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "mattetranslucent"
		
	def DoKdMap(self):
		if self.simplemat.DiffuseMap[0]:
			self.inputs.append(('texture', 'Kr', 'USE_KdMAP'))
			self.inputs.append(('texture', 'Kt', 'USE_KdMAP'))
		
	def DoKd(self):
		if not self.simplemat.DiffuseMap[0]:
			self.inputs.append(('color', 'Kr', 'USE_KD'))
			self.inputs.append(('color', 'Kt', "0.2 0.2 0.2"))
		
	def DoKr(self): pass	
	
	def DoKs(self): pass
class AnalyseGlossy(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "glossy"
	
	def DoKr(self):
		amount = min(1 - self.simplemat.ReflectionValue, 0.8)
		self.inputs.append(('float', 'uroughness', amount))
		self.inputs.append(('float', 'vroughness', amount))
class AnalyseGlossyTranslucent(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "glossytranslucent"
		
	def DoKd(self):
		if not self.simplemat.DiffuseMap[0]:
			self.inputs.append(('color', 'Kr', 'USE_KD'))
			self.inputs.append(('color', 'Kt', "0.9 0.9 0.9"))
		
	def DoKr(self):
		amount = min(1 - self.simplemat.ReflectionValue, 0.8)
		self.inputs.append(('float', 'uroughness', amount))
		self.inputs.append(('float', 'vroughness', amount))
class AnalyseMetal(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "metal"
		
	def DoKdMap(self): pass
	
	def DoKd(self): pass
	
	def DoKr(self): pass
	
	def DoKs(self): pass
class AnalyseShinyMetal(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "shinymetal"
		
	def DoKdMap(self): pass
	
	def DoKd(self): self.inputs.append(('color', 'Kr', 'USE_KD'))
	
	def DoKr(self):
		amount = min(1 - self.simplemat.ReflectionValue, 0.8)
		self.inputs.append(('float', 'uroughness', amount))
		self.inputs.append(('float', 'vroughness', amount))	
class AnalyseMirror(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
	LuxType = "mirror"
	
	def DoKdMap(self): pass
	
	def DoKd(self): pass
	
	def DoKr(self): self.inputs.append(('color', 'Kr', 'USE_KR'))
	
	def DoKs(self): pass
class AnalyseGlass(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "glass"
		
	def DoKdMap(self):
		if self.simplemat.DiffuseMap[0]:
			self.inputs.append(('texture', 'Kr', 'USE_KdMAP'))
			self.inputs.append(('texture', 'Kt', 'USE_KdMAP'))
		
	def DoKd(self):
		if not self.simplemat.DiffuseMap[0]:
			self.inputs.append(('color', 'Kr', 'USE_KD'))
			self.inputs.append(('color', 'Kt', 'USE_KD'))
			
	def DoKr(self): pass
	
	def DoKs(self): pass
class AnalyseGlass2(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "glass2"
class AnalyseRoughGlass(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "roughglass"

	def DoKdMap(self):
		if self.simplemat.DiffuseMap[0]:
			self.inputs.append(('texture', 'Kr', 'USE_KdMAP'))
			self.inputs.append(('texture', 'Kt', 'USE_KdMAP'))
		
	def DoKd(self):
		if not self.simplemat.DiffuseMap[0]:
			self.inputs.append(('color', 'Kr', 'USE_KD'))
			self.inputs.append(('color', 'Kt', "USE_KD"))
			
	def DoKr(self):
		amount = min(1 - self.simplemat.ReflectionValue, 0.8)
		self.inputs.append(('float', 'uroughness', amount))
		self.inputs.append(('float', 'vroughness', amount))	
	
	def DoKs(self): pass		
class AnalyseCarpaint(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "carpaint"
		
	def DoKr(self): pass
class AnalyseScatter(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "scatter"
	
	def DoKr(self): pass

	def DoKs(self): pass
class AnalyseVelvet(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "velvet"
		
	def DoKr(self): pass

	def DoKs(self): pass
class AnalyseNull(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "null"
class AnalysePortal(Analyser):
	def __init__(self, libmat, sm):
		Analyser.__init__(self, libmat, sm)
		LuxType = "portal"

# Custom material classes
class CustomControl:
	def __init__(self, parent, input, type, callback=None):
	
		BG = "grey"
		FG = "black"
		
		self.input = input
		
		# Label and 'Use' option
		topframe = Frame(parent, bg=BG)
		topframe.pack(side=TOP, fill=X)
		Label(topframe, text='%s (%s)' % (input[1], input[0]), bg=BG, fg=FG).pack(side=LEFT)
		self.varUse = IntVar()
		self.varUse.set(1)
		if callback: self.varUse.trace('w', callback)
		Checkbutton(topframe, variable=self.varUse, bg=BG).pack(side=RIGHT)
		
		# Frame for custom control
		lowerframe = Frame(parent)
		lowerframe.pack(side=TOP, fill=X)
		self.control = None
		
		if input[2] == "float":
			self.control = CustomWidgetFloat(lowerframe, input, callback)
		elif input[2] == "bool":
			self.control = CustomWidgetBool(lowerframe, input, callback)		
		elif input[2] == "Bumpmap":
			self.control = CustomWidgetBumpmap(lowerframe, input, callback)	
		elif input[2] == "RGBColor":
			self.control = CustomWidgetRGBColor(lowerframe, input, callback)
		elif input[2] == "list":
			self.control = CustomWidgetList(lowerframe, input, type, callback)
		else:
			self.control = CustomWidgetUnknown(lowerframe, input, callback)
		
	def get(self):
		return self.control.get()
		
	def load(self, value):
		self.control.load(value)
		
	def save(self):
		return self.control.save()
		
	def set(self, value):
		self.control.set(value)

	def SetUsed(self, value):
		self.varUse.set(value)
	
	def Type(self):
		return self.input[0]
		
	def Used(self):
		return self.varUse.get()
class CustomMaterial:
	def __init__(self, parent, libs, callback):
		
		# Make our own copy of the parameters
		self.callback = callback
		self.libs = libs
		self.parent = parent
		
		# other variables
		self.framelist = []							# List of all frames in use
		self.controls = []
		self.current_values = {}
		self.reset = False
		
		# Define frames
		self.frame = VerticalScrolledFrame(parent)
		self.frame.pack(side=TOP, fill=BOTH, expand=1)
		self.mainframe = self.frame.interior
		
		# Initialise
		self.DefineLuxMaterials()
		self.Initialise()
	
	def AddToLibrary(self, key):
		
		newdef = LibDef()
		newdef.SetKey(key)
		newdef.SetType(self.varMat.get())
		
		for c in self.controls:
			if c.Used():
				
				newnode = c.get()
				
				if newnode:			
					if newnode[0] in ("color", "float", "integer", "point", "vector"):
						newnode = '"%s %s" [%s]' % (newnode[0], newnode[1], newnode[2])
					else:
						newnode = '"%s %s" ["%s"]' % (newnode[0], newnode[1], newnode[2])						
					newdef.AddNode(newnode)
				
				if c.Type() == "bumpmap":
					newdef.SetBumpmap(True)
					
		if not self.varUseExt.get(): newdef.SetVolumeExt(MakeKey("P2L", self.varExtVol.get()))
		if not self.varUseInt.get(): newdef.SetVolumeInt(MakeKey("P2L", self.varIntVol.get()))
		
		return newdef
		
	def BuildControls(self):
	
		type = self.varMat.get()
		
		for i in self.luxtypes[type]:
			
			input = self.inputs[i]
			
			# Make container frame
			frame = Frame(self.mainframe, bd=2, relief=RIDGE)
			frame.pack(side=TOP, fill=X, pady=2)
			self.framelist.append(frame)
			
			# Add control
			control = CustomControl(frame, input, type, callback=self.callback)
			self.controls.append(control)
			
			# Check for existing value
			v = self.current_values.get(input[0], None)
			if v: control.load(v)

	def ChangeExtType(self, *dummy):
		newlist = self.GetVolumeList(self.typelist.index(self.varExtType.get()), self.types)
		newlist.sort()
		self.optExtVol.ReplaceList(newlist)
		if not self.varExtVol.get() in newlist:
			self.varExtVol.set(newlist[0])
		
	def ChangeIntType(self, *dummy):
		newlist = self.GetVolumeList(self.typelist.index(self.varIntType.get()), self.types)
		newlist.sort()
		self.optIntVol.ReplaceList(newlist)
		if not self.varIntVol.get() in newlist:
			self.varIntVol.set(newlist[0])
	
	def ChangeMaterial(self, *dummy):
		if self.reset:
			self.reset = False
		else:
			self.CopyAllValues()
		self.DeleteAllControls()
		self.BuildControls()
		if self.callback: self.callback()
		
	def CopyAllValues(self):
		for ctl in self.controls:
			c = ctl.save()
			if c: self.current_values[c[0]] = c[1]
		
	def DefineLuxMaterials(self):
	
		self.lists = {
			"carpaint" : ("ford f8", "polaris silber", "opel titan", "bmw339", "2k acrylack", "white", "blue", "blue matte"),
			"metal" : ("amorphous carbon", "silver", "gold", "copper", "aluminium")
			}
		
		self.inputs = {
			"architectural" : ("architectural", "Architectural", "bool", False),
			"backface_d" : ("backface_d", "Backface depth", "float", (0.0, 1.0, 0.0)),
			"backface_index" : ("backface_index", "Backface IOR", "float", (0.0, 100.0, 1.0)),
			"backface_Ka" : ("backface_Ka", "Backface absorption", "RGBColor", "Colour"),
			"backface_Ks" : ("backface_Ks", "Backface specular", "RGBColor", "USE_KS"),
			"backface_multibounce" : ("backface_multibounce", "Backface multibounce", "bool", False),
			"backface_vroughness" : ("backface_vroughness", "Backface VRough", "float", (0.0, 1.0, 0.5)),
			"backface_uroughness" : ("backface_uroughness", "Backface URough", "float", (0.0, 1.0, 0.5)),
			"bumpmap" : ("bumpmap", "Bump", "Bumpmap", "Use bumpmap if available"),
			"cauchyb" : ("cauchyb", "Cauchy B", "float", (0.0, 1.0, 0.0)),
			"d" : ("d", "Depth", "float", (0.0, 1.0, 0.0)),
			"dispersion" : ("dispersion", "Dispersion", "bool", False),
			"energyconserving" : ("energyconserving", "Conserve energy", "bool", False),
			"film" : ("film", "Thickness", "float", (1.0, 1500.0, 200.0)),
			"filmindex" : ("filmindex", "IOR", "float", (1.0, 6.0, 1.5)),
			"index" : ("index", "IOR", "float", (0.0, 100.0, 1.0)),
			"Ka" : ("Ka", "Absorption", "RGBColor", "Colour"),
			"Kd" : ("Kd", "Diffuse", "RGBColor", "USE_KdMAP"),
			"Kr" : ("Kr", "Reflection colour", "RGBColor", "USE_KD"),
			"Ks" : ("Ks", "Specular", "RGBColor", "USE_KS"),
			"Ks1" : ("Ks1", "Ks1", "RGBColor", "Colour"),
			"Ks2" : ("Ks2", "Ks2", "RGBColor", "Colour"),
			"Ks3" : ("Ks3", "Ks3", "RGBColor", "Colour"),
			"Kt" : ("Kt", "Transmission", "RGBColor", "Colour"),
			"M1" : ("M1", "M1", "float", (0.0, 1.0, 1.0)),
			"M2" : ("M2", "M2", "float", (0.0, 1.0, 1.0)),
			"M3" : ("M3", "M3", "float", (0.0, 1.0, 1.0)),
			"multibounce" : ("multibounce", "Multibounce", "bool", False),
			"onesided" : ("onesided", "Single sided", "bool", False),
			"name" : ("name", "Name", "list", self.lists),
			"p1" : ("p1", "P1", "float", (-100.0, 100.0, -2.0)),
			"p2" : ("p2", "P2", "float", (-100.0, 100.0, 20)),
			"p3" : ("p3", "P3", "float", (-100.0, 100.0, 2)),
			"R1" : ("R1", "R1", "float", (0.0, 1.0, 1.0)),
			"R2" : ("R2", "R2", "float", (0.0, 1.0, 1.0)),
			"R3" : ("R3", "R3", "float", (0.0, 1.0, 1.0)),
			"sigma" : ("sigma", "Sigma", "float", (0.0, 1.0, 0.1)),
			"thickness" : ("thickness", "Thickness", "float", (0.0, 1.0, 0.1)),
			"uroughness" : ("uroughness", "URough", "float", (0.0, 1.0, 0.5)),
			"vroughness" : ("vroughness", "VRough", "float", (0.0, 1.0, 0.5)),
			}
		
		self.luxtypes = {
			"matte" :("Kd", "sigma", "bumpmap"),
			"mattetranslucent" : ("Kr", "Kt", "sigma", "energyconserving", "bumpmap"),
			"glossy" :("Kd", "Ks", "Ka", "index", "d", "uroughness", "vroughness", "bumpmap"),
			"glossytranslucent" : ("Kd", "Kt", "Ks", "index", "Ka", "d", "uroughness", "vroughness", "bumpmap", "multibounce", "onesided", "backface_Ks", "backface_index", "backface_Ka", "backface_d", "backface_uroughness", "backface_vroughness", "backface_multibounce"),
			"metal" : ("name", "uroughness", "vroughness", "bumpmap"),
			"shinymetal" : ("Kr", "Ks", "uroughness", "vroughness", "film", "filmindex", "bumpmap"),
			"mirror" : ("Kr", "film", "filmindex", "bumpmap"),
			"glass" : ("Kr", "Kt", "index", "cauchyb", "film", "filmindex", "architectural", "bumpmap"),
			"glass2" : ("architectural", "dispersion", "bumpmap"),
			"roughglass" : ("Kr", "Kt", "uroughness", "vroughness", "index", "cauchyb", "bumpmap"),
			"carpaint" : ("name", "Kd", "Ka", "d", "Ks1", "Ks2", "Ks3", "R1", "R2", "R3", "M1", "M2", "M3", "bumpmap"),
			#"scatter" : (),
			"velvet" : ("Kd", "p1", "p2", "p3", "thickness"),
			}
		
	def DeleteAllControls(self):
		for f in self.framelist: f.destroy()
		self.framelist = []
		self.controls = []
	
	def FindType(self, vol):
		for x in self.typelist:
			i = self.typelist.index(x)
			list = self.GetVolumeList(i, self.types)
			if vol in list:
				return x
	
	def GetVolumeList(self, i, types):
		list = [x[0] for x in types[i][1]]
		return list
	
	def Initialise(self):
		self.InitialiseVolumes()
		self.InitialiseMaterialMenu()
		self.SetMaterialType("matte")
		
	def InitialiseMaterialMenu(self):
		self.frmMatMenu = Frame(self.mainframe)
		self.frmMatMenu.pack(side=TOP, fill=X, pady=5)
		self.varMat = StringVar()
		self.varMat.trace( 'w', self.ChangeMaterial)
		menulist = self.luxtypes.keys()
		menulist.sort()
		self.optMat = LabelledMenu(self.frmMatMenu, self.varMat, menulist, label="Lux material")
	
	def InitialiseVolumes(self):
		
		self.types, self.iors = self.libs.GetBlenderVolumes()
		self.typelist = [x[0] for x in self.types]
		vollist = self.GetVolumeList(0, self.types)
		
		self.frmVols = Frame(self.mainframe, bd=2, relief=RIDGE)
		self.frmVols.pack(side=TOP, fill=X)
	
		# Interior volume
		Label(self.frmVols, text="Interior volume", bg="grey", fg="black").pack(side=TOP, fill=X)

		self.varIntType = StringVar()
		self.varIntType.trace('w', self.ChangeIntType)
		self.optIntType = LabelledMenu(self.frmVols, self.varIntType, self.typelist, label="Type")
		
		self.varIntVol = StringVar()
		self.varIntVol.set(vollist[0])
		self.optIntVol = LabelledMenu(self.frmVols, self.varIntVol, vollist, label="Volume", callback=self.callback)

		self.varUseInt = IntVar()
		self.varUseInt.set(1)
		self.varUseInt.trace('w', self.callback)
		Checkbutton(self.frmVols, text="Ignore", variable=self.varUseInt, anchor=W).pack(side=TOP, fill=X)
		
		# Exterior volume
		Label(self.frmVols, text="Exterior volume", bg="grey", fg="black").pack(side=TOP, fill=X)

		self.varExtType = StringVar()
		self.optExtType = LabelledMenu(self.frmVols, self.varExtType, self.typelist, label="Type")
		
		self.varExtVol = StringVar()
		self.varExtVol.set(vollist[0])

		self.varIntType.set(self.typelist[0])
		self.optExtVol = LabelledMenu(self.frmVols, self.varExtVol, vollist, label="Volume", callback=self.callback)
		self.varExtType.trace('w', self.ChangeExtType)
		self.varExtType.set("Gases")
			
		self.varUseExt = IntVar()
		self.varUseExt.set(1)
		self.varUseExt.trace('w', self.callback)
		Checkbutton(self.frmVols, text="Ignore", variable=self.varUseExt, anchor=W).pack(side=TOP, fill=X)
	
	def LoadData(self, matdef):
	
		self.varMat.set(matdef.Type())
		
		analysed = {}
		for n in matdef.Nodes():
			analysed[n[1]] = (n[2], n[0])
			
		for ctl in self.controls:
			if ctl.Type() == "bumpmap":
				ctl.SetUsed(matdef.Bumpmap())
			else:
				value = analysed.get(ctl.Type(), None)
				if value:
					ctl.set(value)
				else:
					ctl.SetUsed(False)
					
		ext = matdef.ExternalVolume()
		if ext == None:
			self.varUseExt.set(1)
		else:
			self.varUseExt.set(0)
			vol = ext.split('/')[1]
			self.varExtVol.set(vol)
			self.varExtType.set(self.FindType(vol))
			
		int = matdef.InternalVolume()
		if int == None:
			self.varUseInt.set(1)
		else:
			self.varUseInt.set(0)
			vol = int.split('/')[1]
			self.varIntVol.set(vol)
			self.varIntType.set(self.FindType(vol))
					
	def Reset(self):
		self.current_values = {}
		self.reset = True
		self.varMat.set("matte")
	
	def SetMaterialType(self, type):
		self.varMat.set(type)
class CustomWidget:
	def __init__(self, parent, input, callback):
		self.callback = callback
		self.input = input
		self.menuOptions = {
			"Colour" : ("Custom colour", "color"),
			"USE_KD" : ("Diffuse colour", "color"),
			"USE_KdMAP" : ("Diffuse texture map", "texture"),
			"USE_KS" : ("Specular colour", "color")
			}
	
	def get(self):
		return None

	def load(self, value):
		pass
		
	def save(self):
		return None
	
	def set(self, value):
		pass
class CustomWidgetBool(CustomWidget):
	def __init__(self, parent, input, callback=None):
		CustomWidget.__init__(self, parent, input, callback)	
		self.control = Boolean(parent, label=input[1], text=None, initValue=input[3], callback=self.callback)
		
	def get(self):
		return ("bool", self.input[0], ("false", "true")[self.control.get()])
		
	def load(self, value):
		self.control.set(value)
		
	def save(self):
		return (self.input[0], self.control.get())
		
	def set(self, value):
		self.control.set(eval(value[0].capitalize()))
class CustomWidgetFloat(CustomWidget):
	def __init__(self, parent, input, callback=None):
		CustomWidget.__init__(self, parent, input, callback)
		self.control = LabelEntry(parent, type="float", minVal=input[3][0], maxVal=input[3][1], initValue=input[3][2], callback=self.callback)
	
	def get(self):
		return ("float", self.input[0], self.control.get())
	
	def load(self, value):
		self.control.set(value)
		
	def save(self):
		return (self.input[0], self.control.get())

	def set(self, value):
		self.control.set(value[0])
class CustomWidgetBumpmap(CustomWidget):
	def __init__(self, parent, input, callback=None):
		CustomWidget.__init__(self, parent, input, callback)
		Label(parent, text=input[3]).pack(side=LEFT)
class CustomWidgetList(CustomWidget):
	def __init__(self, parent, input, type, callback=None):
		CustomWidget.__init__(self, parent, input, callback)
		self.varMenu = StringVar()
		self.control = LabelledMenu(parent, self.varMenu, input[3][type], label=None)
		self.control.set("aluminium")
		if callback: self.varMenu.trace('w', callback)

	def get(self):
		return ("string", self.input[0], self.control.get())
	
	def load(self, value):
		self.control.set(value)
		
	def save(self):
		return (self.input[0], self.control.get())

	def set(self, value):
		self.control.set(value[0])		
class CustomWidgetRGBColor(CustomWidget):
	def __init__(self, parent, input, callback=None):
		CustomWidget.__init__(self, parent, input, callback)
		menu = MakeMenu(self.menuOptions,('Colour', 'USE_KdMAP', 'USE_KD', 'USE_KS'))
		self.ColourIsPacked = False
		self.varMenu = StringVar()
		OptionMenu(parent, self.varMenu, *menu).pack(side=LEFT)
		self.Colour = PickColour(parent, width=5, rgb=False, callback=self.callback)
		self.varMenu.set(self.menuOptions[input[3]][0])
		self.Change(False, None)
		self.varMenu.trace('w', self.Change)
	
	def Change(self, usecallback=True, *dummy):
		if self.varMenu.get() == self.menuOptions['Colour'][0]:
			if not self.ColourIsPacked:
				self.Colour.pack(side=RIGHT, fill=None)
				self.ColourIsPacked = True
		else:
			if self.ColourIsPacked:
				self.Colour.pack_forget()
				self.ColourIsPacked = False
		if self.callback and usecallback: self.callback()
				
	def get(self):
		
		if self.varMenu.get() == self.menuOptions['Colour'][0]:
			return ("color", self.input[0], str3(self.Colour.get()))
		
		if	self.varMenu.get() == self.menuOptions['USE_KdMAP'][0]:
			type = "texture"
		else:
			type = "color"
			
		return (type, self.input[0], self.Reverse(self.varMenu.get()))
		
	def load(self, value):
		self.varMenu.set(value[0])
		self.Colour.set(value[1])
		
	def Reverse(self, value):
		temp = {}
		for k, v in self.menuOptions.items():
			temp[v[0]] = k
		return temp[value]
	
	def save(self):
		return (self.input[0], (self.varMenu.get(), self.Colour.get()))	

	def set(self, value):
		if value[0] in ('USE_KdMAP', 'USE_KD', 'USE_KS'):
			self.varMenu.set(self.menuOptions[value[0]][0])
		else:
			self.varMenu.set(self.menuOptions['Colour'][0])
			self.Colour.set(MakeTuple(value[0]))
class CustomWidgetUnknown(CustomWidget):
	def __init__(self, parent, input, callback=None):
		CustomWidget.__init__(self, parent, input, callback)
		Label(parent, text="Unknown data type", anchor=W).pack(side=TOP, fill=X)
		
# Miscellaneous classes		
class LibraryManager:
	def __init__(self):
		
		(LIBRARIES, MATERIALS, PARAMETERS, INDICES, TEXTURES, VOLUMES) = range(6)
		self.libs = self.MakeDefaultLuxMaterials()
		self.AddBlenderVolumes()

	def AddBlenderVolumes(self):
	
		self.iortree = [ ("Liquids", [("Acetone", 1), ("Alcohol, Ethyl (grain)", 2), ("Alcohol, Methyl (wood)", 3), ("Beer", 4), ("Benzene", 5), ("Carbon tetrachloride", 6), ("Carbon disulfide", 7), ("Carbonated Beverages", 8), ("Chlorine (liq)", 9), ("Cranberry Juice (25%)", 10), ("Glycerin", 11), ("Honey, 13% water content", 12), ("Honey, 17% water content", 13), ("Honey, 21% water content", 14), ("Ice", 15), ("Milk", 16), ("Oil, Clove", 17), ("Oil, Lemon", 18), ("Oil, Neroli", 19), ("Oil, Orange", 20), ("Oil, Safflower", 21), ("Oil, vegetable (50 C)", 22), ("Oil of Wintergreen", 23), ("Rum, White", 24), ("Shampoo", 25), ("Sugar Solution 30%", 26), ("Sugar Solution 80%", 27), ("Turpentine", 28), ("Vodka", 29), ("Water (0 C)", 30), ("Water (100 C)", 31), ("Water (20 C)", 32), ("Whisky", 33) ] ), ("Gases", [("Vacuum", 101), ("Air @ STP", 102), ("Air", 103), ("Helium", 104), ("Hydrogen", 105), ("Carbon dioxide", 106) ]), ("Transparent\x20", [("Eye, Aqueous humor", 201), ("Eye, Cornea", 202), ("Eye, Lens", 203), ("Eye, Vitreous humor", 204), ("Glass, Arsenic Trisulfide", 205), ("Glass, Crown (common)", 206), ("Glass, Flint, 29% lead", 207), ("Glass, Flint, 55% lead", 208), ("Glass, Flint, 71% lead", 209), ("Glass, Fused Silica", 210), ("Glass, Pyrex", 211), ("Lucite", 212), ("Nylon", 213), ("Obsidian", 214), ("Plastic", 215), ("Plexiglas", 216), ("Salt", 217)  ]), ("Gemstones", [("Agate", 301), ("Alexandrite", 302), ("Almandine", 303), ("Amber", 304), ("Amethyst", 305), ("Ammolite", 306), ("Andalusite", 307), ("Apatite", 308), ("Aquamarine", 309), ("Axenite", 310), ("Beryl", 311), ("Beryl, Red", 312), ("Chalcedony", 313), ("Chrome Tourmaline", 314), ("Citrine", 315), ("Clinohumite", 316), ("Coral", 317), ("Crystal", 318), ("Crysoberyl, Catseye", 319), ("Danburite", 320), ("Diamond", 321), ("Emerald", 322), ("Emerald Catseye", 323), ("Flourite", 324), ("Garnet, Grossular", 325), ("Garnet, Andradite", 326), ("Garnet, Demantiod", 327), ("Garnet, Mandarin", 328), ("Garnet, Pyrope", 329), ("Garnet, Rhodolite", 330), ("Garnet, Tsavorite", 331), ("Garnet, Uvarovite", 332), ("Hauyn", 333), ("Iolite", 334), ("Jade, Jadeite", 335), ("Jade, Nephrite", 336), ("Jet", 337), ("Kunzite", 338), ("Labradorite", 339), ("Lapis Lazuli", 340), ("Moonstone", 341), ("Morganite", 342), ("Obsidian", 343), ("Opal, Black", 344), ("Opal, Fire", 345), ("Opal, White", 346), ("Oregon Sunstone", 347), ("Padparadja", 348), ("Pearl", 349), ("Peridot", 350), ("Quartz", 351), ("Ruby", 352), ("Sapphire", 353), ("Sapphire, Star", 354), ("Spessarite", 355), ("Spinel", 356), ("Spinel, Blue", 357), ("Spinel, Red", 358), ("Star Ruby", 359), ("Tanzanite", 360), ("Topaz", 361), ("Topaz, Imperial", 362), ("Tourmaline", 363), ("Tourmaline, Blue", 364), ("Tourmaline, Catseye", 365), ("Tourmaline, Green", 366), ("Tourmaline, Paraiba", 367), ("Tourmaline, Red", 368), ("Zircon", 369), ("Zirconia, Cubic", 370) ] ), ("Other ", [("Pyrex (Borosilicate glass)", 401), ("Ruby", 402), ("Water ice", 403), ("Cryolite", 404), ("Acetone", 405), ("Ethanol", 406), ("Teflon", 407), ("Glycerol", 408), ("Acrylic glass", 409), ("Rock salt", 410), ("Crown glass (pure)", 411), ("Salt (NaCl)", 412), ("Polycarbonate", 413), ("PMMA", 414), ("PETg", 415), ("PET", 416), ("Flint glass (pure)", 417), ("Crown glass (impure)", 418), ("Fused Quartz", 419), ("Bromine", 420), ("Flint glass (impure)", 421), ("Cubic zirconia", 422), ("Moissanite", 423), ("Cinnabar (Mercury sulfide)", 424), ("Gallium(III) prosphide", 425), ("Gallium(III) arsenide", 426), ("Silicon", 427) ] ) ]
		
		self.iordict = {1:1.36, 2:1.36, 3:1.329, 4:1.345, 5:1.501, 6:1.000132, 7:1.00045, 8:1.34, 9:1.385, 10:1.351, 11:1.473, 12:1.504, 13:1.494, 14:1.484, 15:1.309, 16:1.35, 17:1.535, 18:1.481, 19:1.482, 20:1.473, 21:1.466, 22:1.47, 23:1.536, 24:1.361, 25:1.362, 26:1.38, 27:1.49, 28:1.472, 29:1.363, 30:1.33346, 31:1.31766, 32:1.33283, 33:1.356, 101:1.0, 102:1.0002926, 103:1.000293, 104:1.000036, 105:1.000132, 106:1.00045, 201:1.33, 202:1.38, 203:1.41, 204:1.34, 205:2.04, 206:1.52, 207:1.569, 208:1.669, 209:1.805, 210:1.459, 211:1.474, 212:1.495, 213:1.53, 214:1.50, 215:1.460, 216:1.488, 217:1.516, 301:1.544, 302:1.746, 303:1.75, 304:1.539, 305:1.532, 306:1.52, 307:1.629, 308:1.632, 309:1.567, 310:1.674, 311:1.57, 312:1.570, 313:1.544, 314:1.61, 315:1.532, 316:1.625, 317:1.486, 318:2.000, 319:1.746, 320:1.627, 321:2.417, 322:1.560, 323:1.560, 324:1.434, 325:1.72, 326:1.88, 327:1.880, 328:1.790, 329:1.73, 330:1.740, 331:1.739, 332:1.74, 333:1.490, 334:1.522, 335:1.64, 336:1.600, 337:1.660, 338:1.660, 339:1.560, 340:1.50, 341:1.518, 342:1.585, 343:1.50, 344:1.440, 345:1.430, 346:1.440, 347:1.560, 348:1.760, 349:1.53, 350:1.635, 351:1.544, 352:1.757, 353:1.757, 354:1.760, 355:1.79, 356:1.712, 357:1.712, 358:1.708, 359:1.76, 360:1.690, 361:1.607, 362:1.605, 363:1.603, 364:1.61, 365:1.61, 366:1.61, 367:1.61, 368:1.61, 369:1.777, 370:2.173, 401:1.47, 402:1.76, 403:1.31, 404:1.388, 405:1.36, 406:1.36, 407:1.35, 408:1.4729, 409:1.49, 410:1.516, 411:1.5, 412:1.544, 413:1.584, 414:1.4893, 415:1.57, 416:1.575, 417:1.6, 418:1.485, 419:1.46, 420:1.661, 421:1.523, 422:2.15, 423:2.419, 424:2.65, 425:3.02, 426:3.5, 427:3.927}

		volumes = self.libs[5]
		for type in self.iortree:
			for vol in type[1]:
				key = MakeKey("P2L", vol[0])
				volumes[key] = ('"clear"', ('"float fresnel" [%s]' % self.iordict[vol[1]],), ())
				
		# Special cases - just a fudge for now!
		self.iortree[4][1].append(("Tissue", 0))
		volumes["P2L/Tissue"] = ('"homogeneous"',('"float fresnel" [1.535]','"color g" [0.9 -0.9 -0.9]','"color sigma_a" [800 1380 1664]','"color sigma_s" [50 13 6]'), ())
				
	def AddMaterial(self, newdef):
		k = newdef.Key()
		key = MakeKey("CUSTOM", k)
		if not k in self.libs[1]["CUSTOM"]:
			self.libs[1]["CUSTOM"].append(k)
		self.libs[2][key] = (
			(False, 0),
			newdef.Type(),
			False, False, False, newdef.Bumpmap(), False,
			"", "", 0, newdef.InternalVolume(), newdef.ExternalVolume(), None, None, False)
		self.libs[3][key] = [x for x in newdef.Nodes()]
	
	def AutoloadMaterials(self, path, menu):
		if YN(path) == "Yes":
			if os.path.isdir(path):
				files = os.listdir(path)
				for f in files:
					if f.endswith('.xml'):
						file = '/'.join((path,f))
						self.LoadMaterialLibrary(file=file, menu=menu)		
		
	def GetAnalyser(self, mat, sm):
		type = mat.GetType()
		if type == "matte": return AnalyseMatte(mat, sm)
		elif type == "mattetranslucent": return AnalyseMatteTranslucent(mat, sm)
		elif type == "glossy": return AnalyseGlossy(mat, sm)
		elif type == "glossytranslucent": return AnalyseGlossyTranslucent(mat, sm)
		elif type == "metal": return AnalyseMetal(mat, sm)
		elif type == "shinymetal": return AnalyseShinyMetal(mat, sm)
		elif type == "mirror": return AnalyseMirror(mat, sm)
		elif type == "glass": return AnalyseGlass(mat, sm)
		elif type == "glass2": return AnalyseGlass2(mat, sm)
		elif type == "roughglass": return AnalyseRoughGlass(mat, sm)
		elif type == "carpaint": return AnalyseCarpaint(mat, sm)
		elif type == "scatter": return AnalyseScatter(mat, sm)
		elif type == "velvet": return AnalyseVelvet(mat, sm)
		elif type == "null": return AnalyseNull(mat, sm)
		elif type == "portal": return AnalysePortal(mat, sm)
		else:
			print "Warning: unknown material type '%s' used for %s, using matte instead" % (type, mat.Key())
			return AnalyseMatte(mat, sm)

	def GetBlenderVolumes(self):
		return self.iortree, self.iordict
		
	def GetInputs(self, key):
		return self.libs[3].get(key, None)
	
	def GetLibraries(self):
		global POSER_VERSION
		if POSER_VERSION == 6:
			newlist = self.libs[0][:]
			newlist.sort()
			return newlist
		else:
			return sorted(self.libs[0])
	
	def GetMaterial(self, key):
		matdef = LibDef()
		matdef.SetKey(key)
		matdef.SetType(self.libs[2][key][1])
		matdef.SetBumpmap(self.libs[2][key][5])
		matdef.SetVolumeInt(self.libs[2][key][10])
		matdef.SetVolumeExt(self.libs[2][key][11])
		for n in self.libs[3][key]:
			matdef.AddNode(self.ParseInput(n))
		
		return matdef
	
	def GetMaterials(self, library):
		global POSER_VERSION
		newlist = self.libs[1].get(library, None)
		if POSER_VERSION == 6:
			if newlist: newlist.sort()
		else:
			if newlist:	newlist = sorted(newlist)
		return newlist
	
	def GetParameters(self, key):
		return self.libs[2].get(key, None)
	
	def GetTexture(self, key):
		return self.libs[4].get(key, None)

	def GetVolume(self, key):
		return self.libs[5].get(key, None)

	def LoadCustomLibrary(self, root):
		cust = root.find('custom')
		if cust:
			parms = cust.findall('CustomMat')
			for p in parms:
				key = p.get('key')
				newdef = LibDef()
				newdef.SetKey(key.split('/', 1)[1])
				newdef.SetType(p.get('type'))
				newdef.SetBumpmap(eval(p.get('bumpmap')))
				newdef.SetVolumeExt(p.get('externalvolume'))
				newdef.SetVolumeInt(p.get('internalvolume'))
				nodes = p.findall('Node')
				for n in nodes:
					newdef.AddNode(n.get('node'))
				self.AddMaterial(newdef)
		
	def LoadMaterialLibrary(self, dir=None, file=None, menu=None ):
	
		(LIBRARIES, MATERIALS, PARAMETERS, INDICES, TEXTURES, VOLUMES) = range(6)
		
		if file == None:
			# Select library file from disk
			xmlfile = tkFileDialog.askopenfilename(
				defaultextension=[("*.xml")],
				filetypes=[("Pose2Lux material library", "*.xml")],
				initialdir=dir,
				initialfile=None,
				title="Load Pose2Lux material library")
		else:
			xmlfile = file
			
		if xmlfile == "": return
		
		# Make sure we have a valid XML file
		try:
			tree = parse(xmlfile)
			root = tree.getroot()
		except IOError, err:
			tkMessageBox.showerror("Pose2Lux", "Pose2Lux could not open the library file:\n%s" % xmlfile); return
		except Exception, err:
			tkMessageBox.showerror("Pose2Lux", "Pose2Lux encountered a problem with:\n%s\n%s" % (xmlfile, err)); return
		
		# Find all libraries. Exit if there aren't any
		libs = root.findall('library')
		if len(libs) == 0:
			tkMessageBox.showerror('Pose2Lux', 'No Pose2Lux libraries found in XML file')
			return
		
		# Add libraries
		for library in libs:
			libraryName = library.attrib.get('name', 'Unnamed library')
			if not libraryName in self.libs[LIBRARIES]:
				self.libs[LIBRARIES].append(libraryName)
				newlib = True
			else:
				newlib = False
			
			# Add materials
			mats = library.findall('material')
			if len(mats) == 0: continue		# Move on to the next library
			
			matnames = []
			for material in mats:
				
				show = material.attrib.get('show', 'yes')
			
				materialType = material.attrib.get('type', 'matte')
				thisMatName = material.attrib.get('name', 'Unnamed material')
				if not show == 'no': matnames.append(thisMatName)
			
				# Get parameters for material
				parms = MatParms(material.find('parameters'))
				parms.LuxType = materialType
				
				if materialType == 'mix':
					parms.Mix[0] = MakeKey(libraryName, material.find('mix_1').text)
					parms.Mix[1] = MakeKey(libraryName, material.find('mix_2').text)
					parms.Mix[2] = material.find('mix_amount').text
					
				# Get used textures
				textures = []
				tex = material.findall('used_texture')
				for t in tex:
					textures.append(t.text.replace('USE_LIB', libraryName))
				parms.Textures = textures
				
				# Get Volume
				vol = material.find('usevolume')
				if not vol is None:
					parms.VolName = vol.text.replace('USE_LIB', libraryName)

				# Get Exterior Volume
				vol = material.find('exterior_volume')
				if not vol is None:
					parms.ExtVolName = vol.text.replace('USE_LIB', libraryName)
					
				# Get displacement
				d0 = material.find('displacement')
				if not d0 is None:
					p0 = d0.text
					d1 = material.find('displacementscale')
					if d1:
						p1 = float(d1.text)
					else:
						p1 = 1.0
					d2 = material.find('sharpboundary')
					if d2:
						p2 = d2.text
					else:
						p2 = "false"
					d3 = material.find('displacementsplit')
					if d3:
						p3 = d3.text
					else:
						p3 = "false"
					parms.Disp = (p0, p1, p2, p3)
				
				# Add to LUXPARMS
				key = MakeKey(libraryName, thisMatName)
				self.libs[PARAMETERS][key] = parms.MakeTuple()
				
				# Get indices
				indices = material.findall('input')
				iList = [i.text for i in indices]
				self.libs[INDICES][key] = iList
			
			if newlib:
				if menu: menu[0].addOption(libraryName, menu[1])
				self.libs[MATERIALS][libraryName] = matnames
			else:
				for mat in matnames:
					if not mat in self.libs[MATERIALS][libraryName]:
						self.libs[MATERIALS][libraryName].append(mat)
			if menu: menu[2](None, None, None)
			
			# Add textures
			tex = library.findall('texture')
			for texture in tex:
				k = MakeKey(libraryName, texture.attrib.get('name', 'Unnamed texture'))
				type = texture.attrib.get('type')
				value = texture.attrib.get('value')
				inputs = []
				for i in texture.findall('input'):
					inputs.append(i.text)
				v = ('"%s" "%s"' % (type, value), inputs)
				self.libs[TEXTURES][k] = v
			
			# Add volumes
			volumes = library.findall('volume')
			for vol in volumes:
				k = MakeKey(libraryName, vol.attrib.get('name', 'Unnamed volume'))
				type = vol.attrib.get('type')
				inputs = []
				for i in vol.findall('input'):
					inputs.append(i.text)
				tex = []
				for t in vol.findall('used_texture'):
					tex.append(t.text.replace('USE_LIB', libraryName))
				v = ('"%s"' % type, inputs, tex)
				self.libs[VOLUMES][k] = v
				
	def MakeDefaultLuxMaterials(self):
		
		defaultLibs = ["Basic", "Car paint", "Gems", "Glasses", "Latex", "Metal", "Misc", "People", "Special", "Water"]
		defaultMats = {
			"Basic":["Glass", "Architectural glass", "Rough glass", "Glossy translucent, smooth", "Glossy translucent, fuzzy", "Glossy", "Not so glossy", "Smooth glossy", "Rough glossy", "Matte", "Matte smooth", "Matte rough", "Matte very rough", "Matte roughest", "Matte Translucent", "Matte Translucent smooth", "Matte Translucent rough", "Mirror", "Shiny metal", "Velvet, smooth", "Velvet, not so smooth", "Velvet, rough"],
			"Special" : ["Null", "Skydome"],
			"Metal":["Aluminium", "Amorphous carbon", "Brass", "Copper", "Copper, polished", "Gold", "Gunmetal", "Pewter", "Phos Bronze", "Platinum", "Rose Gold", "Silver", "Steel"],
			"People":["Basic Skin", "ODF skin", "RhiLau skin", "Snarly skin", "SSS Skin:Normal", "SSS Skin:Shiny", "SSS Skin:Dry", "SSS Skin:useHS", "Hair", "Lip", "Eyesocket", "Tear", "Cornea", "Sclera:iris:pupil", "Eye surface", "Generic nail enamel", "Other bits"],
			"Glasses":["Clear", "Coloured glass", "Greenish blue", "Translucent lilac", "Dark purple", "Ruby Red", "Black with blue sheen", "Rough glass"],
			"Car paint":["Ford F8", "Polaris Silber", "Opel Titan", "BMW 339", "2k Acrylack", "White", "Blue", "Blue matte", "Shifting green", "Vivid Purple", "Black Car Paint 1", "Black Car Paint 2"],
			"Misc":["Candle", "Bees Wax Basic", "Paraffin Wax Solid", "Jade",],
			"Gems":["Amethyst", "Amethyst Dark", "Aquamarine", "Blue Topaz London", "Blue Topaz Montana", "Blue Topaz Swiss", "Blue Zircon", "Chromium Diopside", "Chrome Tourmaline", "Citrine Golden", "Citrine Lemon", "Diamond", "Emerald", "Fire Opal", "Garnet Almandine", "Garnet Rhodolite", "Iolite", "Peridot", "Ruby", "Sapphire", "Tanzanite"],
			"Latex":["Aqua", "Black", "Bright Blue", "Bright Pink", "Bright Red", "Dark Red", "Generic", "Lime", "Purple", "Soft Pink", "White"],
			"Water":["Clear water, flat","Clear water, calm","Clear water, choppy","Clear water, rough"],
			"CUSTOM":[]
		}

			# (DEPTH, LUXTYPE, TEXTURE, DIFFUSE, SPECULAR, BUMP, REFLECTION, MIX_1, MIX_2, MIXAMOUNT, VOLUME, EXTVOLUME, USEDTEXTURES, (DISPMAP, DMSCALE, BOUNDARY), SKIN) = range(15)
		LibParms = {
			"Basic/Glass" : ((False,0), "glass", False, False, False, False, True, "", "", 0, None, None, None, None, False),
			"Basic/Architectural glass" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Basic/Rough glass" : ((False,0), "roughglass", False, False, False, False, True, "", "", 0, None, None, None, None, False),
			"Basic/Glossy translucent, smooth" : ((False,0), "glossytranslucent", True, True, True, True, True, "", "", 0, None, None, None, None, False),
			"Basic/Glossy translucent, fuzzy" : ((False,0), "glossytranslucent", True, True, True, True, True, "", "", 0, None, None, None, None, False),
			"Basic/Glossy" : ((False,0), "glossy", True, True, True, True, True, "", "", 0, None, None, None, None, False),
			"Basic/Not so glossy" : ((False,0), "glossy", True, True, True, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Smooth glossy" : ((False,0), "glossy", True, True, True, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Rough glossy" : ((False,0), "glossy", True, True, True, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte" : ((False,0), "matte", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte smooth" : ((False,0), "matte", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte rough" : ((False,0), "matte", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte very rough" : ((False,0), "matte", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte roughest" : ((False,0), "matte", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte Translucent" : ((False,0), "mattetranslucent", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte Translucent smooth" : ((False,0), "mattetranslucent", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Matte Translucent rough" : ((False,0), "mattetranslucent", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Mirror" : ((False,0), "mirror", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Basic/Shiny metal" : ((False,0), "shinymetal", False, True, False, True, True, "", "", 0, None, None, None, None, False),
			"Basic/Velvet, smooth":((False,0), "velvet", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Velvet, not so smooth":((False,0), "velvet", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"Basic/Velvet, rough":((False,0), "velvet", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			
			"Gems/Amethyst" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Amethyst Dark" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Aquamarine" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Blue Topaz London" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Blue Topaz Montana" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Blue Topaz Swiss" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Blue Zircon" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Chromium Diopside" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Chrome Tourmaline" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Citrine Golden" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Citrine Lemon" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Diamond" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Emerald" : ((False,0), "mix", False, False, False, False, False, "Gems/Emerald@glass", "Gems/Emerald@mattetranslucent", 0.015, None, None, None, None, False),
			"Gems/Emerald@glass" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Emerald@mattetranslucent" : ((False,0), "mattetranslucent", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Fire Opal" : ((False,0), "mix", False, False, False, False, False, "Gems/Fire Opal@glass", "Gems/Fire Opal@mattetranslucent", 0.03, None, None, None, None, False),
			"Gems/Fire Opal@glass" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Fire Opal@mattetranslucent" : ((False,0), "mattetranslucent", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Garnet Almandine" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Garnet Rhodolite" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Iolite" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Peridot" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Ruby" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Sapphire" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Gems/Tanzanite" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			
			"Latex/Black" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/White" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Bright Red" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Dark Red" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Bright Pink" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Soft Pink" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Bright Blue" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Aqua" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Lime" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Purple" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Latex/Generic" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			
			"p2l/Clay" : ((False,0), "matte", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			
			"Special/Null" : ((False,0), "null", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Special/Skydome" : ((False,0), "null", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			
			"Metal/Amorphous carbon" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Silver" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Gold" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Aluminium" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Copper" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Copper, polished" : ((False,0), "metal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			
			"Metal/Rose Gold" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Brass" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Steel" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Pewter" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Platinum" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Phos Bronze" : ((False,0), "shinymetal", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"Metal/Gunmetal" : ((False,0), "glossy", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			
			"People/Basic Skin" : ((False,0), "glossy", True, True, True, True, False, "", "", 0, None, None, None, None, False),
			"People/ODF skin@mt" : ((False,0), "mattetranslucent", True, True, False, False, False, "", "", 0, None, None, None, None, False),
			"People/ODF skin@gl" : ((False,0), "glossy", True, True, False, False, False, "", "", 0, None, None, None, None, False),
			"People/ODF skin" : ((False,0), "mix", False, False, False, False, False, "People/ODF skin@mt", "People/ODF skin@gl", 0.1, None, None, None, None, False),
			"People/Snarly skin@glossy" : ((False,0), "glossy", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Snarly skin@matte" : ((False,0), "mattetranslucent", False, False, False, False, False, "", "", 0, "p2l/Snarly skin", None, None, None, False),
			"People/Snarly skin" : ((False,0), "mix", False, False, False, False, False, "People/Snarly skin@glossy", "People/Snarly skin@matte", 0.8, None, None, None, None, False),
			
			"People/SSS Skin:Normal" : ((False,0), "mix", False, False, False, False, False, "People/Epidermis:normal", "Special/Null", 0.11, None, None, None, None, False),
			"People/SSS Skin:Shiny" : ((False,0), "mix", False, False, False, False, False, "People/Epidermis:shiny", "Special/Null", 0.11, None, None, None, None, False),
			"People/SSS Skin:Dry" : ((False,0), "mix", False, False, False, False, False, "People/Epidermis:dry", "Special/Null", 0.11, None, None, None, None, False),
			"People/SSS Skin:useHS" : ((False,0), "mix", False, False, False, False, False, "People/Epidermis:useHS", "Special/Null", 0.11, None, None, None, None, False),
			"People/Epidermis:normal" : ((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "People/Tissue", None, None, None, False),
			"People/Epidermis:shiny" : ((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "People/Tissue", None, None, None, False),
			"People/Epidermis:dry" : ((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "People/Tissue", None, None, None, False),
			"People/Epidermis:useHS" : ((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "People/Tissue", None, None, None, False),
			
			"People/RhiLau skin" : ((False,0), "mix", False, False, False, False, False, "People/Epidermis1Rough", "Special/Null", 0.10, None, None, None, None, False),
			"People/Epidermis1Rough" : ((False,0), "glossytranslucent", False, False, False, True, False, "", "", 0, "People/SkinRough", None, ("People/color@USE_KEY",), None, False),
			
			"People/Hair" : ((False,0), "mix", False, False, False, False, False, "People/Hair Transmapped@mattetranslucent", "People/Hair Transmapped@glass", 0.18, None, None, None, None, False),
			"People/Hair Transmapped@mattetranslucent" : ((False,0), "mattetranslucent", True, False, False, True, False, "", "", 0, None, None, None, None, False),
			"People/Hair Transmapped@glass" : ((False,0), "glass", True, False, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Lip" : ((False,0), "glossy", True, True, False, True, False, "", "", 0, "People/Tissue", None, None, None, False),
			"People/Eyesocket" : ((False,0), "glossy", True, True, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Tear" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Cornea" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Sclera:iris:pupil" : ((False,0), "glossy", True, True, False, False, False, "", "", 0, None, None, None, None, False),
			"People/Eye surface" : ((False,0), "glass", False, False, False, True, False, "", "", 0, None, None, None, None, False),
			"People/Generic nail enamel" : ((False,0), "carpaint", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			"People/Other bits" : ((False,0), "glossy", True, True, False, True, False, "", "", 0, None, None, None, None, False),
			
			"Glasses/Clear" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Greenish blue" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Translucent lilac@gl" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Translucent lilac@mt" : ((False,0), "matte", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Translucent lilac" : ((False,0), "mix", False, False, False, False, False, "Glasses/Translucent lilac@gl", "Glasses/Translucent lilac@mt", 0.3, None, None, None, None, False),
			"Glasses/Dark purple" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Ruby Red" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Black with blue sheen" : ((False,0), "glass", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Coloured glass" : ((False,0), "glass", True, True, False, False, False, "", "", 0, None, None, None, None, False),
			"Glasses/Rough glass" : ((False,0), "roughglass", True, True, False, False, True, "", "", 0, None, None, None, None, False),
			
			"Car paint/Ford F8" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Polaris Silber" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Opel Titan" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/BMW 339" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/2k Acrylack" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/White" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Blue" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Blue matte" : ((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Shifting green":((False,0), "carpaint", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Vivid Purple":((False,0), "glossy", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Black Car Paint 1":((False,0), "glossy", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			"Car paint/Black Car Paint 2":((False,0), "glossy", False, False, False, False, False, "", "", 0, None, None, None, None, False),
			
			# (TYPE, LUXTYPE, TEXTURE, DIFFUSE, SPECULAR, BUMP, REFLECTION, MIX_1, MIX_2, MIXAMOUNT, VOLUME, EXTVOLUME, USEDTEXTURES, (DISPMAP, DMSCALE, BOUNDARY), SKIN) = range(15)
			"Misc/Jade":((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "p2l/Jade", None, ('p2l/jade::Kd:amount','p2l/jade::Kd','p2l/jade::Kd.scale'), None, False),
			"Misc/Candle":((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "p2l/Wax", None, None, None, False),
			"Misc/Bees Wax Basic":((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "p2l/BeesWax", None, None, None, False),
			"Misc/Paraffin Wax Solid":((False,0), "glossytranslucent", False, False, False, False, False, "", "", 0, "p2l/ParaffinWax", None, None, None, False),
			"Water/Clear water, flat":((False,0), "glass2", False, False, False, False, False, "", "", 0, "p2l/clear water", None, ("p2l/water@disp","p2l/water@scale"), ("p2l/water@scale", 0.0, "false", "false"), False),
			"Water/Clear water, calm":((False,0), "glass2", False, False, False, False, False, "", "", 0, "p2l/clear water", None, ("p2l/water@disp","p2l/water@scale"), ("p2l/water@scale", 0.5, "false", "false"), False),
			"Water/Clear water, choppy":((False,0), "glass2", False, False, False, False, False, "", "", 0, "p2l/clear water", None, ("p2l/water@disp","p2l/water@scale"), ("p2l/water@scale", 1, "false", "false"), False),
			"Water/Clear water, rough":((False,0), "glass2", False, False, False, False, False, "", "", 0, "p2l/clear water", None, ("p2l/water@disp","p2l/water@scale"), ("p2l/water@scale", 1.5, "false", "false"), False)
		}
		
		LibIndices = {
			"Basic/Glass" : ('"color Kr" [1.0 1.0 1.0]','"color Kt" [1.0 1.0 1.0]','"float index" [1.52]'),
			"Basic/Architectural glass" : ('"color Kr" [1.0 1.0 1.0]','"color Kt" [1.0 1.0 1.0]','"float index" [1.52]','"bool architectural" ["true"]'),
			"Basic/Rough glass" : ('"color Kr" [1.0 1.0 1.0]','"color Kt" [1.0 1.0 1.0]','"float index" [1.50]'),
			"Basic/Glossy translucent, smooth" : ('"bool multibounce" ["false"]',),
			"Basic/Glossy translucent, fuzzy" : ('"bool multibounce" ["true"]',),
			"Basic/Glossy" : (),
			"Basic/Not so glossy" : ('"float uroughness" [0.2]', '"float vroughness" [0.2]'),
			"Basic/Smooth glossy" : ('"float uroughness" [0.3]', '"float vroughness" [0.3]'),
			"Basic/Rough glossy" : ('"float uroughness" [0.5]', '"float vroughness" [0.5]'),
			"Basic/Matte" : (),
			"Basic/Matte smooth" : ('"float sigma" [1.0]',),
			"Basic/Matte rough" : ('"float sigma" [5.0]',),
			"Basic/Matte very rough" : ('"float sigma" [10.0]',),
			"Basic/Matte roughest" : ('"float sigma" [20.0]',),
			"Basic/Matte Translucent" : ('"float sigma" [1.0]', '"bool energyconserving" ["true"]'),
			"Basic/Matte Translucent smooth" : ('"float sigma" [8.0]', '"bool energyconserving" ["true"]'),
			"Basic/Matte Translucent rough" : ('"float sigma" [16.0]', '"bool energyconserving" ["true"]'),
			"Basic/Mirror" : (),
			"Basic/Shiny metal" : (),
			
			"Gems/Amethyst" : ('"color Kr" [1.0 0.77 0.86]', '"color Kt" [0.58 0.29 0.74]', '"float index" [1.554]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Amethyst Dark" : ('"color Kr" [0.99 0.64 0.92]', '"color Kt" [0.18 0.02 0.26]', '"float index" [1.554]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Aquamarine" : ('"color Kr" [0.88 0.99 1.0]', '"color Kt" [0.76 0.97 0.92]', '"float index" [1.590]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Blue Topaz London" : ('"color Kr" [0.79 1.0 1.0]', '"color Kt" [0.25 0.56 0.67]', '"float index" [1.610]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Blue Topaz Montana" : ('"color Kr" [0.53 0.64 0.70]', '"color Kt" [0.01 0.29 0.50]', '"float index" [1.610]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Blue Topaz Swiss" : ('"color Kr" [0.75 0.90 1.0]', '"color Kt" [0.43 0.83 1.0]', '"float index" [1.610]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Blue Zircon" : ('"color Kr" [0.76 0.99 0.98]', '"color Kt" [0.2 0.82 0.90]', '"float index" [1.8]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Chromium Diopside" : ('"color Kr" [0.85 0.97 0.45]', '"color Kt" [0.02 0.40 0.0]', '"float index" [1.73]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Chrome Tourmaline" : ('"color Kr" [0.67 1.0 0.76]', '"color Kt" [0.13 0.39 0.15]', '"float index" [1.64]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Citrine Golden" : ('"color Kr" [0.98 1.0 0.36]', '"color Kt" [0.98 0.66 0.00]', '"float index" [1.550]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Citrine Lemon" : ('"color Kr" [0.99 0.99 0.85]', '"color Kt" [0.96 0.88 0.25]', '"float index" [1.550]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Diamond" : ('"color Kr" [0.99 0.99 0.99]', '"color Kt" [0.95 0.95 0.95]', '"float index" [2.417]', '"float cauchyb" [0.005]', '"float film" [10.0]', '"float filmindex" [1.3]'),
			"Gems/Emerald@glass" : ('"color Kr" [0.58 1.0 0.89]', '"color Kt" [0.04 0.40 0.24]', '"float index" [1.576]', '"float cauchyb" [0.0]'),
			"Gems/Emerald@mattetranslucent" : ('"color Kr" [0.14 0.50 0.34]', '"color Kt" [0.14 0.50 0.34]', '"float sigma" [0.5]'),
			"Gems/Fire Opal@glass" : ('"color Kr" [1.0 0.85 0.28]', '"color Kt" [0.96 0.24 0.04]', '"float index" [1.460]', '"float film" [30.0]', '"float filmindex" [1.3]', '"float cauchyb" [0.0003]'),
			"Gems/Fire Opal@mattetranslucent" : ('"color Kr" [1.0 0.95 0.38]', '"color Kt" [0.87 0.44 0.28]', '"float sigma" [0.5]'),
			"Gems/Garnet Almandine" : ('"color Kr" [0.75 0.14 0.27]', '"color Kt" [0.22 0.05 0.07]', '"float index" [1.760]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Garnet Rhodolite" : ('"color Kr" [0.99 0.71 0.98]', '"color Kt" [0.26 0.07 0.17]', '"float index" [1.760]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Iolite" : ('"color Kr" [0.55 0.62 0.80]', '"color Kt" [0.22 0.32 0.65]', '"float index" [1.548]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Peridot" : ('"color Kr" [0.98 1.0 0.43]', '"color Kt" [0.51 0.59 0.11]', '"float index" [1.654]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Ruby" : ('"color Kr" [1.0 0.57 0.85]', '"color Kt" [0.85 0.02 0.20]', '"float index" [1.760]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Sapphire" : ('"color Kr" [0.78 0.93 0.97]', '"color Kt" [0.03 0.10 0.65]', '"float index" [1.760]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),
			"Gems/Tanzanite" : ('"color Kr" [0.67 0.62 0.85]', '"color Kt" [0.24 0.34 0.96]', '"float index" [1.691]', '"float cauchyb" [0.0]', '"float film" [3.0]', '"float filmindex" [1.3]'),

			"Latex/Black" : ('"color Kd" [0.0 0.0 0.0]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/White" : ('"color Kd" [0.93 0.93 0.93]','"color Ka" [0.0 0.0 0.0]','"color Ks" [1 1 1]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Bright Red" : ('"color Kd" [0.71 0.0 0.01]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Dark Red" : ('"color Kd" [0.25 0.0 0.0]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Bright Pink" : ('"color Kd" [1.0 0.25 0.91]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Soft Pink" : ('"color Kd" [0.91 0.72 0.76]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Bright Blue" : ('"color Kd" [0.00 0.40 0.63]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Aqua" : ('"color Kd" [0.42 0.80 0.82]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Lime" : ('"color Kd" [0.62 0.79 0.28]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Purple" : ('"color Kd" [0.40 0.09 0.32]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			"Latex/Generic" : ('"color Kd" [USE_KD]','"color Ka" [0.0 0.0 0.0]','"color Ks" [0.95 0.95 0.95]','"float index" [1.5]','"float uroughness" [0.015]','"float vroughness" [0.015]','"bool multibounce" ["false"]'),
			
			"p2l/Clay" : ('"color Kd" [0.7 0.7 0.7]',),
			
			"Special/Null" : (),
			"Special/Skydome" : (),
			"Basic/Velvet, smooth":('"float thickness" [0.1]',),
			"Basic/Velvet, not so smooth":('"float thickness" [0.3]',),
			"Basic/Velvet, rough":('"float thickness" [0.6]',),
			"Metal/Amorphous carbon" : ('"string name" ["amorphous carbon"]',),
			"Metal/Silver" : ('"string name" ["silver"]', '"float uroughness" [0.05]', '"float vroughness" [0.05]'),
			"Metal/Gold" : ('"string name" ["gold"]', '"float uroughness" [0.025]', '"float vroughness" [0.025]'),
			"Metal/Aluminium" : ('"string name" ["aluminium"]',),
			"Metal/Copper, polished" : ('"string name" ["copper"]', '"float uroughness" [0.025]', '"float vroughness" [0.025]'),
			"Metal/Copper" : ('"string name" ["copper"]', '"float uroughness" [0.125]', '"float vroughness" [0.125]'),
			
			"Metal/Rose Gold" : ('"float uroughness" [0.02]', '"float vroughness" [0.02]', '"color Kr" [0.45 0.13 0.08]', '"color Ks" [0.85 0.53 0.48]', '"float film" [0.1]', '"float filmindex" [1.2]'),
			"Metal/Brass" : ('"float uroughness" [0.02]', '"float vroughness" [0.02]', '"color Kr" [0.46 0.35 0.07]', '"color Ks" [0.66 0.55 0.27]', '"float film" [0.1]', '"float filmindex" [1.2]'),
			"Metal/Steel" : ('"float uroughness" [0.02]', '"float vroughness" [0.02]', '"color Kr" [0.33 0.32 0.28]', '"color Ks" [0.74 0.74 0.74]', '"float film" [0.1]', '"float filmindex" [1.2]'),
			"Metal/Pewter" : ('"float uroughness" [0.04]', '"float vroughness" [0.04]', '"color Kr" [0.27 0.27 0.26]', '"color Ks" [0.37 0.37 0.36]', '"float film" [0.15]', '"float filmindex" [1.01]'),
			"Metal/Platinum" : ('"float uroughness" [0.02]', '"float vroughness" [0.02]', '"color Kr" [0.36 0.36 0.32]', '"color Ks" [0.99 0.95 0.95]', '"float film" [0.1]', '"float filmindex" [1.2]'),
			"Metal/Phos Bronze" : ('"float uroughness" [0.4]', '"float vroughness" [0.4]', '"color Kr" [0.40 0.22 0.15]', '"color Ks" [0.40 0.22 0.15]', '"float film" [0.15]', '"float filmindex" [1.01]'),
			"Metal/Gunmetal" : ('"float uroughness" [0.02]', '"float vroughness" [0.02]', '"color Kd" [0.0 0.0 0.0]', '"color Ks" [0.69 0.69 0.64]', '"float index" [0.20]'),
			
			"People/Basic Skin" : ('"float uroughness" [0.3]', '"float vroughness" [0.3]'),
			"People/ODF skin@mt" : ('"color Kt" [0.25 0.1 0.1]', '"float sigma" [0.5]'),
			"People/ODF skin@gl" : ('"color Ks" [0.15 0.2 0.25]', '"float uroughness" [0.236836278392]', '"float vroughness" [0.236836278392]'),
			"People/Snarly skin@glossy" : ('"texture Kd" [USE_KdMAP]', '"color Ks" [0.08 0.10 0.10]','"color Ka" [0.000 0.114 0.137]','"float d" [0.300]','"texture bumpmap" [USE_BUMPMAP]','"float uroughness" [0.38]','"float vroughness" [0.38]','"float index" [0.0]','"bool multibounce" ["true"]'),
			"People/Snarly skin@matte" : ('"texture Kr" [USE_KdMAP]', '"color Kt" [0.40 0.35 0.35]','"bool energyconserving" ["true"]'),
			
			"People/Epidermis1Rough" : ('"float d" [0.025]','"color Ka" [0.0144 0.029 0.223]','"bool onesided" ["false"]','"bool multibounce" ["false"]','"texture Kd" ["People/color@USE_KEY"]','"float index" [1.3]','"float uroughness" [0.28]','"float vroughness" [0.28]','"bool backface_multibounce" ["false"]','"float backface_d" [0.015]','"color backface_Ka" [0.0 0.0 0.0]','"color backface_Ks" [0.0 0.0 0.0]','"float backface_index" [0]','"float backface_uroughness" [0.8]','"float backface_vroughness" [0.8]'),
			
			"People/Epidermis:normal" : ('"texture bumpmap" [USE_BUMPMAP]','"float d" [0.025]','"color Ka" [0.03 0.028 0.22]','"bool onesided" ["false"]','"bool multibounce" ["false"]','"texture Kd" [USE_KdMAP]','"float index" [1.535]','"color Ks" [USE_KS]','"color Kt" [0.85 0.85 0.85]','"float uroughness" [0.3]','"float vroughness" [0.3]','"bool backface_multibounce" ["false"]','"float backface_d" [0.015]','"color backface_Ka" [0.0 0.0 0.0]','"color backface_Ks" [0.0 0.0 0.0]','"float backface_index" [0]','"float backface_uroughness" [0.8]','"float backface_vroughness" [0.8]'),
			"People/Epidermis:shiny" : ('"texture bumpmap" [USE_BUMPMAP]','"float d" [0.025]','"color Ka" [0.03 0.028 0.22]','"bool onesided" ["false"]','"bool multibounce" ["false"]','"texture Kd" [USE_KdMAP]','"float index" [1.535]','"color Ks" [USE_KS]','"color Kt" [0.85 0.85 0.85]','"float uroughness" [0.2]','"float vroughness" [0.2]','"bool backface_multibounce" ["false"]','"float backface_d" [0.015]','"color backface_Ka" [0.0 0.0 0.0]','"color backface_Ks" [0.0 0.0 0.0]','"float backface_index" [0]','"float backface_uroughness" [0.8]','"float backface_vroughness" [0.8]'),
			"People/Epidermis:dry" : ('"texture bumpmap" [USE_BUMPMAP]','"float d" [0.025]','"color Ka" [0.03 0.028 0.22]','"bool onesided" ["false"]','"bool multibounce" ["false"]','"texture Kd" [USE_KdMAP]','"float index" [1.535]','"color Ks" [USE_KS]','"color Kt" [0.85 0.85 0.85]','"float uroughness" [0.4]','"float vroughness" [0.4]','"bool backface_multibounce" ["false"]','"float backface_d" [0.015]','"color backface_Ka" [0.0 0.0 0.0]','"color backface_Ks" [0.0 0.0 0.0]','"float backface_index" [0]','"float backface_uroughness" [0.8]','"float backface_vroughness" [0.8]'),
			"People/Epidermis:useHS" : ('"texture bumpmap" [USE_BUMPMAP]','"float d" [0.025]','"color Ka" [0.03 0.028 0.22]','"bool onesided" ["false"]','"bool multibounce" ["false"]','"texture Kd" [USE_KdMAP]','"float index" [1.535]','"color Ks" [USE_KS]','"color Kt" [0.85 0.85 0.85]','"float uroughness" [USE_HS]','"float vroughness" [USE_HS]','"bool backface_multibounce" ["false"]','"float backface_d" [0.015]','"color backface_Ka" [0.0 0.0 0.0]','"color backface_Ks" [0.0 0.0 0.0]','"float backface_index" [0]','"float backface_uroughness" [0.8]','"float backface_vroughness" [0.8]'),
			
			"People/Hair Transmapped@mattetranslucent" : ('"bool energyconserving" ["true"]','"texture Kt" [USE_KdMAP]','"float sigma" [0.6]'),
			"People/Hair Transmapped@glass" : ('"float index" [1.55]',),
			"People/Lip" : ('"color Ks" [0.000 0.000 0.000]','"float index" [0.0]','"float uroughness" [0.057864]','"float vroughness" [0.057864]'),
			"People/Eyesocket" : ('"color Ks" [0.007 0.001 0.002]','"float index" [0.0]','"float uroughness" [0.09]','"float vroughness" [0.09]','"color Ka" [0.000 0.114 0.137]','"float d" [0.200]'),
			"People/Tear" : ('"color Kr" [1.000 1.000 1.000]','"color Kt" [0.992 0.992 0.992]','"float index" [1.4]','"bool architectural" ["true"]'),
			"People/Cornea" : ('"color Kr" [1.000 1.000 1.000]','"texture Kt" [USE_KdMAP]','"float index" [1.4]','"bool architectural" ["true"]'),
			"People/Sclera:iris:pupil" : ('"color Ks" [0.007 0.001 0.002]','"float index" [0.0]','"float uroughness" [0.09]','"float vroughness" [0.09]'),
			"People/Eye surface" : ('"color Kr" [1.000 1.000 1.000]','"texture Kt" [USE_KdMAP]','"float index" [1.4]','"bool architectural" ["true"]'),
			"People/Other bits" : ('"float uroughness" [0.5]','"float vroughness" [0.5]','"color Ks" [0 0 0]'),
			"People/Generic nail enamel" : ('"color Ks1" [0.80000001 0.80000001 0.80000001]','"color Ks2" [0.07323252 0.07323252 0.07323252]','"color Ks3" [0.03740784 0.03740784 0.03740784]','"float M1" [0.032999999821186]','"float M2" [0.054999999701977]','"float M3" [0.100000001490116]','"float R1" [0.079999998211861]','"float R2" [0.029999999329448]','"float R3" [0.059999998658895]'),
			
			"Glasses/Clear" : ('"color Kr" [1.0 1.0 1.0]', '"color Kt" [1.0 1.0 1.0]', '"float index" [1.52]'),
			"Glasses/Greenish blue" : ('"color Kr" [1.0 1.0 1.0]', '"color Kt" [0.95 0.97 0.95]', '"float index" [1.6]'),
			"Glasses/Translucent lilac@gl" : ('"color Kr" [0.7 0.7 0.7]', '"color Kt" [0.345424920321 0.290522754192 0.583333333333]', '"float index" [2.0]', '"float cauchyb" [0.1]'),
			"Glasses/Translucent lilac@mt" : ('"color Kd" [0.545424920321 0.490522754192 0.783333333333]',),
			"Glasses/Translucent lilac" : (),
			"Glasses/Dark purple" : ('"color Kr" [0.7 0.0 0.7]', '"color Kt" [0.4 0.0 0.4]', '"float index" [1.5]'),
			"Glasses/Ruby Red" : ('"color Kr" [0.752941184243 0.0734640540431 0.0]', '"color Kt" [0.452941184243 0.0434640540431 0.0]', '"float index" [1.5]', '"float cauchyb" [0.5]'),
			"Glasses/Black with blue sheen" : ('"color Kr" [0.0 0.0 0.51]', '"color Kt" [0.0 0.0 0.0]', '"float index" [1.52]'),
			"Glasses/Coloured glass" : ('"float index" [1.52]',),
			"Glasses/Rough glass" : ('"float index" [1.52]',),
			
			"Car paint/Ford F8" : ('"string name" ["ford f8"]',),
			"Car paint/Polaris Silber" : ('"string name" ["polaris silber"]',),
			"Car paint/Opel Titan" : ('"string name" ["opel titan"]',),
			"Car paint/BMW 339" : ('"string name" ["bmw339"]',),
			"Car paint/2k Acrylack" : ('"string name" ["2k acrylack"]',),
			"Car paint/White" : ('"string name" ["white"]',),
			"Car paint/Blue" : ('"string name" ["blue"]',),
			"Car paint/Blue matte" : ('"string name" ["blue matte"]',),
			"Car paint/Shifting green":('"color Kd" [0.124526 0.294963 0.203289]', '"color Ks1" [0.754852 0.773852 0.500392]', '"color Ks2" [0.442047 0.629024 0.290352]', '"color Ks3" [0.227329 0.546751 0.374624]', '"float R1" [1.000000]', '"float R2" [1.000000]', '"float R3" [1.000000]', '"float M1" [1.000000]', '"float M2" [1.000000]', '"float M3" [1.000000]'),
			"Car paint/Vivid Purple":('"color Kd" [0.1 0.0 0.50]', '"color Ks" [0.5 0.0 0.7]', '"color Ka" [0.43 0.48 0.22]', '"float uroughness" [-0.005]', '"float vroughness" [0.005]', '"float d" [0.4]', '"float index" [2.0]'),
			"Car paint/Black Car Paint 1":('"color Kd" [0.0 0.0 0.0]', '"color Ks" [0.02 0.02 0.02]', '"float uroughness" [-0.258086156304]', '"float vroughness" [0.258086156304]', '"float film" [0.1]', '"float index" [2.3]'),
			"Car paint/Black Car Paint 2":('"color Kd" [0.0 0.0 0.0] ', '"color Ks" [0.02 0.02 0.02]', '"color Ka" [0.0 0.0 0.0]', '"float uroughness" [-0.005]', '"float vroughness" [0.005]', '"float d" [0.1]', '"float index" [2.0]'),
			
			"Misc/Jade":('"texture Kd" ["p2l/jade::Kd.scale"]','"color Kt" [1.000000 1.000000 1.000000]','"color Ks" [0.060778 0.060778 0.060778]','"float index" [0.0]','"float uroughness" [0.010000]','"float vroughness" [0.010000]','"bool multibounce" ["false"]'),
			"Misc/Candle":('"bool onesided" ["true"]','"bool multibounce" ["false"]','"color Kt" [0.9 0.9 0.9]','"color Kd" [0.25 0.25 0.25]','"color Ks" [0.28 0.28 0.28]','"float index" [0.0]','"float uroughness" [0.075]','"float vroughness" [0.075]'),
			"Misc/Bees Wax Basic":('"bool onesided" ["true"]','"bool multibounce" ["false"]','"color Kt" [0.3 0.3 0.3]','"color Kd" [USE_KD]','"color Ks" [0.28 0.28 0.28]','"float index" [1.443]','"float uroughness" [0.050]','"float vroughness" [0.050]'),
			"Misc/Paraffin Wax Solid":('"bool onesided" ["true"]','"bool multibounce" ["false"]','"color Kt" [0.3 0.3 0.3]','"color Kd" [USE_KD]','"color Ks" [0.28 0.28 0.28]','"float index" [1.5366]','"float uroughness" [0.050]','"float vroughness" [0.050]'),
			
			"Water/Clear water, flat":('"bool architectural" ["false"]','"bool dispersion" ["false"]'),
			"Water/Clear water, calm":('"bool architectural" ["false"]','"bool dispersion" ["false"]'),
			"Water/Clear water, choppy":('"bool architectural" ["false"]','"bool dispersion" ["false"]'),
			"Water/Clear water, rough":('"bool architectural" ["false"]','"bool dispersion" ["false"]')
		}
		
		LibTextures = {
			"p2l/jade::Kd:amount":('"float" "blender_clouds"',('"string type" ["default"]','"string noisetype" ["hard_noise"]','"float noisesize" [0.100000]','"integer noisedepth" [6]','"string noisebasis" ["blender_original"]','"float bright" [1.000000]','"float contrast" [4.000000]','"vector scale" [0.200000 0.200000 0.200000]','"vector rotate" [0 0 0]','"vector translate" [-0.000000 -0.000000 -0.000000]','"float tex1" [0.0]', '"float tex2" [1.0]')),
			"p2l/jade::Kd":('"color" "mix" ',('"texture amount" ["p2l/jade::Kd:amount"]','"color tex1" [0.116576 0.238828 0.116576]','"color tex2" [0.680021 0.796917 0.612066]')),
			"p2l/jade::Kd.scale":('"color" "scale"',('"texture tex1" ["p2l/jade::Kd"]','"color tex2" [0.549303 0.549303 0.549303]')),
			"p2l/jade:volume":('"fresnel" "constant"',('"float value" [1.640000]',)),
			"p2l/water":('"fresnel" "constant"',('"float value" [1.33]',)),
			"p2l/water@disp":('"float" "blender_distortednoise"',('"float distamount" [1.6] ','"float noisesize" [0.25]','"float nabla" [0.025]','"string type" ["blender_original"]','"string noisebasis" ["blender_original"]','"float bright" [1.0]','"float contrast" [0.5]','"vector scale" [1.0 1.0 1.0]','"vector rotate" [0 0 0]','"vector translate" [0 0 0]')),
			"p2l/water@scale":('"float" "scale"',('"texture tex1" ["p2l/water@disp"]','"float tex2" [0.1]')),
			"p2l/skin":('"fresnel" "constant"',('"float value" [1.45]',)),
			"People/color@USE_KEY":('"color" "imagemap"', ('"string filename" [USE_KdTEX]','"integer discardmipmaps" [0]','"string filtertype" ["bilinear"]','"float gain" [1.0]','"float gamma" [3.0]','"float maxanisotropy" [8.0]','"string wrap" ["repeat"]','"string mapping" ["uv"]','"float udelta" [0.0]','"float uscale" [1.0]','"float vscale" [-1.0]','"float vdelta" [0.0]'))
		}
		
		LibVolumes = {
			"p2l/Jade":('"homogeneous" "texture fresnel" ["p2l/jade:volume"]',('"color sigma_a" [236.413304397 157.521281655 236.413304397]','"color sigma_s" [37.116092 50.000000 37.116092]','"color g" [0.0 0.0 0.0]'),('p2l/jade:volume',)),
			"p2l/clear water":('"clear" "texture fresnel" ["p2l/water"]',('"color absorption" [0 0 0]',), ("p2l/water",)),
			"p2l/Snarly skin":('"homogeneous" "texture fresnel" ["p2l/skin"]',('"color sigma_a" [0.0 0.0 0.0]','"color sigma_s" [40.0 40.0 40.0]','"color g" [0.5 0.5 0.5]'), ("p2l/skin",)),
			"People/Tissue":('"homogeneous"',('"float fresnel" [1.535]','"color g" [0.9 -0.9 -0.9]','"color sigma_a" [800 1380 1664]','"color sigma_s" [50 13 6]'), ()),
			"People/SkinRough":('"homogeneous"',('"float fresnel" [1.3]','"color g" [0.7 -0.9 -0.9]','"color sigma_a" [150.0 690.45 831.60]','"color sigma_s" [150.0 37.7 18.0]'), ()),
			"p2l/Wax":('"homogeneous"',('"float fresnel" [1.35]','"color g" [0.0 0.0 0.0]','"color sigma_a" [30 50 50]','"color sigma_s" [0.01 0.01 0.01]'), ()),
			"p2l/BeesWax":('"homogeneous"',('"float fresnel" [1.443]','"color g" [0.0 0.0 0.0]','"color sigma_a" [40 60 60]','"color sigma_s" [0.1 0.1 0.1]'), ()),
			"p2l/ParaffinWax":('"homogeneous"',('"float fresnel" [1.5366]','"color g" [0.0 0.0 0.0]','"color sigma_a" [20 80 80]','"color sigma_s" [0.1 0.1 0.1]'), ()),
			"P2L/Tissue":('"homogeneous"',('"float fresnel" [1.535]','"color g" [0.9 -0.9 -0.9]','"color sigma_a" [800 1380 1664]','"color sigma_s" [50 13 6]'), ()),
		}
		
		defaultData = (defaultLibs, defaultMats, LibParms, LibIndices, LibTextures, LibVolumes)
		
		return defaultData
	
	def MakeMaterial(self, attr, material, key):
	
		parameters = self.GetParameters(key)
		
		if parameters:
		
			# Convert dictionary parameters to a LibMaterial object 
			libmat = LibMaterial(parameters)
			
			# Define which Lux material it is
			attr.AddNode("string", "type", libmat.GetType())
			
			# The 'mix' material is a special case
			if libmat.GetType() != "mix":
				# Analyse all the 'USE_KD' style parameters
				analyser = self.GetAnalyser(libmat, material.simplemat)
				for i in analyser.Inputs():
					attr.AddNode(i[0], i[1], i[2])
					
				# Now get the associated indices from the library
				inputs = self.GetInputs(key)
				if inputs:
					for i in inputs:
						p = self.ParseInput(i)
						attr.AddNode(p[0], p[1], p[2])
			
		else:
			libmat = None
			
		return libmat
		
	def MakeTexture(self, texKey, attr):
	
		# Get the texture parameters
		t = self.GetTexture(texKey)
		if t:
			attr.SetTextureType(t[0])
			for i in t[1]:
				p = self.ParseInput(i)
				attr.AddNode(p[0], p[1], p[2])
				
	def MakeVolume(self, volKey, attr):
		
		v = self.GetVolume(volKey)
		if v:
			attr.SetVolumeType(v[0])
			for i in v[1]:
				p = self.ParseInput(i)
				attr.AddNode(p[0], p[1], p[2])
		return v[2]
	
	def ParseInput(self, index):
		global POSER_VERSION
		index = index.strip()
		x, z = index.split('[')
		x, y = x[1:-2].split()
		if POSER_VERSION == 6:
			if z.startswith('"'): z = z[1:]
			if z.endswith('"]'): z = z[:-2]
			elif z.endswith('"') or z.endswith(']'): z = z[:-1]
		else:
			z = z.strip('"]')
		return x, y, z
	
	def ParseTexture(self, index):
		pass

	def SaveCustomLibrary(self, root):
		
		cust = Element('custom')
		root.append(cust)
		
		for m in self.libs[1]['CUSTOM']:
			key = MakeKey('CUSTOM', m)
			p = self.libs[2][key]
			attrib = {'key' : key, 'type' : p[1], 'bumpmap' : str(p[5]), 'internalvolume' : str(p[10]), 'externalvolume' : str(p[11])}
			element = Element('CustomMat', attrib)
			cust.append(element)
			i = self.libs[3].get(key, [])
			for node in i:
				element.append(Element('Node', {'node' : node}))
		
	def ValidMaterial(self, mat):
		return self.libs[2].has_key(mat)
class LibDef:
	def __init__(self):
		self.bumpmap = False
		self.key = None
		self.nodes = []
		self.type = None
		self.volext = None
		self.volint = None
	
	def AddNode(self, value):
		self.nodes.append(value)
	
	def Bumpmap(self):
		return self.bumpmap
	
	def ExternalVolume(self):
		return self.volext
		
	def InternalVolume(self):
		return self.volint
	
	def Key(self):
		return self.key
		
	def Nodes(self):
		return self.nodes
		
	def Print(self):
		print "Key:", self.key
		print "Type:", self.type
		print "Bumpmap?", self.bumpmap
		print "External Volume:", self.volext
		print "Internal Volume:", self.volint
		print "Nodes:"
		for n in self.nodes: print n
	
	def SetBumpmap(self, value):
		self.bumpmap = value
		
	def SetKey(self, value):
		self.key = value
	
	def SetType(self, value):
		self.type = value
		
	def SetVolumeExt(self, vol):
		if vol == None or vol == "None":
			self.volext = None
		else:
			self.volext = vol
		
	def SetVolumeInt(self, vol):
		if vol == None or vol == "None":
			self.volint = None
		else:		
			self.volint = vol

	def Type(self):
		return self.type
class LibObject:
	def __init__(self, parms):
		self.parms = parms
		self.type = self.GetType()
		self.used_textures = self.GetUsedTextures()
	
	def GetType(self):
		return ("Unknown")
		
	def GetUsedtextures(self):
		return None
		
	def Type(self):
		return self.type
		
	def UsedTextures(self):
		return self.used_textures
class LibMaterial(LibObject):
	def __init__(self, parms):
		LibObject.__init__(self, parms)
		
		self.backface = parms[0][0]
		self.backface_amount = parms[0][1]
		self.use_kdmap = parms[2]
		self.use_kd = parms[3]
		self.use_ks = parms[4]
		self.use_bump = parms[5]
		self.use_kr = parms[6]
		self.mix1 = parms[7]
		self.mix2 = parms[8]
		self.mix_amount = parms[9]
		self.volume_int = parms[10]
		self.volume_ext = parms[11]
		self.displacement = parms[13]	
		self.skin = parms[14]
		
	def Displacement(self):
		return self.displacement
	
	def ExteriorVolume(self):
		return self.volume_ext
	
	def GetType(self):
		return self.parms[1]
		
	def GetUsedTextures(self):
		if self.parms[12]:
			return [x for x in self.parms[12]]
		else:
			return []
			
	def InteriorVolume(self):
		return self.volume_int
			
	def Mix(self, opt):
		if opt == 1: return self.mix1
		else: return self.mix2
		
	def MixAmount(self):
		return self.mix_amount
class MatParms:
	def __init__(self, p=None):
		
		self.OtherSide = [False, 0]
		self.LuxType = ""
		self.Use = [False,False,False,False,False]
		self.Mix = ["", "", 0]
		self.VolName = None
		self.ExtVolName = None
		self.Disp = None
		self.Skin = False
		self.Textures = []
		
		if not p == None: self.GetParameters(p)
	
	def GetParameters(self, p):
		self.OtherSide[0] = self.GetBool(p.attrib.get('MAKE2SIDED'))
		self.OtherSide[1] = float(p.attrib.get('THICKNESS', 0))
		self.Use[0] = self.GetBool(p.attrib.get('USE_KdMAP'))
		self.Use[1] = self.GetBool(p.attrib.get('USE_KD'))
		self.Use[2] = self.GetBool(p.attrib.get('USE_KS'))
		self.Use[3] = self.GetBool(p.attrib.get('USE_BUMPMAP'))
		self.Use[4] = self.GetBool(p.attrib.get('USE_KR'))
		self.Skin = self.GetBool(p.attrib.get('USE_SKIN'))
		
	def MakeTuple(self):
		return (
			self.OtherSide,
			self.LuxType,
			self.Use[0],
			self.Use[1],
			self.Use[2],
			self.Use[3],
	        self.Use[4],
			self.Mix[0],
			self.Mix[1],
			self.Mix[2],
			self.VolName,
			self.ExtVolName,
			self.Textures,
			self.Disp,
			self.Skin
			)
			
	def GetBool(self, x):
		if x:
			return (False, True)[x.lower() == "yes"]	
		else:
			return False
class MultiListbox(Frame):
	def __init__(self, master, h, lists, multiselect=False, labelpad=3, preselect=None, postselect=None):
		Frame.__init__(self, master)
		self.postselect = postselect
		self.preselect = preselect
		self.lists = [  ]
		self.show = {}
		for l, w, sort_command in lists:
			frame = Frame(self)
			self.show[l] = frame
			if sort_command:
				Button(frame, text=l, borderwidth=1, relief=RAISED, command=sort_command).pack(fill=X)
			else:
				Label(frame, text=l, anchor=W, borderwidth=1, relief=RAISED, pady=labelpad).pack(fill=X)
			lb = Listbox(frame, width=w, height=h, borderwidth=0, selectborderwidth=0,
						 relief=FLAT, exportselection=FALSE)
			lb.pack(expand=YES, fill=BOTH)
			self.lists.append(lb)
			lb.bind('<B1-Motion>', lambda e, s=self: s._select(e.y))
			lb.bind('<Button-1>', lambda e, s=self: s._select(e.y))
			if multiselect:
				lb.bind('<Control-Button-1>', lambda e, s=self: s._ctlselect(e.y))
				lb.bind('<Shift-Button-1>', lambda e, s=self: s._shiftselect(e.y))
			lb.bind('<Leave>', lambda e: 'break')
			lb.bind('<B2-Motion>', lambda e, s=self: s._b2motion(e.x, e.y))
			lb.bind('<Button-2>', lambda e, s=self: s._button2(e.x, e.y))
		self.sbframe = Frame(self)
		#frame.pack(side=LEFT, fill=Y)
		Label(frame, borderwidth=1, relief=RAISED).pack(fill=X)
		self.sb = Scrollbar(self.sbframe, orient=VERTICAL, command=self._scroll)
		self.sb.pack(expand=YES, fill=Y)
		self.lists[0]['yscrollcommand'] = self.sb.set
		self.current=0
	def _select(self, y):
		row = self.lists[0].nearest(y)
		if self.preselect: self.preselect(row)
		self.selection_clear(0, END)
		self.selection_set(row)
		if self.postselect: self.postselect(row)
		return 'break'
	def _ctlselect(self, y):
		row = self.lists[0].nearest(y)
		self.selection_toggle(row)
		return 'break'
	def _shiftselect(self, y):
		row = self.lists[0].nearest(y)
		self.selection_clear(0, END)
		self.selection_set(self.current, row)
		return 'break'
	def _button2(self, x, y):
		for l in self.lists:
			l.scan_mark(x, y)
		return 'break'
	def _b2motion(self, x, y):
		for l in self.lists:
			l.scan_dragto(x, y)
		return 'break'
	def _scroll(self, *args):
		for l in self.lists:
			apply(l.yview, args)
		return 'break'
	def curselection(self):
		return self.lists[0].curselection( )
	def delete(self, first, last=None):
		for l in self.lists:
			l.delete(first, last)
	def Display(self, columns):
		for x in self.show.values():
			x.pack_forget()
		self.sbframe.pack_forget()
		for c in columns:
			self.show[c].pack(side=LEFT, expand=YES, fill=BOTH)
		self.sbframe.pack(side=LEFT, expand=YES, fill=BOTH)	
	def get(self, first, last=None):
		result = [  ]
		for l in self.lists:
			result.append(l.get(first,last))
		if last: return apply(map, [None] + result)
		return result	
	def	getCurrent(self):
		result = [  ]
		for l in self.lists:
			result.append(l.get(self.current))
		return result
	def index(self, index):
		self.lists[0].index(index)
	def insert(self, index, *elements):
		for e in elements:
			i = 0
			for l in self.lists:
				l.insert(index, e[i])
				i = i + 1
	def nearest(self, y):
		return self.lists[0].nearest(y)
	def size(self):
		return self.lists[0].size( )
	def scrollbar(self):
		return self.sb
	def see(self, index):
		for l in self.lists:
			l.see(index)
	def selection_anchor(self, index):
		for l in self.lists:
			l.selection_anchor(index)
			self.current=index
	def selection_clear(self, first, last=None):
		for l in self.lists:
			l.selection_clear(first, last)
	def selection_includes(self, index):
		return self.lists[0].selection_includes(index)
	def selection_set(self, first, last=None):
		self.current=first
		for l in self.lists:
			l.selection_set(first, last)
	def selection_toggle(self, index):
		if self.lists[0].selection_includes(index) == 1:
			self.selection_clear(index)
		else:
			self.selection_set(index)
			self.current=index			
	def yview_scroll(self, number, what):
		for l in self.lists:
			l.yview_scroll(number, what)
class Notebook(object):
	def __init__(self, master, side=LEFT ):
		self.active_fr = None
		self.count = 0
		self.choice = IntVar(0)
		if side in (TOP, BOTTOM): self.side = LEFT
		else: self.side = TOP
		self.rb_fr = Frame(master, borderwidth=2, relief=SUNKEN)
		self.rb_fr.pack(side=side, fill=BOTH)
		self.screen_fr = Frame(master)
		self.screen_fr.pack(fill=BOTH, padx="3m", pady="3m", expand=1)
		
		self.buttons = {}
		self.btn_order = []
		
	def __call__(self):
		return self.screen_fr

	def ShowHide(self, mode, button=None):
		if mode == "advanced":
			for b in self.btn_order: b.pack_forget()
			for b in self.btn_order: b.pack(fill=BOTH, side=self.side)
		elif mode == "hide":
			b = self.buttons[button]
			b.pack_forget()
	
	def AddScreen(self, fr, title):
		b = Radiobutton(self.rb_fr, text=title, indicatoron=0,
			variable=self.choice,
			value=self.count, command=lambda: self.display(fr))
		b.pack(fill=BOTH, side=self.side)
		self.buttons[title] = b; self.btn_order.append(b)
		
		if not self.active_fr:
			fr.pack(fill=BOTH, expand=1)
			self.active_fr = fr
		self.count += 1
		
	def display(self, fr):
		self.active_fr.forget( )
		fr.pack(fill=BOTH, expand=1)
		self.active_fr = fr	
	
	def DisplayByName(self, name):
		btn = self.buttons.get(name, None)
		if btn: btn.invoke()
class OptionMenu(tk.OptionMenu):
	def __init__(self, *args, **kw):
		self._command = kw.get("command")
		tk.OptionMenu.__init__(self, *args, **kw)
	def addOption(self, label, variable):
		self["menu"].add_command(label=label, command=tk._setit(variable, label, self._command))
class ProgressBarView:
  
  def __init__(self, master=None, orientation='horizontal',
	  min=0, max=100, width=100, height=None,
	  doLabel=1, appearance=None,
	  fillColor=None, background=None,
	  labelColor=None, labelFont=None,
	  labelText='', labelFormat="%d%%",
	  value=0.1, bd=2):
	# preserve various values
	self.master=master
	self.orientation=orientation
	self.min=min
	self.max=max
	self.doLabel=doLabel
	self.labelText=labelText
	self.labelFormat=labelFormat
	self.value=value
	if (fillColor == None) or (background == None) or (labelColor == None):
	  # We have no system color names under linux. So use a workaround.
	  #btn = Button(font=labelFont)
	  btn = Button(master, text='0', font=labelFont)
	  if fillColor == None:
		fillColor  = btn['foreground']
	  if background == None:
		background = btn['disabledforeground']
	  if labelColor == None:
		labelColor = btn['background']
	if height == None:
	  l = Label(font=labelFont)
	  height = l.winfo_reqheight()
	self.width      = width
	self.height     = height
	self.fillColor  = fillColor
	self.labelFont  = labelFont
	self.labelColor = labelColor
	self.background = background
	#
	# Create components
	#
	self.frame=Frame(master, relief=appearance, bd=bd, width=width, height=height)
	self.canvas=Canvas(self.frame, bd=0,
		highlightthickness=0, background=background, width=width, height=height)
	self.scale=self.canvas.create_rectangle(0, 0, width, height,
		fill=fillColor)
	self.label=self.canvas.create_text(width / 2, height / 2,
		text=labelText, anchor=CENTER, fill=labelColor, font=self.labelFont)
	self.canvas.pack(fill=BOTH)
	self.update()
	self.canvas.bind('<Configure>', self.onResize) # monitor size changes

  def onResize(self, event):
	if (self.width == event.width) and (self.height == event.height):
	  return
	# Set new sizes
	self.width  = event.width
	self.height = event.height
	# Move label
	self.canvas.coords(self.label, event.width/2, event.height/2)
	# Display bar in new sizes
	self.update()

  def updateProgress(self, newValue, newMax=None, newCol=None):
	if newMax:
	  self.max = newMax
	self.value = newValue
	if newCol:
		self.canvas.itemconfigure(self.scale, fill=newCol)
	self.update()

  def pack(self, *args, **kw):
	self.frame.pack(*args, **kw)

  def update(self):
	# Trim the values to be between min and max
	value=self.value
	if value > self.max:
		value = self.max
	if value < self.min:
		value = self.min
	# Adjust the rectangle
	if self.orientation == "horizontal":
		self.canvas.coords(self.scale, 0, 0,
		float(value) / self.max * self.width, self.height)
	else:
		self.canvas.coords(self.scale, 0,
		self.height - (float(value) / self.max*self.height),
		self.width, self.height)
	# And update the label
	if self.doLabel:
		if value:
			if value >= 0:
				pvalue = int((float(value) / float(self.max)) * 100.0)
			else:
				pvalue = 0
			self.canvas.itemconfig(self.label, text=self.labelFormat % pvalue)
		else:
			self.canvas.itemconfig(self.label, text='')
	else:
		#self.canvas.itemconfig(self.label, text=self.labelFormat % self.labelText)
		pass
	#self.canvas.update_idletasks()
	self.canvas.update()
class VerticalScrolledFrame(Frame):
	def __init__(self, parent, *args, **kw):
		Frame.__init__(self, parent, *args, **kw)            

		# create a canvas object and a vertical scrollbar for scrolling it
		vscrollbar = Scrollbar(self, orient=VERTICAL)
		vscrollbar.pack(fill=Y, side=RIGHT, expand=FALSE)
		canvas = Canvas(self, bd=0, highlightthickness=0, yscrollcommand=vscrollbar.set)
		canvas.pack(side=LEFT, fill=BOTH, expand=TRUE)
		vscrollbar.config(command=canvas.yview)

		# reset the view
		canvas.xview_moveto(0)
		canvas.yview_moveto(0)

		# create a frame inside the canvas which will be scrolled with it
		self.interior = interior = Frame(canvas)
		interior_id = canvas.create_window(0, 0, window=interior, anchor=NW)

		# track changes to the canvas and frame width and sync them, also updating the scrollbar
		def _configure_interior(event):
			# update the scrollbars to match the size of the inner frame
			size = (interior.winfo_reqwidth(), interior.winfo_reqheight())
			canvas.config(scrollregion="0 0 %s %s" % size)
			if interior.winfo_reqwidth() != canvas.winfo_width():
				# update the canvas's width to fit the inner frame
				canvas.config(width=interior.winfo_reqwidth())
		interior.bind('<Configure>', _configure_interior)

		def _configure_canvas(event):
			if interior.winfo_reqwidth() != canvas.winfo_width():
				# update the inner frame's width to fill the canvas
				canvas.itemconfigure(interior_id, width=canvas.winfo_width())
		canvas.bind('<Configure>', _configure_canvas)

		return
		
# Custom controls
class Boolean(object):
	def __init__(self, parent, label=None, text=None, initValue=False, callback=None):
		
		self.varBool = IntVar()
		self.varBool.set(initValue)
		if callback: self.varBool.trace('w', callback)
		
		self.frame = Frame(parent)
		self.frame.pack(side=TOP, fill=X)
		
		if label:
			Label(self.frame, text="%s: " % label).pack(side=LEFT)
		
		Checkbutton(self.frame, text=text, variable=self.varBool).pack(side=LEFT)
		
	def get(self):
		return self.varBool.get()
		
	def set(self, value):
		self.varBool.set(value)		
class LabelEntry(object):
	def __init__(self, parent, initValue=None, text=None, type="int", formatstring=None, expand=None, minVal=None, maxVal=None, callback=None):
	
		self.initValue = initValue or ""
		self.minVal = minVal
		self.maxVal = maxVal
		self.text = text
		self.type = type
		self.template = formatstring or "%.8f"
	
		# Use own frame
		self.frame = Frame(parent)
		self.frame.pack(side=TOP, fill=X)
		
		# Label
		if text:
			l = "%s: " % text
			Label(self.frame, text=l).pack(side=LEFT)
			
		# Value
		self.varEntry = StringVar()
		i = self.template % float(initValue)
		self.varEntry.set(i)
		if callback: self.varEntry.trace('w', callback)
		self.distance = Label(self.frame, textvariable=self.varEntry, anchor=W)
		
		if expand == 0:
			self.distance.pack(side=LEFT)
		else:
			self.distance.pack(side=LEFT, fill=X, expand=1)
		
		# Button
		Button(self.frame, text="#", command=self.input).pack(side=LEFT)
		
	def get(self):
		if self.type == "int":
			return int(self.varEntry.get())
		else:
			return float(self.varEntry.get())
	
	def input(self):
		v = self.varEntry
		title = self.text
		
		if self.minVal == 0:
			minimum = 0
		else:
			minimum = self.minVal or "anything"
		if self.maxVal == 0:
			maximum = 0
		else:
			maximum = self.maxVal or "anything"
			
		if self.type == "float":
			text = 'Enter a floating point value\nbetween %s and %s' % (minimum, maximum)
			result = tkSimpleDialog.askfloat(title, text, initialvalue=v.get(), minvalue=(minimum, None)[minimum == "anything"], maxvalue=(maximum, None)[maximum == "anything"])
			if result: result = self.template % result
		else:
			text = 'Enter an integer\nbetween %s and %s' % (minimum, maximum)
			result = tkSimpleDialog.askinteger(title, text, initialvalue=v.get(), minvalue=(minimum, None)[minimum == "anything"], maxvalue=(maximum, None)[maximum == "anything"])
		
		if result: v.set(result)		
	
	def set(self, value):
		self.varEntry.set(value)
class LabelledMenu(object):
	def __init__(self, parent, variable, list, label=None, callback=None):
	
		self.variable = variable
		if callback: self.variable.trace('w', callback)
		
		# Use our own frame
		self.frame = Frame(parent)
		self.frame.pack(side=TOP, fill=X)
		
		# Label
		if label:
			l = "%s: " % label
			Label(self.frame, text=l).pack(side=LEFT)
			
		self.menu = OptionMenu(self.frame, variable, *list)
		self.menu.pack(side=LEFT)
		
	def addOption(self, newgroup, var):
		self.menu.addOption(newgroup, var)
		
	def get(self):
		return self.variable.get()
	
	def ReplaceList(self, newlist):
		self.menu.pack_forget()
		self.menu = OptionMenu(self.frame, self.variable, *newlist)
		self.menu.pack(side=LEFT)
	
	def set(self, value):
		self.variable.set(value)
class PickColour(object):
	def __init__(self, parent, label=None, width=20, initcol=(1,1,1), rgb=True, callback=None):
		
		self.frame = Frame(parent)
		self.callback = callback
		self.colour = initcol
		self.rgb = rgb
		
		if label:
			Label(self.frame, text="%s: " % label).pack(side=LEFT)
		
		self.button = Button(self.frame, width=width, bg=PoserColToHTML(initcol), command=self.SelectColour)
		self.button.pack(side=LEFT)
		
		if rgb:
			self.varRGB = StringVar()
			self.varRGB.set(self.FormatColourLabel(initcol))
			Label(self.frame, textvariable=self.varRGB).pack(side=LEFT, padx=3)
	
	def FormatColourLabel(self, col):
		return '(R: %03d G:%03d B:%03d)' % PoserColToRGB(col)
		
	def get(self):
		return self.colour
		
	def pack(self, side=TOP, fill=X, expand=0, padx=0, pady=0):
		self.frame.pack(side=side, fill=fill, expand=expand, padx=padx, pady=pady)
	
	def pack_forget(self):
		self.frame.pack_forget()
	
	def SelectColour(self):
		newcol = tkColorChooser.askcolor(PoserColToHTML(self.colour), title="Pick a colour")
		if newcol[0]:
			self.set(RGBToPoserCol(newcol[0]))
	
	def set(self, colour):
		self.colour = colour
		self.button.config(bg=PoserColToHTML(self.colour))
		if self.rgb: self.varRGB.set(self.FormatColourLabel(self.colour))
		if self.callback: self.callback()
class PickFile(object):
	def __init__(self,
		parent,
		label=None,
		initDir=None,
		pady=0,
		callback=None):
		
		self.initdir = initDir
		
		# Our own frame
		self.frame = Frame(parent)
		self.frame.pack(side=TOP, fill=X)
		
		# The label
		if label:
			Label(self.frame, text=label).pack(side=LEFT)
			
		# The entry box
		self.varEntry = StringVar()
		if callback: self.varEntry.trace('w', callback)
		self.entry = Entry(self.frame, textvariable=self.varEntry)
		self.entry.pack(side=LEFT, fill=X, expand=1, pady=pady)
		
		# The button
		Button(self.frame, text="#", command=self.input).pack(side=LEFT)
		
	def get(self):
		return self.varEntry.get()
		
	def input(self):
		initDir = self.initdir
		initFile=None
		result = tkFileDialog.askopenfilename(
			defaultextension=[("*.ies")],
			filetypes=[("IES data file", "*.ies")],
			initialdir=initDir,
			initialfile=initFile,
			title="Select IES file")
		if result != "": self.varEntry.set(result)	
		
	def set(self, value):
		self.varEntry.set(value)
class PickScale(object):
	def __init__(self, 
		parent, 
		type="int", 
		orient=HORIZONTAL, 
		length=None, 
		from_=1, to=10, 
		resolution=1, 
		variable=None, 
		label=None,
		prefix=None,
		callback=None,
		initval=None):
		
		self.type = type
		if variable:
			self.variable = variable
		else:
			if self.type == "int":
				self.variable = IntVar()
			else:
				self.variable = DoubleVar()
				
		if initval:
			self.variable.set(initval)
		
		if callback:
			self.variable.trace('w', callback)
	
		# Our own frame
		self.frame = Frame(parent)
		self.frame.pack(side=TOP, fill=X)
		
		# Label prefix
		if prefix:
			Label(self.frame, text="%s: " % prefix).pack(side=LEFT)
		
		#The scale
		self.scale = Scale(self.frame, orient=orient, length=length, from_=from_, to=to, resolution=resolution, variable=self.variable, label=label)
		self.scale.pack(side=LEFT, fill=X, expand=1)
		
		# The button
		self.button = Button(self.frame, text="#", command=self.input)
		self.button.pack(side=BOTTOM, pady=3)
	
	def get(self):
		return self.scale.get()
		
	def input(self):
		s = self.scale
		title = s.cget("label")
	
		if self.type == "float":
			text = 'Enter a value from %s to %s\nin units of %s' % (s.cget("from"), s.cget("to"), s.cget("resolution"))
			result = tkSimpleDialog.askfloat(title, text)
		else:
			text = 'Enter a value from %s to %s\nin units of %s' % (int(s.cget("from")), int(s.cget("to")), int(s.cget("resolution")))
			result = tkSimpleDialog.askinteger(title, text)
		
		if result: self.scale.set(result)
		
	def set(self, value):
		self.scale.set(value)
		
# Utility functions
def xrot(p, drot):
        s = sin(radians(drot))
        c = cos(radians(drot))
        x,y,z = p
        return x, y * c - z * s, y * s + z * c
def yrot(p, drot):
        s = sin(radians(drot))
        c = cos(radians(drot))
        x,y,z = p
        return z * s + x * c, y, z * c - x * s
def zrot(p, drot):
        s = sin(radians(drot))
        c = cos(radians(drot))
        x,y,z = p
        return x * c - y * s, x * s + y * c, z
def rollPitchYaw(vec, roll, pitch, yaw):
	vec = zrot(vec, roll)
	vec = xrot(vec, pitch)
	vec = yrot(vec, yaw)
	return vec
def add3(a, b): return a[0] + b[0], a[1] + b[1], a[2] + b[2]
def dotp(a, b):	return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]
def mcol(wm, c): return wm[0][c], wm[1][c], wm[2][c]
def mul3by1(a, b): return a[0] * b, a[1] * b, a[2] * b
def mul3(a, b): return a[0] * b[0], a[1] * b[1], a[2] * b[2]
def str3(v): return ' '.join([str(x) for x in v])
def xfm(x): return x[0], -x[2], x[1]
def bw(x): return x[0] + x[1] + x[2]

def ChooseLoadFile(initDir, initFile, extension, type):
	
	path = tkFileDialog.askopenfilename(
		defaultextension=[("*.%s" % extension)],
		filetypes=[("Pose2Lux %s" % type, "*.%s" % extension)],
		initialdir=initDir,
		initialfile=initFile,
		title="Load Pose2Lux %s" % type)
	
	input = None	
	if not path == "":
		try:
			input = open(path)
		except:
			pass
	
	return path
def ChooseSaveFile(initDir, initFile, extension, type):
	
	result = tkFileDialog.asksaveasfilename(
		defaultextension=[("*.%s" % extension)],
		filetypes=[("Pose2Lux %s" % type, "*.%s" % extension)],
		initialdir=initDir,
		initialfile=initFile,
		title="Save Pose2Lux %s" % type)
	
	output = None	
	if not result == "":
		try:
			output = open(result, 'w')
		except:
			tkMessageBox.showwarning("Pose2Lux message", "Unable to write to settings file")
			output = None
	
	return output
def ColourValue(dc, v):
	newcolour = (dc[0]*v, dc[1]*v, dc[2]*v)
	return newcolour
def GC(colours):
	return [x**2 for x in colours]
def RGBToPoserCol(rgb):
	return tuple([x/255.0 for x in rgb])
def HTMLColorToRGB(colorstring):
    """ convert #RRGGBB to an (R, G, B) tuple """
    colorstring = colorstring.strip()
    if colorstring[0] == '#': colorstring = colorstring[1:]
    if len(colorstring) != 6:
        raise ValueError, "input #%s is not in #RRGGBB format" % colorstring
    r, g, b = colorstring[:2], colorstring[2:4], colorstring[4:]
    r, g, b = [int(n, 16) for n in (r, g, b)]
    return (r, g, b)
def LuxColour(colour, factor=None):
	if factor:
		r = colour[0] * factor
		g = colour[1] * factor
		b = colour[2] * factor
	else:
		r, g, b = colour
	return '%s %s %s' % (r, g, b)
def MakeKey(owner, name):
	return "/".join([owner, name])	
def MakeMenu(opts, menulist):
	menu = []
	for m in menulist:
		menu.append(opts[m][0])
	return menu
def MakeTuple(value):
	r, g, b = value.split()
	return (float(r), float(g), float(b))
def PoserColToHTML(col_tuple):
	rgb = tuple([int(x*255) for x in col_tuple])
	return RGBToHTMLColor(rgb)
def PoserColToRGB(col_tuple):
	rgb = tuple([int(x*255) for x in col_tuple])
	return rgb
def RGBToHTMLColor(rgb_tuple):
	hexcolor = '#%02x%02x%02x' % rgb_tuple
	return hexcolor
def Unpack(value, type):
	if type in ("integer", "float", "bool", "tuple"):
		return eval(value)
	elif type == "string":
		if value == "None":
			return None
		else:
			return value
def YN(x):
	try :
		result = ("No", "Yes")[x]
	except:
		result = ("Yes", "No")[x == None or x == ""]
	return result
	
# Initialise and run
root = Tk( )
defaultColour = root.cget("bg")
titleFont = tkFont.Font ( family="Helvetica", size=12, weight="bold" )
nb = Notebook(root, LEFT)
gui = GUIManager(root, nb)